//go:generate go mod vendor
//go:generate protoc --go_out=plugins=grpc,paths=source_relative:. --proto_path=. --proto_path=vendor/gitlab.mashroom.com/mashroom/backend mashroom_listing.proto
//go:generate rm -r vendor
//go:generate mockgen -package=mocklisting -destination=mock/client.go . ListingServiceClient
//go:generate mockgen -package=mocklisting -destination=mock/client_publication.go . ListingPublicationServiceClient
//go:generate go fmt ./...
//go:generate go fix ./...
//go:generate goimports -w ./
//go:generate go mod tidy

// Package listing provides invoice service api
package listing

import (
	"strings"

	_ "gitlab.mashroom.com/mashroom/backend/common/api/generic"       // required vendor
	_ "gitlab.mashroom.com/mashroom/backend/common/api/generic/query" // required vendor
	_ "gitlab.mashroom.com/mashroom/backend/common/api/protobuf"      // required vendor
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	_ "gitlab.mashroom.com/mashroom/backend/geo/api" // required vendor
)

// BPM workflow process ID.
const (
	PropertyListProcessID           = "property-list-process"
	PropertyInviteProcessID         = "property-invite-process"
	PropertyOfferProcessID          = "property-offer-process"
	TenancyPostTerminationProcessID = "tenancy-post-termination"
)

// VersionSpecPropertyID returns version selector specification by user ID, property ID and version statuses
func VersionSpecPropertyID(userID, propertyID string, statuses ...VersionStatus) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_Select{
			Select: &VersionSpecSelect{
				Statuses: statuses,
				Id: &VersionSpecSelect_PropertyId{
					PropertyId: propertyID,
				},
			},
		},
	}
}

// VersionSpecListingID returns version selector specification by user ID, listing ID and version statuses
func VersionSpecListingID(userID, listingID string, statuses ...VersionStatus) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_Select{
			Select: &VersionSpecSelect{
				Statuses: statuses,
				Id: &VersionSpecSelect_ListingId{
					ListingId: listingID,
				},
			},
		},
	}
}

// VersionSpecTenancyID returns version selector specification by user ID, tenancy ID and version statuses
func VersionSpecTenancyID(userID, tenancyID string, statuses ...VersionStatus) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_Select{
			Select: &VersionSpecSelect{
				Statuses: statuses,
				Id: &VersionSpecSelect_TenancyId{
					TenancyId: tenancyID,
				},
			},
		},
	}
}

// VersionSpecVersionID returns version selector specification by user ID and version ID
func VersionSpecVersionID(userID, versionID string) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_VersionId{
			VersionId: versionID,
		},
	}
}

// VersionSpecClientID returns version selector specification by user ID and client ID
func VersionSpecClientID(userID, clientID string) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_ClientId{
			ClientId: clientID,
		},
	}
}

// VersionSpecInvitationID returns version selector specification by user ID and invitation ID
func VersionSpecInvitationID(userID, invitationID string) *VersionSpec {
	return &VersionSpec{
		UserId: userID,
		Details: &VersionSpec_InvitationId{
			InvitationId: invitationID,
		},
	}
}

// FormatAddress returns full address string
func (x *Location) FormatAddress() string {
	flat := strings.TrimSpace(strings.ReplaceAll(strings.ToLower(x.GetFlat().GetValue()), "flat", ""))

	if flat != "" {
		flat = "Flat " + flat
	}

	return util.JoinNonblankStrings(
		" ",
		flat,
		x.GetBuilding().GetValue(),
		x.GetStreet().GetValue(),
		x.GetCity().GetValue(),
		x.GetPostcode().GetValue(),
	)
}

// Source returns VersionOriginSource calculated from VersionOrigin fields
func (x *VersionOrigin) Source() VersionOriginSource {
	switch {
	case x.GetListing():
		return VersionOriginSource_VERSION_ORIGIN_SOURCE_LISTING
	case x.GetPreface():
		return VersionOriginSource_VERSION_ORIGIN_SOURCE_PREFACE
	case x.GetTool() != nil:
		switch {
		case x.GetTool().GetValuation():
			return VersionOriginSource_VERSION_ORIGIN_SOURCE_TOOL_VALUATION
		case x.GetTool().GetExpenseTracker():
			return VersionOriginSource_VERSION_ORIGIN_SOURCE_TOOL_EXPENSE_TRACKER
		case x.GetTool().GetCertificateRepository():
			return VersionOriginSource_VERSION_ORIGIN_SOURCE_TOOL_CERTIFICATE_REPOSITORY
		default:
			return VersionOriginSource_VERSION_ORIGIN_SOURCE_UNKNOWN
		}
	case x.GetProduct() != nil:
		return VersionOriginSource_VERSION_ORIGIN_SOURCE_PRODUCT_PURCHASE
	default:
		return VersionOriginSource_VERSION_ORIGIN_SOURCE_UNKNOWN
	}
}
