package main

import (
	"context"
	"fmt"
	"net"
	"net/http"
	"os"
	"time"

	gmailparser "gitlab.mashroom.com/mashroom/backend/gmailparser/api"

	offer "gitlab.mashroom.com/mashroom/backend/offer/api"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/worker"
	"github.com/camunda-cloud/zeebe/clients/go/pkg/zbc"
	"github.com/jmoiron/sqlx"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	address "gitlab.mashroom.com/mashroom/backend/address/api"
	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/api/process"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bpms"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/error/sqlerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	logger "gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"
	"gitlab.mashroom.com/mashroom/backend/common/generic/tracer"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	geo "gitlab.mashroom.com/mashroom/backend/geo/api"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/intel"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model/postgres"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service"
	listingArchiveInactive "gitlab.mashroom.com/mashroom/backend/listing/internal/task/listing/archive/inactive"
	publicationSyncListed "gitlab.mashroom.com/mashroom/backend/listing/internal/task/publication/sync/listed"
	publicationSyncUnlisted "gitlab.mashroom.com/mashroom/backend/listing/internal/task/publication/sync/unlisted"
	publicationUnviewed "gitlab.mashroom.com/mashroom/backend/listing/internal/task/publication/unviewed"
	"gitlab.mashroom.com/mashroom/backend/listing/version"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	viewing "gitlab.mashroom.com/mashroom/backend/viewing/api"
	zeebe "gitlab.mashroom.com/mashroom/backend/zeebe/api"
)

var conns = connections.Connections{
	Account: account.NewAccountServiceClient(grpc.Dial(
		env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000"),
		"Account",
	)),
	Publication: connections.Publications{
		Zoopla: listing.NewListingPublicationServiceClient(grpc.Dial(
			env.ResolveString("ZOOPLA_SERVICE_URL", "mashroom-zoopla:10000"),
			"Zoopla",
		)),
		Rightmove: listing.NewListingPublicationServiceClient(grpc.Dial(
			env.ResolveString("RIGHTMOVE_SERVICE_URL", "mashroom-rightmove:10000"),
			"Rightmove",
		)),
		Boomin: listing.NewListingPublicationServiceClient(grpc.Dial(
			env.ResolveString("BOOMIN_SERVICE_URL", "mashroom-boomin:10000"),
			"Boomin",
		)),
	},
	Geo: geo.NewGeoServiceClient(grpc.Dial(
		env.ResolveString("GEO_SERVICE_URL", "mashroom-geo:10000"),
		"Geo",
	)),
	Document: document.NewDocumentServiceClient(grpc.Dial(
		env.ResolveString("DOCUMENT_SERVICE_URL", "mashroom-document:10000"),
		"Document",
	)),
	Address: address.NewAddressServiceClient(grpc.Dial(
		env.ResolveString("ADDRESS_SERVICE_URL", "mashroom-address:10000"),
		"Address",
	)),
	Invoice: invoice.NewInvoiceServiceClient(grpc.Dial(
		env.ResolveString("INVOICE_SERVICE_URL", "mashroom-invoice:10000"),
		"Invoice",
	)),
	Extras: extras.NewExtrasServiceClient(grpc.Dial(
		env.ResolveString("EXTRAS_SERVICE_URL", "mashroom-extras:10000"),
		"Extras",
	)),
	Notification: notification.NewNotificationServiceClient(grpc.Dial(
		env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000"),
		"Notification",
	)),
	Closecom: closecom.NewClosecomServiceClient(grpc.Dial(
		env.ResolveString("CLOSECOM_SERVICE_URL", "mashroom-closecom:10000"),
		"Closecom",
	)),
	Viewing: viewing.NewViewingServiceClient(grpc.Dial(
		env.ResolveString("VIEWING_SERVICE_URL", "mashroom-viewing:10000"),
		"Viewing",
	)),
	Offer: offer.NewOfferServiceClient(grpc.Dial(
		env.ResolveString("OFFER_SERVICE_URL", "mashroom-offer:10000"),
		"Offer",
	)),
	Gmailparser: gmailparser.NewGmailparserServiceClient(grpc.Dial(
		env.ResolveString("GMAILPARSER_SERVICE_URL", "mashroom-gmailparser:10000"),
		"Gmailparser",
	)),
	Property: property.NewPropertyServiceClient(grpc.Dial(
		env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000"),
		"Property",
	)),
	Invitation: invitation.NewInvitationServiceClient(grpc.Dial(
		env.ResolveString("INVITATION_SERVICE_URL", "mashroom-invitation:10000"),
		"Invitation",
	)),
	Schedule: schedule.NewScheduleServiceClient(grpc.Dial(
		env.ResolveString("SCHEDULE_SERVICE_URL", "mashroom-schedule:10000"),
		"Schedule",
	)),
	ZeebeProcess: zeebe.NewZeebeProcessClient(grpc.Dial(
		env.ResolveString("BPMS_EXPORTER_URL", "mashroom-zeebe:10000"),
		"ZeebeProcess",
	)),
	SelfRemote: env.ResolveString("SELF_REMOTE", "mashroom-listing:10000"),
}

// envSiteURL is a URL where this service are working
var envSiteURL = env.ResolveString("SITE_URL", "dev.env.mashroom.com")

// envAvmServiceURL is a URL of AVM service
var envIntelServiceURL = env.ResolveString("INTEL_SERVICE_URL", "http://mashroom-intel:5000")

func scheduledTasks(repository model.Repository) *schedule.Registry {
	taskListingArchiveInactive, err := listingArchiveInactive.New(&listingArchiveInactive.Config{
		Repository: repository,
	})
	if err != nil {
		panic(err)
	}

	taskPublicationSyncListed, err := publicationSyncListed.New(&publicationSyncListed.Config{
		Repository:  repository,
		Connections: conns.Publication,
	})
	if err != nil {
		panic(err)
	}

	taskPublicationSyncUnlisted, err := publicationSyncUnlisted.New(&publicationSyncUnlisted.Config{
		Repository:  repository,
		Connections: conns.Publication,
	})
	if err != nil {
		panic(err)
	}

	taskPublicationUnviewed, err := publicationUnviewed.New(&publicationUnviewed.Config{
		Repository:   repository,
		Viewing:      conns.Viewing,
		Offer:        conns.Offer,
		Gmailparser:  conns.Gmailparser,
		Invoice:      conns.Invoice,
		Notification: conns.Notification,
	})
	if err != nil {
		panic(err)
	}

	return &schedule.Registry{
		Periods: map[string]*schedule.Period{
			"hourly": {
				Interval: time.Hour,
				Handlers: []schedule.Handler{
					taskListingArchiveInactive,
					taskPublicationSyncListed,
				},
			},
			"daily": {
				Interval: time.Hour * 24,
				Handlers: []schedule.Handler{
					taskPublicationSyncUnlisted,
					taskPublicationUnviewed,
				},
			},
		},
	}
}

func main() {
	ctx := context.Background()

	defer tracer.Close()

	conn := fmt.Sprintf(
		"postgres://%s:%s@%s:%s/%s?sslmode=disable",
		os.Getenv("POSTGRES_USERNAME"),
		os.Getenv("POSTGRES_PASSWORD"),
		os.Getenv("POSTGRES_HOST"),
		os.Getenv("POSTGRES_PORT"),
		os.Getenv("POSTGRES_DATABASE"),
	)

	db := sqlx.MustOpen("postgres", conn)

	repo, err := postgres.NewRepository(db)
	if err != nil {
		panic(err)
	}

	err = repo.Migrate(ctx, &conns)
	if err != nil {
		panic(err)
	}

	var zbcClient zbc.Client

	if bpms.URL != "" {
		zbcClient, err = zbc.NewClient(&zbc.ClientConfig{
			UsePlaintextConnection: true,
			GatewayAddress:         bpms.URL,
		})
		if err != nil {
			panic(err)
		}
	}

	taskSchedules := scheduledTasks(repo)

	srv, err := service.NewService(&service.Config{
		Repo:         repo,
		Connections:  conns,
		SiteURL:      envSiteURL,
		ZeebeClient:  zbcClient,
		IntelClient:  intel.New(&http.Client{}, envIntelServiceURL),
		TaskRegistry: taskSchedules,
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("listing", grpc.ServerOptionErrorTranslator(sqlerr.Translate))
	if err != nil {
		panic(err)
	}

	listing.RegisterListingServiceServer(grpcServer, srv)
	process.RegisterProcessMonitorServiceServer(grpcServer, srv)
	invoice.RegisterInvoiceNotifyServiceServer(grpcServer, srv)
	schedule.RegisterReminderConsumerServiceServer(grpcServer, srv)
	invitation.RegisterInvitationNotifyServiceServer(grpcServer, srv)

	logger.Infoln("listing started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	go func() {
		var w worker.JobWorker

		w, err = srv.RegisterProcessHandlers()
		if err != nil {
			panic(err)
		}

		err = srv.UploadProcesses()
		if err != nil {
			panic(err)
		}

		go func() {
			<-ctx.Done()

			w.Close()
		}()

		if w != nil {
			w.AwaitClose()
		}
	}()

	go func() {
		err = srv.SetupSchedule(ctx)
		if err != nil {
			panic(err)
		}
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
