// Package intel provides integration with valuation service (intel)
package intel

import (
	"context"
	"fmt"
	"net/http"
	"strings"
)

// Config provides configuration for avm client
type Config struct {
	AvmEndpoint       string
	LettingsEndpoint  string
	LiquidityEndpoint string
	Client            *http.Client
}

// FurnishedValue type as used by avm service
type FurnishedValue string

// FurnishedValue known values
const (
	FurnishedValueFurnished   FurnishedValue = "furnished"
	FurnishedValueUnfurnished FurnishedValue = "unfurnished"
)

// TypeValue property type as used by avm service
type TypeValue int

// TypeValue known values
const (
	TypeValueTerraced     TypeValue = 1
	TypeValueSemiDetached TypeValue = 3
	TypeValueDetached     TypeValue = 4
	TypeValueStudio       TypeValue = 9
	TypeValueOther        TypeValue = 8
)

// ValuationRequestLocation location portion of avm valuation request
type ValuationRequestLocation struct {
	Postcode string `json:"postcode"`
	Building string `json:"building"`
	Flat     string `json:"flat,omitempty"`
	Street   string `json:"street"`
}

// ValuationRequestDetails details portion of avm valuation request
type ValuationRequestDetails struct {
	Type       TypeValue      `json:"type"`
	BedroomsNo int            `json:"bedrooms_no"`
	Furnished  FurnishedValue `json:"furnished"`
	Ownership  int            `json:"ownership"`
}

// ValuationRequestEstimatedValue estimation portion of avm valuation request
type ValuationRequestEstimatedValue struct {
	InitialAskingPriceLow  int `json:"initial_asking_price_low"`
	InitialAskingPriceHigh int `json:"initial_asking_price_high"`
}

// ValuationRequest request for property valuation by avm service
type ValuationRequest struct {
	Location       ValuationRequestLocation       `json:"location"`
	Details        ValuationRequestDetails        `json:"details"`
	EstimatedValue ValuationRequestEstimatedValue `json:"estimated_value"`
}

// OutputAdditional additional information on valuation estimate
type OutputAdditional struct {
	VariationConfidence        float64 `json:"variation_confidence"`
	ComparatorNumberConfidence float64 `json:"comparator_number_confidence"`
}

// AvmOutput estimation portion of response of valuation response
type AvmOutput struct {
	Estimate   float64          `json:"estimate"`
	Confidence float64          `json:"confidence"`
	Additional OutputAdditional `json:"additional"`
	LowerBound float64          `json:"lower_bound"`
	UpperBound float64          `json:"upper_bound"`
}

// ValuationResponse from avm
type ValuationResponse struct {
	AvmOutput AvmOutput `json:"avm_output"`
	Raw       string    `json:"raw"`
}

// Client for avm service
type Client interface {
	Avm(ctx context.Context, request *ValuationRequest) (*ValuationResponse, error)
	Lettings(ctx context.Context, request *ValuationRequest) (*ValuationResponse, error)
	Liquidity(ctx context.Context, request *ValuationRequest) (*ValuationResponse, error)
}

// TokenTransport auth token transport
type TokenTransport struct {
	Token string
}

// RoundTrip sets authentication token to http transport
func (tr *TokenTransport) RoundTrip(req *http.Request) (*http.Response, error) {
	req.Header.Set("Authorization", "Token "+tr.Token)

	return http.DefaultTransport.RoundTrip(req)
}

// New returns new avm client
func New(httpClient *http.Client, url string) Client {
	return &clientImpl{
		config: Config{
			AvmEndpoint:       joinURL(url, "avm"),
			LettingsEndpoint:  joinURL(url, "lettings"),
			LiquidityEndpoint: joinURL(url, "liquidity"),
			Client:            httpClient,
		},
	}
}

func joinURL(base, endpoint string) string {
	return fmt.Sprintf("%s/%s/", strings.TrimRight(base, "/"), strings.Trim(endpoint, "/"))
}
