// Package model provides structures for persistent repository
package model

import (
	"context"
	"database/sql/driver"
	"encoding/json"
	"strconv"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
)

// VersionSpec returns set of filtering options based on version specification
func VersionSpec(ctx context.Context, spec *listing.VersionSpec, published bool) *VersionSpecOptions {
	var id uint64

	if spec.GetVersionId() != "" {
		id, _ = strconv.ParseUint(spec.GetVersionId(), 10, 64)
	}

	statuses := spec.GetSelect().GetStatuses()

	if len(statuses) == 0 && spec.GetUserId() == "" &&
		!grpcctx.ContextIsAdminUser(ctx) && !grpcctx.ContextIsServiceUser(ctx) {
		statuses = []listing.VersionStatus{
			listing.VersionStatus_VERSION_VERIFIED,
		}
	}

	return &VersionSpecOptions{
		ListingID:      spec.GetSelect().GetListingId(),
		PropertyID:     spec.GetSelect().GetPropertyId(),
		TenancyID:      spec.GetSelect().GetTenancyId(),
		InvitationID:   spec.GetSelect().GetInvitationId(),
		UserID:         spec.GetUserId(),
		Statuses:       statuses,
		VersionID:      id,
		Published:      published,
		ServiceContext: grpcctx.ContextIsAdminUser(ctx) || grpcctx.ContextIsServiceUser(ctx),
	}
}

// VersionSpecOptions is options for get query
type VersionSpecOptions struct {
	ListingID      string
	PropertyID     string
	TenancyID      string
	InvitationID   string
	UserID         string
	Statuses       []listing.VersionStatus
	VersionID      uint64
	Published      bool
	ServiceContext bool
}

// SearchInRadius  is using for searching by radius
type SearchInRadius struct {
	Longitude      float64 `json:"longitude"`
	Latitude       float64 `json:"latitude"`
	RadiusInMeters uint64  `json:"radius_in_meters"`
}

// FlagType is using for updating flag
type FlagType string

var (
	// FlagDisabled is disable sign of version
	FlagDisabled FlagType = "{disabled}"
	// FlagTest is test sign of version
	FlagTest FlagType = "{test}"
	// FlagAuto is auto sign of version
	FlagAuto FlagType = "{auto}"
	// FlagActiveOffer shows that the version currently has an active offer
	FlagActiveOffer FlagType = "{active_offer}"
)

// Value returns serialized representation
func (v SearchInRadius) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *SearchInRadius) Scan(value interface{}) error {
	if value == nil {
		return nil
	}

	bs, ok := value.([]byte)
	if !ok {
		return ErrByteArrayAssert
	}

	return json.Unmarshal(bs, v)
}

// PropertyListingID pair
type PropertyListingID struct {
	PropertyID string `db:"property_id"`
	ListingID  string `db:"listing_id"`
}

// Repository interface abstracting specific repository
type Repository interface {
	Migrate(ctx context.Context, conns *connections.Connections) error
	VersionCreate(ctx context.Context, lv *Version, newstatus listing.VersionStatus) error
	VersionServiceCreate(ctx context.Context, lv *Version, newstatus listing.VersionStatus) error
	VersionUpdate(ctx context.Context, lv *Version, newstatus listing.VersionStatus) error
	VersionUpdateStatus(ctx context.Context, lvID uint64, oldstatus, newstatus listing.VersionStatus) error
	VersionGet(ctx context.Context, options *VersionSpecOptions) (*Version, error)
	VersionsList(ctx context.Context, options *VersionSpecOptions) ([]*Version, error)
	VersionFilterList(
		ctx context.Context,
		filter *query.QueryParams,
		rad *SearchInRadius,
	) ([]*Version, uint64, error)
	VersionPublicationEntrySave(ctx context.Context, m *PublicationEntry) error
	VersionPublicationEntryPublished(
		ctx context.Context,
		platform listing.PublicationPlatform,
		propertyID string,
	) (bool, error)
	VersionPublicationEntryListUnsynced(
		ctx context.Context,
		platform listing.PublicationPlatform,
	) ([]*PublicationEntry, error)
	VersionNearbyGet(ctx context.Context, listing uint64) (*NearbyCache, error)
	VersionNearbyInsert(ctx context.Context, nearby *NearbyCache) error
	VersionSetAdmin(ctx context.Context, options *VersionSpecOptions, admin string) error
	VersionAverageRentPrices(ctx context.Context, opts *VersionAverageRentOptions) ([]VersionAverageRentItem, error)
	VersionArchiveInactive(ctx context.Context, now time.Time) error
	VersionPublicationPropertyListingIDs(
		ctx context.Context,
		platform listing.PublicationPlatform,
		since time.Time,
	) ([]PropertyListingID, error)
	VersionPublicationLetPropertyIDs(
		ctx context.Context,
	) (ids []string, err error)
	VersionPropertiesCount(ctx context.Context, userIDs ...string) (map[string]int64, error)
	VersionFlagUpdate(ctx context.Context, options *VersionSpecOptions, flag FlagType, value interface{}) error
	VersionNotificationUpdate(
		ctx context.Context,
		options *VersionSpecOptions,
		notification VersionNotificationType,
		value bool,
	) error
	PrefaceGet(ctx context.Context, clientID string) (*Preface, error)
	PrefaceSave(ctx context.Context, preface *Preface) error
	PrefaceRemove(ctx context.Context, clientID string) error
}
