// Package postgres provides repository implementation for postgresql database
package postgres

import (
	"context"
	"fmt"

	"gitlab.mashroom.com/mashroom/backend/listing/internal/model/postgres/migrations/data"

	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model/postgres/migrations/schema"

	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"

	"github.com/golang-migrate/migrate/v4"
	"github.com/golang-migrate/migrate/v4/database/postgres"
	"github.com/jmoiron/sqlx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
)

type repositoryImpl struct {
	DB *sqlx.DB
	repositoryVersion
	repositoryVersionNearby
	repositoryVersionPublication
	repositoryPreface
}

func prepareNamed(db *sqlx.DB, s string) (r *sqlx.NamedStmt, err error) {
	r, err = db.PrepareNamed(s)

	if err != nil {
		fmt.Printf("error %v preparing %s\n", err, s)
	}

	return
}

func preparex(db *sqlx.DB, s string) (r *sqlx.Stmt, err error) {
	r, err = db.Preparex(s)

	if err != nil {
		fmt.Printf("error %v preparing %s\n", err, s)
	}

	return
}

// NewRepository returns new postgresql repository
func NewRepository(db *sqlx.DB) (repos model.Repository, err error) {
	repo := &repositoryImpl{
		DB: db,
	}

	return repo, nil
}

func (repo *repositoryImpl) prepare() (err error) {
	repo.repositoryVersion, err = prepareVersion(repo.DB)
	if err != nil {
		return err
	}

	repo.repositoryVersionNearby, err = prepareVersionNearby(repo.DB)
	if err != nil {
		return err
	}

	repo.repositoryVersionPublication, err = prepareVersionPublication(repo.DB)
	if err != nil {
		return err
	}

	repo.repositoryPreface, err = preparePreface(repo.DB)
	if err != nil {
		return err
	}

	return
}

func (repo *repositoryImpl) Migrate(ctx context.Context, conns *connections.Connections) error {
	source, err := schema.NewSource()
	if err != nil {
		return err
	}

	driver, err := postgres.WithInstance(repo.DB.DB, &postgres.Config{})
	if err != nil {
		return err
	}

	m, err := migrate.NewWithInstance("httpfs", source, "postgres", driver)
	if err != nil {
		return err
	}

	err = m.Up()

	switch {
	case err == nil:
		log.Debug("Migrated.")
	case err == migrate.ErrNoChange:
		log.Debug("Migration not required.")
	default:
		return err
	}

	err = repo.prepare()
	if err != nil {
		return err
	}

	go repo.migrateData(ctx, conns)

	return nil
}

func (repo *repositoryImpl) migrateData(ctx context.Context, conns *connections.Connections) {
	for _, f := range data.CurrentMigrationList {
		err := f(ctx, repo.DB, conns)
		if err != nil {
			log.WithError(err).Error("migrating data")
		}
	}
}
