package postgres

import (
	"context"
	"database/sql"
	"errors"
	"time"

	"github.com/jmoiron/sqlx"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
)

type repositoryPreface struct {
	db     *sqlx.DB
	save   *sqlx.NamedStmt
	get    *sqlx.NamedStmt
	remove *sqlx.NamedStmt
}

func preparePreface(db *sqlx.DB) (stmt repositoryPreface, err error) {
	stmt.db = db

	stmt.save, err = prepareNamed(db, queryPrefaceSave)
	if err != nil {
		return
	}

	stmt.get, err = prepareNamed(db, queryPrefaceGet)
	if err != nil {
		return
	}

	stmt.remove, err = prepareNamed(db, queryPrefaceRemove)
	if err != nil {
		return
	}

	return
}

const queryPrefaceSave = `
insert into preface_versions(
  client_id,
  description,
  details,
  rental,
  location,
  valuation,
  gps,
  validation,
  migrated_id,
  timezone,
  created,
  updated
) values (
  :client_id,
  :description,
  :details,
  :rental,
  :location,
  :valuation,
  ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  :validation,
  :migrated_id,
  coalesce(nullif(:timezone, ''), 'Europe/London'),
  now(),
  now()
) on conflict(client_id) do update set
  description = preface_versions.description || :description,
  details = preface_versions.details || :details,
  rental = preface_versions.rental || :rental,
  location = preface_versions.location || :location,
  valuation = preface_versions.valuation || :valuation,
  gps = ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  validation = preface_versions.validation || :validation,
  updated = now()
returning
  id,
  client_id,
  description,
  details,
  rental,
  location || jsonb_build_object(
    'gps', jsonb_build_object('longitude', st_x(gps::::geometry), 'latitude', st_y(gps::::geometry))
  ) as location,
  valuation,
  validation,
  timezone,
  created,
  updated,
  migrated_id
`

func (repo *repositoryPreface) PrefaceSave(
	ctx context.Context,
	lv *model.Preface,
) (err error) {
	return repo.save.GetContext(ctx, lv, lv)
}

const queryPrefaceGet = `
select
  id,
  client_id,
  description,
  details,
  rental,
  location || jsonb_build_object(
    'gps', jsonb_build_object('longitude', st_x(gps::::geometry), 'latitude', st_y(gps::::geometry))
  ) as location,
  valuation,
  validation,
  timezone,
  created,
  updated,
  migrated_id
from preface_versions where
  client_id = :client_id
`

func (repo *repositoryPreface) PrefaceGet(
	ctx context.Context,
	clientID string,
) (preface *model.Preface, err error) {
	preface = &model.Preface{}

	err = repo.get.GetContext(ctx, preface, struct {
		ClientID string `db:"client_id"`
	}{
		ClientID: clientID,
	})
	if err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return &model.Preface{
				Created:  time.Now(),
				Updated:  time.Now(),
				ClientID: clientID,
			}, nil
		}

		return nil, err
	}

	return preface, nil
}

const queryPrefaceRemove = `
delete from preface_versions where
  client_id = :client_id
`

func (repo *repositoryPreface) PrefaceRemove(
	ctx context.Context,
	clientID string,
) (err error) {
	_, err = repo.remove.ExecContext(ctx, struct {
		ClientID string `db:"client_id"`
	}{
		ClientID: clientID,
	})

	return err
}
