package postgres

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"reflect"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/jmoiron/sqlx"
	"github.com/lib/pq"
	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/generic/psqlquery"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
)

type repositoryVersion struct {
	db                 *sqlx.DB
	create             *sqlx.NamedStmt
	serviceCreate      *sqlx.NamedStmt
	update             *sqlx.NamedStmt
	updateStatus       *sqlx.NamedStmt
	get                *sqlx.NamedStmt
	list               *sqlx.NamedStmt
	setadmin           *sqlx.NamedStmt
	averageRentPrices  *sqlx.NamedStmt
	archiveInactive    *sqlx.NamedStmt
	propertiesCount    *sqlx.NamedStmt
	flagUpdate         *sqlx.NamedStmt
	notificationUpdate *sqlx.NamedStmt
}

func prepareVersion(db *sqlx.DB) (stmt repositoryVersion, err error) {
	stmt.db = db

	stmt.create, err = prepareNamed(db, queryVersionCreate)
	if err != nil {
		return
	}

	stmt.serviceCreate, err = prepareNamed(db, queryVersionServiceCreate)
	if err != nil {
		return
	}

	stmt.update, err = prepareNamed(db, queryVersionUpdate)
	if err != nil {
		return
	}

	stmt.updateStatus, err = prepareNamed(db, queryVersionUpdateStatus)
	if err != nil {
		return
	}

	stmt.get, err = prepareNamed(db, queryVersionGet)
	if err != nil {
		return
	}

	stmt.list, err = prepareNamed(db, queryVersionList)
	if err != nil {
		return
	}

	stmt.setadmin, err = prepareNamed(db, queryVersionSetAdmin)
	if err != nil {
		return
	}

	stmt.averageRentPrices, err = prepareNamed(db, queryVersionAverageRentPrices)
	if err != nil {
		return
	}

	stmt.archiveInactive, err = prepareNamed(db, queryVersionArchiveInactive)
	if err != nil {
		return
	}

	stmt.propertiesCount, err = prepareNamed(db, queryVersionPropertiesCount)
	if err != nil {
		return
	}

	stmt.flagUpdate, err = prepareNamed(db, queryVersionFlagUpdate)
	if err != nil {
		return
	}

	stmt.notificationUpdate, err = prepareNamed(db, queryVersionNotificationUpdate)
	if err != nil {
		return
	}

	return
}

const queryVersionCreate = `
with
vid as (
  select coalesce(
    nullif(:id, 0),
    nextval('versions_id_seq') 
  ) as vid
),
lid as (
  select coalesce(
    nullif(:listing_id, ''),
   (select listing_id from versions where migrated_id = :migrated_id and migrated_id != '' limit 1),
    nextval('listing_id_seq')::::text
  ) as lid
),
sub as (
  update versions set
    status = :alt_status
  where
    :alt_status != 0 and 
    status = :new_status and
    listing_id = (select lid from lid) and
    id != (select vid from vid)
)
insert into versions(
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  gps,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  migrated_id,
  timezone,
  created,
  updated
) values (
  (select vid from vid),
  (select lid from lid),
  :property_id,
  :tenancy_id,
  :ivr,
  coalesce(nullif(:new_status, 0), 1),
  :description,
  :details,
  :people,
  :rental,
  :location,
  :valuation,
  ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  :verification,
  :validation,
  :origin,
  :viewing_availability,
  :invitation,
  :flags,
  :notification,
  :migrated_id,
  coalesce(nullif(:timezone, ''), 'Europe/London'),
  now(),
  now()
) on conflict(id) do update set
  listing_id = (select lid from lid),
  tenancy_id = :tenancy_id,
  status = coalesce(nullif(:new_status, 0), 1),
  description = :description,
  details = :details,
  people = :people,
  rental = :rental,
  location = :location,
  valuation = :valuation,
  gps = ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  verification = :verification,
  validation = :validation,
  viewing_availability = :viewing_availability,
  invitation = :invitation,
  flags = :flags,
  notification = :notification,
  updated = now()
where
  versions.status = :status
returning
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  timezone,
  created,
  updated,
  migrated_id,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 1 order by id asc
  ) foo) as publications_mashroom,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 2 order by id asc
  ) foo) as publications_rightmove,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 3 order by id asc
  ) foo) as publications_zoopla,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 4 order by id asc
  ) foo) as publications_boomin,
  coalesce((select published_at from version_publications where
      version_publications.listing_id = versions.listing_id and
      version_publications.unpublished_at <= '1970-01-01'::::timestamp
      order by published_at desc limit 1), '0001-01-01'::::timestamp) as published,
  coalesce(
    (select max(v.id) from versions v where v.listing_id = versions.listing_id and v.status in (3,4)),
    0
  ) as last_verified
`

func (repo *repositoryVersion) VersionCreate(
	ctx context.Context,
	lv *model.Version,
	newstatus listing.VersionStatus,
) (err error) {
	err = repo.create.GetContext(ctx, lv, struct {
		*model.Version
		NewStatus listing.VersionStatus `db:"new_status"`
		AltStatus listing.VersionStatus `db:"alt_status"`
	}{
		Version:   lv,
		NewStatus: newstatus,
		AltStatus: versionAltStatus(lv.Status),
	})

	return
}

const queryVersionServiceCreate = `
with 
vid as (
  select coalesce(
    nullif(:id, 0),
    nextval('versions_id_seq') 
  ) as vid
),
lid as (
  select coalesce(
    nullif(:listing_id, ''),
   (select listing_id from versions where migrated_id = :migrated_id and migrated_id != '' limit 1),
    nextval('listing_id_seq')::::text
  ) as lid
),
sub as (
  update versions set
    status = :alt_status
  where
    :alt_status != 0 and 
    status = :new_status and 
    listing_id = (select lid from lid) and
    id != (select vid from vid)
)
insert into versions(
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  gps,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  migrated_id,
  timezone,
  created,
  updated
) values (
  (select vid from vid),
  (select lid from lid),
  :property_id,
  :tenancy_id,
  :ivr,
  coalesce(nullif(:new_status, 0), 1),
  :description,
  :details,
  :people,
  :rental,
  :location,
  :valuation,
  ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  :verification,
  :validation,
  :origin,
  :viewing_availability,
  :invitation,
  :flags,
  :notification,
  :migrated_id,
  coalesce(nullif(:timezone, ''), 'Europe/London'),
  :created,
  :updated
) on conflict(id) do update set
  listing_id = (select lid from lid),
  tenancy_id = :tenancy_id,
  status = coalesce(nullif(:new_status, 0), 1),
  description = :description,
  details = :details,
  people = :people,
  rental = :rental,
  location = :location,
  valuation = :valuation,
  gps = ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  verification = :verification,
  validation = :validation,
  viewing_availability = :viewing_availability,
  invitation = :invitation,
  flags = :flags,
  notification = :notification,
  updated = :updated
where
  versions.status = :status
returning
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  timezone,
  created,
  updated,
  migrated_id,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 1 order by id asc
  ) foo) as publications_mashroom,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 2 order by id asc
  ) foo) as publications_rightmove,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 3 order by id asc
  ) foo) as publications_zoopla,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 4 order by id asc
  ) foo) as publications_boomin,
  coalesce((select published_at from version_publications where
      version_publications.listing_id = versions.listing_id and
      version_publications.unpublished_at <= '1970-01-01'::::timestamp
      order by published_at desc limit 1), '0001-01-01'::::timestamp) as published,
  coalesce(
    (select max(v.id) from versions v where v.listing_id = versions.listing_id and v.status in (3,4)),
    0
  ) as last_verified
`

func (repo *repositoryVersion) VersionServiceCreate(
	ctx context.Context,
	lv *model.Version,
	newstatus listing.VersionStatus,
) (err error) {
	return repo.serviceCreate.GetContext(ctx, lv, struct {
		*model.Version
		NewStatus listing.VersionStatus `db:"new_status"`
		AltStatus listing.VersionStatus `db:"alt_status"`
	}{
		Version:   lv,
		NewStatus: newstatus,
		AltStatus: versionAltStatus(newstatus),
	})
}

const queryVersionUpdate = `
with sub as (
  update versions set
    status = :alt_status
  where
    :alt_status != 0 and 
    status = :new_status and 
    listing_id = :listing_id and
    id != :id
)
update versions set
  tenancy_id = :tenancy_id,
  status = :new_status,
  description = :description,
  details = :details,
  people = :people,
  rental = :rental,
  location = :location,
  valuation = :valuation,
  gps = ST_SetSRID(
    ST_MakePoint(
      ((:location)::::jsonb -> 'gps' -> 'longitude')::::float, 
      ((:location)::::jsonb -> 'gps' -> 'latitude')::::float
   ), 
   4326
  ),
  verification = :verification,
  validation = :validation,
  viewing_availability = :viewing_availability,
  invitation = :invitation,
  flags = :flags,
  notification = :notification,
  updated = now()
where
  id = :id and 
  status = :status
returning
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  timezone,
  created,
  updated,
  migrated_id,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 1 order by id asc
  ) foo) as publications_mashroom,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 2 order by id asc
  ) foo) as publications_rightmove,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 3 order by id asc
  ) foo) as publications_zoopla,
  (select jsonb_agg(foo.*) from (select * from version_publications where
    version_publications.listing_id = versions.listing_id and platform = 4 order by id asc
  ) foo) as publications_boomin,
  coalesce((select published_at from version_publications where
      version_publications.listing_id = versions.listing_id and
      version_publications.unpublished_at <= '1970-01-01'::::timestamp
      order by published_at desc limit 1), '0001-01-01'::::timestamp) as published,
  coalesce(
    (select max(v.id) from versions v where v.listing_id = versions.listing_id and v.status in (3,4)),
    0
  ) as last_verified
`

func (repo *repositoryVersion) VersionUpdate(
	ctx context.Context,
	lv *model.Version,
	newstatus listing.VersionStatus,
) (err error) {
	return repo.update.GetContext(ctx, lv, struct {
		*model.Version
		NewStatus listing.VersionStatus `db:"new_status"`
		AltStatus listing.VersionStatus `db:"alt_status"`
	}{
		Version:   lv,
		NewStatus: newstatus,
		AltStatus: versionAltStatus(newstatus),
	})
}

func versionAltStatus(newstatus listing.VersionStatus) listing.VersionStatus {
	if newstatus == listing.VersionStatus_VERSION_VERIFIED {
		return listing.VersionStatus_VERSION_ARCHIVE
	}

	return 0
}

const queryVersionUpdateStatus = `
with sub as (
  update versions set
    status = :alt_status
  where
    :alt_status != 0 and 
    status = :new_status and
    listing_id = (select listing_id from versions where id = :id) and
    id != :id 
)
update versions set
  status = :new_status
where
  id = :id and
  status = :old_status
`

func (repo *repositoryVersion) VersionUpdateStatus(
	ctx context.Context,
	id uint64,
	oldstatus, newstatus listing.VersionStatus,
) (err error) {
	_, err = repo.updateStatus.ExecContext(ctx, struct {
		ID        uint64                `db:"id"`
		OldStatus listing.VersionStatus `db:"old_status"`
		NewStatus listing.VersionStatus `db:"new_status"`
		AltStatus listing.VersionStatus `db:"alt_status"`
	}{
		ID:        id,
		OldStatus: oldstatus,
		NewStatus: newstatus,
		AltStatus: versionAltStatus(newstatus),
	})

	return
}

func arrInt(v interface{}) (arr []int64) {
	val := reflect.ValueOf(v)

	if val.Kind() != reflect.Slice {
		return nil
	}

	for i := 0; i < val.Len(); i++ {
		idx := val.Index(i)
		switch idx.Kind() {
		case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64,
			reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
		default:
			return nil
		}

		arr = append(arr, idx.Int())
	}

	return
}

const queryVersionGet = `
select
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  timezone,
  created,
  updated,
  migrated_id,
  publications_mashroom,
  publications_rightmove,
  publications_zoopla,
  publications_boomin,
  published,
  last_verified
from versions_list where
  (:id = 0 or id = :id) and
  (:listing_id = '' or listing_id = :listing_id) and
  (:property_id = '' or property_id = :property_id) and
  (:tenancy_id = '' or tenancy_id = :tenancy_id) and
  (:user_id = '' or 
    (people->>'landlord' = :user_id) or
    (people->>'submitter' = :user_id) or
    (people->>'manager' = :user_id) or
    (people->'tenants' @> ('"' || :user_id || '"')::::jsonb) or
    (people->'viewers'->>:user_id is not null) or
    (people->'offerers'->>:user_id is not null) or
    (invitation->'manager_invitation'->>'invitee_id' = :user_id) or
    (invitation->'landlord_invitation'->>'invitee_id' = :user_id) or
    (status = 3) or
    :service_context
  ) and
  (:invitation_id = '' or
    (invitation->'manager_invitation'->>'invitation_id' = :invitation_id) or
    (invitation->'landlord_invitation'->>'invitation_id' = :invitation_id)
  ) and
  (array_length(cast(:statuses as int8[]), 1) is null 
    or status = any(cast(:statuses as int8[]))) and
  (:published = false or coalesce(versions_list.published, '0001-01-01'::::timestamp) > '1970-01-01'::::timestamp) and
  (not coalesce((flags->'disabled')::::bool, false) or :service_context)
order by id desc
limit 1 
`

func (repo *repositoryVersion) VersionGet(
	ctx context.Context,
	options *model.VersionSpecOptions,
) (m *model.Version, err error) {
	if options.VersionID == 0 && options.ListingID == "" && options.TenancyID == "" && options.PropertyID == "" {
		return nil, sql.ErrNoRows
	}

	m = &model.Version{}

	err = repo.get.GetContext(ctx, m, struct {
		ID             uint64        `db:"id"`
		ListingID      string        `db:"listing_id"`
		TenancyID      string        `db:"tenancy_id"`
		PropertyID     string        `db:"property_id"`
		UserID         string        `db:"user_id"`
		InvitationID   string        `db:"invitation_id"`
		Statuses       pq.Int64Array `db:"statuses"`
		Published      bool          `db:"published"`
		ServiceContext bool          `db:"service_context"`
	}{
		ID:             options.VersionID,
		ListingID:      options.ListingID,
		TenancyID:      options.TenancyID,
		PropertyID:     options.PropertyID,
		InvitationID:   options.InvitationID,
		UserID:         options.UserID,
		Statuses:       arrInt(options.Statuses),
		ServiceContext: options.ServiceContext,
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryVersionList = `
select
  id,
  listing_id,
  property_id,
  tenancy_id,
  ivr,
  status,
  description,
  details,
  people,
  rental,
  location,
  valuation,
  verification,
  validation,
  origin,
  viewing_availability,
  invitation,
  flags,
  notification,
  timezone,
  created,
  updated,
  migrated_id,
  publications_mashroom,
  publications_rightmove,
  publications_zoopla,
  publications_boomin,
  published,
  last_verified
from versions_list where
  (:id = 0 or id = :id) and
  (:listing_id = '' or listing_id = :listing_id) and
  (:property_id = '' or property_id = :property_id) and 
  (:tenancy_id = '' or tenancy_id = :tenancy_id) and
  (:user_id = '' or 
    (people->>'landlord' = :user_id) or
    (people->>'submitter' = :user_id) or
    (people->>'manager' = :user_id) or
    (people->'tenants' @> ('"' || :user_id || '"')::::jsonb) or
    (people->'viewers'->>:user_id is not null) or
    (people->'offerers'->>:user_id is not null) or
    (invitation->'manager_invitation'->>'invitee_id' = :user_id) or
    (invitation->'landlord_invitation'->>'invitee_id' = :user_id) or
    (status = 3)
  ) and
  (:invitation_id = '' or
    (invitation->'manager_invitation'->>'invitation_id' = :invitation_id) or
    (invitation->'landlord_invitation'->>'invitation_id' = :invitation_id)
  ) and
  (array_length(cast(:statuses as int8[]), 1) is null or
    status = any(cast(:statuses as int8[]))) and
  (:published = false or coalesce(versions_list.published, '0001-01-01'::::timestamp) > '1970-01-01'::::timestamp) and
  (not coalesce((flags->'disabled')::::bool, false) or :service_context)
`

func (repo *repositoryVersion) VersionsList(
	ctx context.Context,
	options *model.VersionSpecOptions,
) (ms []*model.Version, err error) {
	err = repo.list.SelectContext(ctx, &ms, struct {
		ID             uint64        `db:"id"`
		ListingID      string        `db:"listing_id"`
		TenancyID      string        `db:"tenancy_id"`
		PropertyID     string        `db:"property_id"`
		InvitationID   string        `db:"invitation_id"`
		UserID         string        `db:"user_id"`
		Statuses       pq.Int64Array `db:"statuses"`
		Published      bool          `db:"published"`
		ServiceContext bool          `db:"service_context"`
	}{
		ID:             options.VersionID,
		ListingID:      options.ListingID,
		TenancyID:      options.TenancyID,
		PropertyID:     options.PropertyID,
		InvitationID:   options.InvitationID,
		UserID:         options.UserID,
		Statuses:       arrInt(options.Statuses),
		Published:      options.Published,
		ServiceContext: options.ServiceContext,
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryVersionSetAdmin = `
update versions 
  set people = jsonb_set(people, '{admin}', to_jsonb(cast(:admin as text)))
where
  (:id = 0 or id = :id) and
  (:listing_id = '' or listing_id = :listing_id) and
  (:property_id = '' or property_id = :property_id) and 
  (:tenancy_id = '' or tenancy_id = :tenancy_id) and
  (:user_id = '' or 
    (people->>'landlord' = :user_id) or
    (people->>'submitter' = :user_id) or
    (people->>'manager' = :user_id) or
    (people->'tenants' @> ('"' || :user_id || '"')::::jsonb) or
    (people->'viewers'->>:user_id is not null) or
    (people->'offerers'->>:user_id is not null) or
    (invitation->'manager_invitation'->>'invitee_id' = :user_id) or
    (invitation->'landlord_invitation'->>'invitee_id' = :user_id) or
    (status = 3)
  ) and
  (:invitation_id = '' or
    (invitation->'manager_invitation'->>'invitation_id' = :invitation_id) or
    (invitation->'landlord_invitation'->>'invitation_id' = :invitation_id)
  ) and
  (array_length(cast(:statuses as int8[]), 1) is null or
    status = any(cast(:statuses as int8[])))
`

func (repo *repositoryVersion) VersionSetAdmin(
	ctx context.Context,
	options *model.VersionSpecOptions,
	admin string,
) (err error) {
	_, err = repo.setadmin.ExecContext(ctx, struct {
		ID           uint64        `db:"id"`
		ListingID    string        `db:"listing_id"`
		TenancyID    string        `db:"tenancy_id"`
		PropertyID   string        `db:"property_id"`
		InvitationID string        `db:"invitation_id"`
		UserID       string        `db:"user_id"`
		Statuses     pq.Int64Array `db:"statuses"`
		Admin        string        `db:"admin"`
	}{
		ID:           options.VersionID,
		ListingID:    options.ListingID,
		TenancyID:    options.TenancyID,
		PropertyID:   options.PropertyID,
		InvitationID: options.InvitationID,
		UserID:       options.UserID,
		Statuses:     arrInt(options.Statuses),
		Admin:        admin,
	})

	return
}

func (repo *repositoryVersion) VersionFilterList(
	ctx context.Context,
	filter *query.QueryParams,
	radius *model.SearchInRadius,
) (ms []*model.Version, count uint64, err error) {
	b, err := psqlquery.ListQueryBuilder(model.Version{}, "versions_list", filter)
	if err != nil {
		return nil, 0, err
	}

	if radius != nil {
		b = b.Where(squirrel.Expr(
			"ST_DWithin(gps, ST_SetSRID(ST_MakePoint(?, ?), 4326), ?)",
			radius.Longitude,
			radius.Latitude,
			radius.RadiusInMeters/2,
		))
	}

	qq, args, err := b.PlaceholderFormat(squirrel.Dollar).ToSql()
	if err != nil {
		return nil, 0, err
	}

	err = repo.db.SelectContext(ctx, &ms, qq, args...)
	if err != nil {
		return nil, 0, err
	}

	b, err = psqlquery.CountQueryBuilder(model.Version{}, "versions_list", filter)
	if err != nil {
		return nil, 0, err
	}

	if radius != nil {
		b = b.Where(squirrel.Expr(
			"ST_DWithin(gps, ST_SetSRID(ST_MakePoint(?, ?), 4326), ?)",
			radius.Longitude,
			radius.Latitude,
			radius.RadiusInMeters/2,
		))
	}

	qq, args, err = b.PlaceholderFormat(squirrel.Dollar).ToSql()
	if err != nil {
		return nil, 0, err
	}

	err = repo.db.GetContext(ctx, &count, qq, args...)
	if err != nil {
		return nil, 0, err
	}

	return
}

const queryVersionAverageRentPrices = `
with cte as (
  select
    coalesce((details->'bedrooms_no')::::int, 0) as bedrooms,
    coalesce((rental->'rental_price')::::numeric, 0) as price
  from versions where
    status = any(cast(:statuses as int8[])) and
    (:country = '' or location->>'country' = :country) and
    (:region = '' or location->>'geo_region' = :region) and
    (:county = '' or location->>'geo_county' = :county) and
    (:town = '' or location->>'geo_town' = :town) and
    (rental->'rental_price')::::numeric > 0.0 and
    (details->'type')::::int != 0
), 
adjusted as (
  select
    (case when bedrooms < 1 then 1 when bedrooms > 4 then 4 else bedrooms end),
    price
  from cte
) 
select
  bedrooms,
  avg(price) as price,
  count(*) as total
from adjusted
group by bedrooms;
`

func (repo *repositoryVersion) VersionAverageRentPrices(
	ctx context.Context,
	options *model.VersionAverageRentOptions,
) (ms []model.VersionAverageRentItem, err error) {
	err = repo.averageRentPrices.SelectContext(ctx, &ms, struct {
		Country  string        `db:"country"`
		Region   string        `db:"region"`
		County   string        `db:"county"`
		Town     string        `db:"town"`
		Statuses pq.Int64Array `db:"statuses"`
	}{
		Country:  options.Country,
		Region:   options.Region,
		County:   options.County,
		Town:     options.Town,
		Statuses: arrInt(options.Statuses),
	})
	if err != nil {
		return nil, err
	}

	return
}

const queryVersionArchiveInactive = `
update versions set status = 5 where status = 1 and listing_id in (
  select listing_id from (
    select array_agg(status) as statuses, max(updated) as updated, listing_id from versions group by listing_id
  ) as f
  where not(statuses && '{2,3,4}') and statuses && '{1}' and (:now - updated) > '30 days'::::interval
)
`

func (repo *repositoryVersion) VersionArchiveInactive(ctx context.Context, now time.Time) error {
	_, err := repo.archiveInactive.ExecContext(ctx, struct {
		Now time.Time `db:"now"`
	}{
		Now: now,
	})

	return err
}

const queryVersionPropertiesCount = `
select
  count(distinct property_id) as count,
  people->>'landlord' as user_id
from versions where 
  people->>'landlord' = any(cast(:user_ids as text[]))
group by
  people->>'landlord'
`

func (repo *repositoryVersion) VersionPropertiesCount(
	ctx context.Context,
	userIDs ...string,
) (count map[string]int64, err error) {
	var res []struct {
		Count  int64  `db:"count"`
		UserID string `db:"user_id"`
	}

	err = repo.propertiesCount.SelectContext(ctx, &res, struct {
		UserID pq.StringArray `db:"user_ids"`
	}{
		UserID: userIDs,
	})
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, err
	}

	count = make(map[string]int64)

	for _, r := range res {
		count[r.UserID] = r.Count
	}

	return
}

const queryVersionFlagUpdate = `
update versions set flags = jsonb_set(flags, :flag, :value) where 
  (:id = 0 or id = :id) and
  (:property_id = '' or property_id = :property_id)
`

func (repo *repositoryVersion) VersionFlagUpdate(
	ctx context.Context,
	options *model.VersionSpecOptions,
	flag model.FlagType,
	value interface{},
) (err error) {
	_, err = repo.flagUpdate.ExecContext(ctx, struct {
		ID         uint64      `db:"id"`
		PropertyID string      `db:"property_id"`
		Flag       string      `db:"flag"`
		Value      interface{} `db:"value"`
	}{
		ID:         options.VersionID,
		PropertyID: options.PropertyID,
		Flag:       string(flag),
		Value:      value,
	})

	return err
}

const queryVersionNotificationUpdate = `
update versions set notification = jsonb_set(notification, :notification, :value) where
  (:property_id = '' or property_id = :property_id)
`

func (repo *repositoryVersion) VersionNotificationUpdate(
	ctx context.Context,
	options *model.VersionSpecOptions,
	notification model.VersionNotificationType,
	value bool,
) (err error) {
	_, err = repo.notificationUpdate.ExecContext(ctx, struct {
		PropertyID   string      `db:"property_id"`
		Notification string      `db:"notification"`
		Value        interface{} `db:"value"`
	}{
		PropertyID:   options.PropertyID,
		Notification: fmt.Sprintf("{%s}", string(notification)),
		Value:        value,
	})

	return err
}
