package postgres

import (
	"context"
	"database/sql"
	"time"

	listing "gitlab.mashroom.com/mashroom/backend/listing/api"

	"github.com/jmoiron/sqlx"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
)

type repositoryVersionPublication struct {
	save               *sqlx.NamedStmt
	published          *sqlx.Stmt
	listUnsynced       *sqlx.Stmt
	propertyListingIDs *sqlx.NamedStmt
	letPropertyIDs     *sqlx.Stmt
}

func prepareVersionPublication(db *sqlx.DB) (stmt repositoryVersionPublication, err error) {
	stmt.save, err = prepareNamed(db, queryPublicationEntrySave)
	if err != nil {
		return
	}

	stmt.published, err = preparex(db, queryPublicationEntryPublished)
	if err != nil {
		return
	}

	stmt.listUnsynced, err = preparex(db, queryPublicationEntryListUnsynced)
	if err != nil {
		return
	}

	stmt.propertyListingIDs, err = prepareNamed(db, queryPublicationPropertyListingIDs)
	if err != nil {
		return
	}

	stmt.letPropertyIDs, err = preparex(db, queryPublicationLetPropertyIDs)
	if err != nil {
		return
	}

	return
}

const queryPublicationEntrySave = `
with deactivate_old as (
  update version_publications set
    publication_id = '', 
    publication_url = '',
    unpublished_by = coalesce(nullif(:unpublished_by, ''), :published_by),
    unpublished_at = (:published_at)::::timestamp
  where
    :published_at > '1970-01-01'::::timestamp and
    version_publications.unpublished_at <= '1970-01-01'::::timestamp and
    version_publications.platform = :platform and
    version_publications.listing_id != :listing_id and
    version_publications.property_id = :property_id
)
insert into version_publications(
  platform,
  listing_id,
  publication_id,
  publication_url,
  published_by,
  unpublished_by,
  published_at,
  unpublished_at,
  expire_at,
  synced_at,
  property_id,
  invoice_ids,
  let_agreed
) values (
  :platform,
  :listing_id,
  :publication_id,
  :publication_url,
  :published_by,
  :unpublished_by,
  :published_at,
  coalesce(nullif(:unpublished_at, 'null'), '0001-01-01')::::timestamp,
  :expire_at,
  :synced_at,
  :property_id,
  :invoice_ids,
  :let_agreed
) on conflict (listing_id, platform) do update set
  publication_id = :publication_id,
  publication_url = :publication_url,
  unpublished_by = :unpublished_by,
  unpublished_at = coalesce(nullif(:unpublished_at, 'null'), '0001-01-01')::::timestamp,
  published_by = :published_by,
  published_at = coalesce(nullif(:published_at, 'null'), '0001-01-01')::::timestamp,
  expire_at = :expire_at,
  synced_at = :synced_at,
  invoice_ids = array(select distinct unnest(version_publications.invoice_ids || cast(:invoice_ids as text[]))),
  let_agreed = :let_agreed
returning
  id,
  platform,
  listing_id,
  publication_id,
  publication_url,
  published_by,
  unpublished_by,
  published_at,
  unpublished_at,
  expire_at,
  synced_at,
  property_id,
  invoice_ids,
  let_agreed
`

func (repo *repositoryVersionPublication) VersionPublicationEntrySave(
	ctx context.Context,
	e *model.PublicationEntry,
) (err error) {
	return repo.save.GetContext(ctx, e, e)
}

const queryPublicationEntryPublished = `
update version_publications set
  synced_at = now()
where
  platform = $1 and 
  version_publications.unpublished_at <= '1970-01-01'::timestamp and
  version_publications.published_at > '1970-01-01'::timestamp and
  version_publications.publication_id != '' and
  version_publications.publication_url != '' and
  version_publications.property_id = $2
returning
  true
`

func (repo *repositoryVersionPublication) VersionPublicationEntryPublished(
	ctx context.Context,
	platform listing.PublicationPlatform,
	propertyID string,
) (res bool, err error) {
	err = repo.published.GetContext(ctx, &res, platform, propertyID)
	if err == sql.ErrNoRows {
		err = nil
	}

	return
}

const queryPublicationEntryListUnsynced = `
select 
  *
from version_publications where
  platform = $1 and
  version_publications.unpublished_at <= '1970-01-01'::timestamp and
  version_publications.published_at > '1970-01-01'::timestamp and
  (
    (version_publications.publication_id = '' and version_publications.publication_url = '') or 
    (now() - version_publications.synced_at > '48 hours'::interval)
  ) 
`

func (repo *repositoryVersionPublication) VersionPublicationEntryListUnsynced(
	ctx context.Context,
	platform listing.PublicationPlatform,
) (entries []*model.PublicationEntry, err error) {
	err = repo.listUnsynced.SelectContext(ctx, &entries, platform)

	return
}

const queryPublicationPropertyListingIDs = `
select distinct 
  versions.property_id,
  max(versions.listing_id) as listing_id
from version_publications
join versions on 
  versions.listing_id = version_publications.listing_id and
  versions.status = 3 and
  coalesce((versions.notification->'viewing_rarity')::::bool, false) = false
where
  version_publications.published_at > '1970-01-01'::::timestamp and
  version_publications.unpublished_at <= '1970-01-01'::::timestamp and 
  version_publications.published_at <= :since and
  platform = :platform
group by versions.property_id
`

func (repo *repositoryVersionPublication) VersionPublicationPropertyListingIDs(
	ctx context.Context,
	platform listing.PublicationPlatform,
	since time.Time,
) (ids []model.PropertyListingID, err error) {
	err = repo.propertyListingIDs.SelectContext(ctx, &ids, struct {
		Platform listing.PublicationPlatform `db:"platform"`
		Since    time.Time                   `db:"since"`
	}{
		Platform: platform,
		Since:    since,
	})

	return
}

const queryPublicationLetPropertyIDs = `
select distinct 
  versions.property_id
from version_publications
join versions on
  versions.listing_id = version_publications.listing_id and
  versions.status = 4
where
  version_publications.published_at > '1970-01-01'::timestamp and
  version_publications.unpublished_at <= '1970-01-01'::timestamp and
  not exists(
    select
    from versions s
    join version_publications vp on
      s.listing_id = versions.listing_id and
      s.status = 3 and
      s.id > versions.id and
      vp.published_at > '1970-01-01'::timestamp and
      vp.unpublished_at <= '1970-01-01'::timestamp
  )
`

func (repo *repositoryVersionPublication) VersionPublicationLetPropertyIDs(
	ctx context.Context,
) (ids []string, err error) {
	err = repo.letPropertyIDs.SelectContext(ctx, &ids)

	return
}
