package model

import (
	"context"
	"database/sql/driver"
	"encoding/json"
	"errors"
	"strings"
	"time"

	"github.com/lib/pq"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/optional"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	viewing "gitlab.mashroom.com/mashroom/backend/viewing/api"
)

// ServiceTime time formatted in database json
type ServiceTime time.Time

// ErrByteArrayAssert returned when non-matching type is encountered
var ErrByteArrayAssert = errors.New("type assertion to []byte failed")

func jsonUnmarshal(v, value interface{}) error {
	if value == nil {
		return nil
	}

	bs, ok := value.([]byte)
	if !ok {
		return ErrByteArrayAssert
	}

	return json.Unmarshal(bs, v)
}

// MarshalJSON returns json representation
func (v ServiceTime) MarshalJSON() ([]byte, error) {
	if time.Time(v).IsZero() {
		return []byte("null"), nil
	}

	bs := (time.Time)(v).Format(time.RFC3339)

	return []byte(`"` + bs + `"`), nil
}

// UnmarshalJSON parses json representation
func (v *ServiceTime) UnmarshalJSON(data []byte) error {
	d := strings.Trim(string(data), `"`)
	if d == "" || d == "null" {
		return nil
	}

	if !strings.Contains(d, "Z") {
		d += "Z"
	}

	t, err := time.Parse(time.RFC3339, d)
	if err != nil {
		return err
	}

	*v = ServiceTime(t)

	return nil
}

// Value returns serialized representation
func (v ServiceTime) Value() (driver.Value, error) {
	if time.Time(v).IsZero() {
		return time.Time{}, nil
	}

	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *ServiceTime) Scan(value interface{}) error {
	if value == nil {
		return nil
	}

	t, ok := value.(time.Time)
	if ok {
		*v = (ServiceTime)(t)
		return nil
	}

	bs, ok := value.([]byte)
	if !ok {
		return ErrByteArrayAssert
	}

	return json.Unmarshal(bs, v)
}

// VersionDescription represents property description data model
type VersionDescription struct {
	PostHeadline      string   `json:"post_headline,omitempty"`
	PostPropertyPitch string   `json:"post_property_pitch,omitempty"`
	PostAreaPitch     string   `json:"post_area_pitch,omitempty"`
	PostComment       string   `json:"post_comment,omitempty"`
	PostAdminPitch    string   `json:"post_admin_pitch,omitempty"`
	Photos            []string `json:"photos,omitempty"`
	Plans             []string `json:"plans,omitempty"`
	VideoEmbeds       []string `json:"video_embeds,omitempty"`
	Tags              []string `json:"tags,omitempty"`
}

// Value returns serialized representation
func (v VersionDescription) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionDescription) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionDetails represents property details data model
type VersionDetails struct {
	AmenityOutsideSpace []listing.PropertyOutsideSpace `json:"amenity_outside_space,omitempty"`
	AmenityParking      []listing.PropertyParking      `json:"amenity_parking,omitempty"`
	TypeOther           *optional.String               `json:"type_other,omitempty"`
	Area                *optional.Float64              `json:"area,omitempty"`
	BedroomsNo          *optional.Int32                `json:"bedrooms_no,omitempty"`
	BathroomsNo         *optional.Int32                `json:"bathrooms_no,omitempty"`
	LivingroomsNo       *optional.Int32                `json:"livingrooms_no,omitempty"`
	KitchenNo           *optional.Int32                `json:"kitchen_noomitempty,omitempty"`
	PetsAllowed         *optional.Bool                 `json:"pets_allowed,omitempty"`
	Type                listing.PropertyType           `json:"type,omitempty"`
	AmenityFurniture    listing.PropertyFurniture      `json:"amenity_furniture,omitempty"`
}

// Value returns serialized representation
func (v VersionDetails) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionDetails) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// ViewersMap nil-guarded abstraction
type ViewersMap map[string][]viewing.ViewingState

// Uncensored returns true if a user has access to uncensored property
func (m *ViewersMap) Uncensored(user string) bool {
	if m == nil || *m == nil {
		return false
	}

	vs := (*m)[user]

	for _, v := range vs {
		switch v {
		case viewing.ViewingState_VIEWING_STATE_MANAGER_DECISION,
			viewing.ViewingState_VIEWING_STATE_REJECTED,
			viewing.ViewingState_VIEWING_STATE_UNKNOWN:
		default:
			return true
		}
	}

	return false
}

// Set viewing state safely
func (m *ViewersMap) Set(k string, v viewing.ViewingState) {
	if *m == nil {
		*m = make(ViewersMap)
	}

	if len((*m)[k]) > 0 && (*m)[k][len((*m)[k])-1] == v {
		return
	}

	(*m)[k] = append((*m)[k], v)
}

// Get viewing state safely
func (m ViewersMap) Get(k string) viewing.ViewingState {
	if m == nil {
		return 0
	}

	vs := m[k]

	if len(vs) == 0 {
		return 0
	}

	return vs[len(vs)-1]
}

// OfferersMap nil-guarded abstraction
type OfferersMap map[string][]offer.OfferState

// Uncensored returns true if a user has access to uncensored property
func (m *OfferersMap) Uncensored(user string) bool {
	if m == nil || *m == nil {
		return false
	}

	vs := (*m)[user]

	for _, v := range vs {
		switch v {
		case offer.OfferState_CANCELED,
			offer.OfferState_DENIED,
			offer.OfferState_PENDING,
			offer.OfferState_ACTIVE,
			offer.OfferState_UNKNOWN_OFFER_STATE:
		default:
			return true
		}
	}

	return false
}

// Set viewing state safely
func (m *OfferersMap) Set(k string, v offer.OfferState) {
	if *m == nil {
		*m = make(OfferersMap)
	}

	if len((*m)[k]) > 0 && (*m)[k][len((*m)[k])-1] == v {
		return
	}

	(*m)[k] = append((*m)[k], v)
}

// Get viewing state safely
func (m OfferersMap) Get(k string) offer.OfferState {
	if m == nil {
		return 0
	}

	vs := m[k]

	if len(vs) == 0 {
		return 0
	}

	return vs[len(vs)-1]
}

// VersionPeople represents property people data model.
type VersionPeople struct {
	Manager   string      `json:"manager,omitempty"`
	Landlord  string      `json:"landlord,omitempty"`
	Submitter string      `json:"submitter,omitempty"`
	Admin     string      `json:"admin,omitempty"`
	Tenants   []string    `json:"tenants,omitempty"`
	Viewers   ViewersMap  `json:"viewers,omitempty"`
	Offerers  OfferersMap `json:"offerers,omitempty"`
}

// Uncensored returns true if a user has access to uncensored property
func (v *VersionPeople) Uncensored(ctx context.Context, user string) bool {
	if grpcctx.ContextIsAdminUser(ctx) || grpcctx.ContextIsServiceUser(ctx) {
		return true
	}

	if user == "" {
		user = grpcctx.ContextUserID(ctx)
	}

	for _, u := range v.Tenants {
		if u == user {
			return true
		}
	}

	return user == v.Landlord || user == v.Submitter || v.Viewers.Uncensored(user) || v.Offerers.Uncensored(user)
}

// Value returns serialized representation
func (v VersionPeople) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionPeople) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionRental represents property rental data model
type VersionRental struct {
	BillsCovered            []listing.PropertyBillType `json:"bills_covered,omitempty"`
	EarliestMovein          ServiceTime                `json:"earliest_movein,omitempty"`
	RentalPrice             *optional.Float64          `json:"rental_price,omitempty"`
	DepositReplacementFee   *optional.Float64          `json:"deposit_replacement_fee,omitempty"`
	DurationMinMonths       *optional.Int32            `json:"duration_min_months,omitempty"`
	DurationMaxMonths       *optional.Int32            `json:"duration_max_months,omitempty"`
	DssAvailable            *optional.Bool             `json:"dss_available,omitempty"`
	BillsIncluded           *optional.Bool             `json:"bills_included,omitempty"`
	TenancyDepositInsurance *optional.Bool             `json:"tenancy_deposit_insurance,omitempty"`
	InventoryReport         *optional.Bool             `json:"inventory_report,omitempty"`
	BankDetails             BankDetails                `json:"bank_details,omitempty"`
}

// BankDetails for payouts
type BankDetails struct {
	Society           string `json:"society,omitempty"`
	SortCode          string `json:"sort_code,omitempty"`
	AccountNumber     string `json:"account_number,omitempty"`
	AccountName       string `json:"account_name,omitempty"`
	PersonalReference string `json:"personal_reference,omitempty"`
}

// Value returns serialized representation
func (v VersionRental) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionRental) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionLocation represents property location data model
type VersionLocation struct {
	Postcode             string         `json:"postcode,omitempty"`
	Address              string         `json:"address,omitempty"`
	Address1             string         `json:"address1,omitempty"`
	Address2             string         `json:"address2,omitempty"`
	Address3             string         `json:"address3,omitempty"`
	Address4             string         `json:"address4,omitempty"`
	Flat                 string         `json:"flat,omitempty"`
	Building             string         `json:"building,omitempty"`
	Street               string         `json:"street,omitempty"`
	City                 string         `json:"city,omitempty"`
	GPS                  GpsCoordinates `json:"gps,omitempty"`
	PostTown             string         `json:"post_town,omitempty"`
	Country              string         `json:"country,omitempty"`
	County               string         `json:"county,omitempty"`
	AdministrativeCounty string         `json:"administrative_county,omitempty"`
	PostalCounty         string         `json:"postal_county,omitempty"`
	TraditionalCounty    string         `json:"traditional_county,omitempty"`
	District             string         `json:"district,omitempty"`
	GeoRegion            string         `json:"geo_region,omitempty"`
	GeoCounty            string         `json:"geo_county,omitempty"`
	GeoTown              string         `json:"geo_town,omitempty"`
}

// Value returns serialized representation
func (v VersionLocation) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionLocation) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValuationAdditional represents property valuation nested data model
type VersionValuationAdditional struct {
	VariationConfidence        float64 `json:"variation_confidence,omitempty"`
	ComparatorNumberConfidence float64 `json:"comparator_number_confidence,omitempty"`
}

// VersionValuation represents property valuation data model
type VersionValuation struct {
	Estimate   float64                    `json:"estimate,omitempty"`
	Confidence float64                    `json:"confidence,omitempty"`
	Additional VersionValuationAdditional `json:"additional,omitempty"`
	LowerBound float64                    `json:"lower_bound,omitempty"`
	UpperBound float64                    `json:"upper_bound,omitempty"`
}

// Value returns serialized representation
func (v VersionValuation) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValuation) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// GpsCoordinates represents GPS coordinates data model.
type GpsCoordinates struct {
	Longitude float64 `json:"longitude,omitempty"`
	Latitude  float64 `json:"latitude,omitempty"`
}

// Value returns serialized representation
func (v GpsCoordinates) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *GpsCoordinates) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// PublicationEntries wraps publications historical list
type PublicationEntries []*PublicationEntry

// Published returns true if latest publication is active
func (pe PublicationEntries) Published() bool {
	if len(pe) == 0 {
		return false
	}

	p := pe[len(pe)-1]

	return p.ShouldBePublished() && (p.PublicationID != "" || p.PublicationURL != "")
}

// ShouldBePublished returns true if this publication is supposed to be published
func (p *PublicationEntry) ShouldBePublished() bool {
	if time.Time(p.UnpublishedAt).Unix() > 0 {
		return false
	}

	return time.Time(p.PublishedAt).Unix() > 0
}

// Last returns latest publication
func (pe PublicationEntries) Last() *PublicationEntry {
	if len(pe) == 0 {
		return &PublicationEntry{}
	}

	return pe[len(pe)-1]
}

// Scan parses serialized representation
func (pe *PublicationEntries) Scan(value interface{}) error {
	return jsonUnmarshal(pe, value)
}

// PublicationEntry represents listing publication item data model
type PublicationEntry struct {
	PublishedAt    ServiceTime                 `db:"published_at" json:"published_at"`
	UnpublishedAt  ServiceTime                 `db:"unpublished_at" json:"unpublished_at"`
	ExpireAt       ServiceTime                 `db:"expire_at" json:"expire_at"`
	SyncedAt       ServiceTime                 `db:"synced_at" json:"synced_at"`
	ListingID      string                      `db:"listing_id" json:"listing_id,omitempty"`
	PropertyID     string                      `db:"property_id" json:"property_id,omitempty"`
	PublicationID  string                      `db:"publication_id" json:"publication_id,omitempty"`
	PublicationURL string                      `db:"publication_url" json:"publication_url,omitempty"`
	PublishedBy    string                      `db:"published_by" json:"published_by,omitempty"`
	UnpublishedBy  string                      `db:"unpublished_by" json:"unpublished_by,omitempty"`
	InvoiceIDs     pq.StringArray              `db:"invoice_ids" json:"invoice_ids"`
	ID             uint64                      `db:"id" json:"id,omitempty"`
	Platform       listing.PublicationPlatform `db:"platform" json:"platform,omitempty"`
	LetAgreed      bool                        `db:"let_agreed" json:"let_agreed"`
}

// VersionNotificationType is key for verification flag
type VersionNotificationType string

const (
	// VersionNotificationVerificationPending key for notify flag about penging verification
	VersionNotificationVerificationPending VersionNotificationType = "verification_pending"
	// VersionNotificationViewingRarity key for notify flag about property receives less than 3 viewings a week
	VersionNotificationViewingRarity VersionNotificationType = "viewing_rarity"
)

// VersionVerification represents admin verifications data model
type VersionVerification struct {
	Address     AdminVerifyItem `json:"address,omitempty"`
	Description AdminVerifyItem `json:"description,omitempty"`
	Photos      AdminVerifyItem `json:"photos,omitempty"`
	Documents   AdminVerifyItem `json:"documents,omitempty"`
	Fee         AdminVerifyItem `json:"fee,omitempty"`
}

// Value returns serialized representation
func (v VersionVerification) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionVerification) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// AdminVerifyItem represents admin verify item data model
type AdminVerifyItem struct {
	Admin      string      `json:"admin,omitempty"`
	Verified   bool        `json:"verified,omitempty"`
	VerifiedAt ServiceTime `json:"verified_at,omitempty"`
}

// Value returns serialized representation
func (v AdminVerifyItem) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *AdminVerifyItem) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// ValidationLocation stores location validation errors
type ValidationLocation struct {
	Postcode             []listing.ValidationError `json:"postcode,omitempty"`
	Address1             []listing.ValidationError `json:"address1,omitempty"`
	Address2             []listing.ValidationError `json:"address2,omitempty"`
	Address3             []listing.ValidationError `json:"address3,omitempty"`
	Address4             []listing.ValidationError `json:"address4,omitempty"`
	Flat                 []listing.ValidationError `json:"flat,omitempty"`
	Building             []listing.ValidationError `json:"building,omitempty"`
	Street               []listing.ValidationError `json:"street,omitempty"`
	City                 []listing.ValidationError `json:"city,omitempty"`
	GPS                  []listing.ValidationError `json:"gps,omitempty"`
	Country              []listing.ValidationError `json:"country,omitempty"`
	PostTown             []listing.ValidationError `json:"post_town,omitempty"`
	County               []listing.ValidationError `json:"county,omitempty"`
	AdministrativeCounty []listing.ValidationError `json:"administrative_county,omitempty"`
	PostalCounty         []listing.ValidationError `json:"postal_county,omitempty"`
	TraditionalCounty    []listing.ValidationError `json:"traditional_county,omitempty"`
	District             []listing.ValidationError `json:"district,omitempty"`
	Address              []listing.ValidationError `json:"address,omitempty"`
	GeoRegion            []listing.ValidationError `json:"geo_region,omitempty"`
	GeoCounty            []listing.ValidationError `json:"geo_county,omitempty"`
	GeoTown              []listing.ValidationError `json:"geo_town,omitempty"`
}

// Value returns serialized representation
func (v ValidationLocation) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *ValidationLocation) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidationDetails stores details validation errors
type VersionValidationDetails struct {
	Type                []listing.ValidationError `json:"type,omitempty"`
	TypeOther           []listing.ValidationError `json:"type_other,omitempty"`
	Area                []listing.ValidationError `json:"area,omitempty"`
	BedroomsNo          []listing.ValidationError `json:"bedrooms_no,omitempty"`
	BathroomsNo         []listing.ValidationError `json:"bathrooms_no,omitempty"`
	LivingroomsNo       []listing.ValidationError `json:"livingrooms_no,omitempty"`
	KitchenNo           []listing.ValidationError `json:"kitchen_no,omitempty"`
	AmenityFurniture    []listing.ValidationError `json:"amenity_furniture,omitempty"`
	AmenityOutsideSpace []listing.ValidationError `json:"amenity_outside_space,omitempty"`
	AmenityParking      []listing.ValidationError `json:"amenity_parking,omitempty"`
	PetsAllowed         []listing.ValidationError `json:"pets_allowed,omitempty"`
}

// Value returns serialized representation
func (v VersionValidationDetails) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidationDetails) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidationRental stores rental validation errors
type VersionValidationRental struct {
	RentalPrice             []listing.ValidationError `json:"rental_price,omitempty"`
	DssAvailable            []listing.ValidationError `json:"dss_available,omitempty"`
	BillsIncluded           []listing.ValidationError `json:"bills_included,omitempty"`
	BillsCovered            []listing.ValidationError `json:"bills_covered,omitempty"`
	DurationMinMonths       []listing.ValidationError `json:"duration_min_months,omitempty"`
	DurationMaxMonths       []listing.ValidationError `json:"duration_max_months,omitempty"`
	EarliestMovein          []listing.ValidationError `json:"earliest_movein,omitempty"`
	TenancyDepositInsurance []listing.ValidationError `json:"tenancy_deposit_insurance,omitempty"`
	InventoryReport         []listing.ValidationError `json:"inventory_report,omitempty"`
	DepositReplacementFee   []listing.ValidationError `json:"deposit_replacement_fee,omitempty"`
}

// Value returns serialized representation
func (v VersionValidationRental) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidationRental) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidationDescription stores description validation errors
type VersionValidationDescription struct {
	PostHeadline         []listing.ValidationError `json:"post_headline,omitempty"`
	PostPropertyPitch    []listing.ValidationError `json:"post_property_pitch,omitempty"`
	PostOuterDescription []listing.ValidationError `json:"post_outer_description,omitempty"`
	PostAreaPitch        []listing.ValidationError `json:"post_area_pitch,omitempty"`
	PostComment          []listing.ValidationError `json:"post_comment,omitempty"`
	PostAdminPitch       []listing.ValidationError `json:"post_admin_pitch,omitempty"`
	Photos               []listing.ValidationError `json:"photos,omitempty"`
	Plans                []listing.ValidationError `json:"plans,omitempty"`
	VideoEmbeds          []listing.ValidationError `json:"video_embeds,omitempty"`
	Tags                 []listing.ValidationError `json:"tags,omitempty"`
}

// Value returns serialized representation
func (v VersionValidationDescription) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidationDescription) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidationPeople stores people validation errors
type VersionValidationPeople struct {
	Manager   []listing.ValidationError `json:"manager,omitempty"`
	Landlord  []listing.ValidationError `json:"landlord,omitempty"`
	Submitter []listing.ValidationError `json:"submitter,omitempty"`
}

// Value returns serialized representation
func (v VersionValidationPeople) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidationPeople) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidationDocuments stores document validation errors
type VersionValidationDocuments struct {
	EPC []listing.ValidationError `json:"epc,omitempty"`
}

// Value returns serialized representation
func (v VersionValidationDocuments) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidationDocuments) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionValidation stores listing version validation errors and overall result
type VersionValidation struct {
	Valid       bool                         `json:"valid,omitempty"`
	Verifiable  bool                         `json:"verifiable,omitempty"`
	Location    ValidationLocation           `json:"location,omitempty"`
	Details     VersionValidationDetails     `json:"details,omitempty"`
	Rental      VersionValidationRental      `json:"rental,omitempty"`
	Description VersionValidationDescription `json:"description,omitempty"`
	People      VersionValidationPeople      `json:"people,omitempty"`
	Documents   VersionValidationDocuments   `json:"documents,omitempty"`
}

// Value returns serialized representation
func (v VersionValidation) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionValidation) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionOriginProduct stores listing version origin product details
type VersionOriginProduct struct {
	ProductID   string `json:"product_id,omitempty"`
	ProductName string `json:"product_name,omitempty"`
}

// Value returns serialized representation
func (v VersionOriginProduct) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionOriginProduct) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionOriginTool stores listing version origin tool details
type VersionOriginTool struct {
	Valuation             bool `json:"valuation,omitempty"`
	ExpenseTracker        bool `json:"expense_tracker,omitempty"`
	CertificateRepository bool `json:"certificate_repository,omitempty"`
}

// Value returns serialized representation
func (v VersionOriginTool) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionOriginTool) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionOrigin stores listing version origin details
type VersionOrigin struct {
	Listing bool                  `json:"listing,omitempty"`
	Preface bool                  `json:"preface,omitempty"`
	Product *VersionOriginProduct `json:"product,omitempty"`
	Tool    *VersionOriginTool    `json:"tool,omitempty"`
}

// Value returns serialized representation
func (v VersionOrigin) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionOrigin) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionViewingAvailabilitySlot stores interval when viewing can be scheduled
type VersionViewingAvailabilitySlot struct {
	Period      listing.Period `json:"period,omitempty"`
	StartOffset time.Duration  `json:"start_offset,omitempty"`
	EndOffset   time.Duration  `json:"end_offset,omitempty"`
}

// Value returns serialized representation
func (v VersionViewingAvailabilitySlot) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionViewingAvailabilitySlot) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionViewingAvailability stores list of viewing availability slots
type VersionViewingAvailability []*VersionViewingAvailabilitySlot

// Value returns serialized representation
func (v VersionViewingAvailability) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionViewingAvailability) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionInvitationStatus stores invitation state
type VersionInvitationStatus struct {
	Pending      bool        `json:"pending,omitempty"`
	Sent         ServiceTime `json:"sent,omitempty"`
	HostID       string      `json:"host_id,omitempty"`
	InviteeID    string      `json:"invitee_id,omitempty"`
	InviteeEmail string      `json:"invitee_email,omitempty"`
	InviteePhone string      `json:"invitee_phone,omitempty"`
	InviteeName  string      `json:"invitee_name,omitempty"`
	InvitationID string      `json:"invitation_id,omitempty"`
}

// Value returns serialized representation
func (v VersionInvitationStatus) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionInvitationStatus) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionInvitation stores invitation states
type VersionInvitation struct {
	ManagerInvitation  VersionInvitationStatus `json:"manager_invitation,omitempty"`
	LandlordInvitation VersionInvitationStatus `json:"landlord_invitation,omitempty"`
}

// Value returns serialized representation
func (v VersionInvitation) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionInvitation) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionFlags stores listing version flags
type VersionFlags struct {
	Disabled    bool   `json:"disabled,omitempty"`
	Test        bool   `json:"test,omitempty"`
	Auto        bool   `json:"auto,omitempty"`
	ActiveOffer string `json:"active_offer,omitempty"`
}

// Value returns serialized representation
func (v VersionFlags) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *VersionFlags) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// Version represents listing statement
type Version struct {
	Description           VersionDescription         `db:"description"`
	Details               VersionDetails             `db:"details"`
	People                VersionPeople              `db:"people"`
	Rental                VersionRental              `db:"rental"`
	Location              VersionLocation            `db:"location"`
	Valuation             VersionValuation           `db:"valuation"`
	Verification          VersionVerification        `db:"verification"`
	Validation            VersionValidation          `db:"validation"`
	Origin                VersionOrigin              `db:"origin"`
	ViewingAvailability   VersionViewingAvailability `db:"viewing_availability"`
	Invitation            VersionInvitation          `db:"invitation"`
	Created               time.Time                  `db:"created"`
	Updated               time.Time                  `db:"updated"`
	Published             time.Time                  `db:"published"`
	PublicationsMashroom  PublicationEntries         `db:"publications_mashroom"`
	PublicationsZoopla    PublicationEntries         `db:"publications_zoopla"`
	PublicationsRightmove PublicationEntries         `db:"publications_rightmove"`
	PublicationsBoomin    PublicationEntries         `db:"publications_boomin"`
	Flags                 VersionFlags               `db:"flags"`
	Timezone              string                     `db:"timezone"`
	PropertyID            string                     `db:"property_id"`
	TenancyID             string                     `db:"tenancy_id"`
	ListingID             string                     `db:"listing_id"`
	IVR                   string                     `db:"ivr"`
	MigratedID            string                     `db:"migrated_id"` // temporary
	LastVerified          uint64                     `db:"last_verified"`
	ID                    uint64                     `db:"id"`
	Status                listing.VersionStatus      `db:"status"`
	Latest                bool                       `db:"latest"` // only exists in view
	Notification          Notification               `db:"notification"`
}

// Notification represents send notifications of listing
type Notification map[VersionNotificationType]bool

// Set notification in type-safe manner
func (v *Notification) Set(k VersionNotificationType, done bool) {
	if *v == nil {
		*v = make(Notification)
	}

	(*v)[k] = done
}

// Get notification status in type-safe manner
func (v Notification) Get(k VersionNotificationType) bool {
	if v == nil {
		return false
	}

	return v[k]
}

// Value returns serialized representation
func (v Notification) Value() (driver.Value, error) {
	return json.Marshal(v)
}

// Scan parses serialized representation
func (v *Notification) Scan(value interface{}) error {
	return jsonUnmarshal(v, value)
}

// VersionAverageRentOptions average rent query options
type VersionAverageRentOptions struct {
	Country  string
	Region   string
	County   string
	Town     string
	Statuses []listing.VersionStatus
}

// VersionAverageRentItem average rent distribution per bedrooms number
type VersionAverageRentItem struct {
	Bedrooms int     `db:"bedrooms"`
	Total    uint64  `db:"total"`
	Price    float64 `db:"price"`
}
