// Package service provides implementation for invoices service
package service

import (
	"context"
	"fmt"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/worker"
	"github.com/camunda-cloud/zeebe/clients/go/pkg/zbc"
	"gitlab.mashroom.com/mashroom/backend/common/api/process"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/intel"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
)

// Service contract
type Service interface {
	listing.ListingServiceServer
	process.ProcessMonitorServiceServer
	invoice.InvoiceNotifyServiceServer
	schedule.ReminderConsumerServiceServer
	invitation.InvitationNotifyServiceServer
	UploadProcesses() error
	RegisterProcessHandlers() (worker.JobWorker, error)
	SetupSchedule(ctx context.Context) error
}

// Config for Service
type Config struct {
	Repo         model.Repository
	ZeebeClient  zbc.Client
	Connections  connections.Connections
	SiteURL      string
	IntelClient  intel.Client
	TaskRegistry *schedule.Registry
}

func validateConfigConnections(connections *connections.Connections) error {
	if connections.SelfRemote == "" {
		return fmt.Errorf("no self remote address")
	}

	if connections.Account == nil {
		return fmt.Errorf("no account client")
	}

	if connections.Geo == nil {
		return fmt.Errorf("no geo client")
	}

	if connections.Document == nil {
		return fmt.Errorf("no document client")
	}

	if connections.Address == nil {
		return fmt.Errorf("no address client")
	}

	if connections.Invoice == nil {
		return fmt.Errorf("no invoice client")
	}

	if connections.Extras == nil {
		return fmt.Errorf("no extras client")
	}

	if connections.Notification == nil {
		return fmt.Errorf("no notification client")
	}

	if connections.Closecom == nil {
		return fmt.Errorf("no closecom client")
	}

	err := connections.Publication.Validate()
	if err != nil {
		return err
	}

	return nil
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid service config")
	}

	if err := validateConfigConnections(&config.Connections); err != nil {
		return err
	}

	if config.Repo == nil {
		return fmt.Errorf("no repository in service config")
	}

	return nil
}

// NewService returns new Service instance
func NewService(config *Config) (Service, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	return &serviceImpl{
		Config: *config,
	}, nil
}

func (srv *serviceImpl) SetupSchedule(ctx context.Context) error {
	return srv.TaskRegistry.Setup(ctx, srv.Connections.Schedule, "listing", srv.Connections.SelfRemote)
}
