// Package publication provides utilities to publish and unpublish listing versions
package publication

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service/render"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

// VersionUnpublishExternal publish on external platform
func VersionUnpublishExternal(
	ctx context.Context,
	repo model.Repository,
	external listing.ListingPublicationServiceClient,
	entries model.PublicationEntries,
	propertyID, userID string,
) error {
	if !entries.Published() {
		return nil
	}

	last := entries.Last()

	_, err := external.ListingPublicationRemove(ctx, &listing.ListingPublicationRemoveRequest{
		PublicationId: last.PublicationID,
		PropertyId:    propertyID,
	})
	if err != nil {
		log.WithError(err).WithField("property", propertyID).Error("unpublishing")
	} else {
		last.SyncedAt = model.ServiceTime(time.Now())
		last.PublicationID = ""
		last.PublicationURL = ""
	}

	last.UnpublishedAt = model.ServiceTime(time.Now())
	last.UnpublishedBy = userID

	return repo.VersionPublicationEntrySave(ctx, last)
}

// VersionUnpublishEverywhere unpublish on every platform
func VersionUnpublishEverywhere(
	ctx context.Context,
	repo model.Repository,
	conns *connections.Publications,
	propertyID string,
	statuses ...listing.VersionStatus,
) error {
	lv, err := repo.VersionGet(ctx, &model.VersionSpecOptions{
		PropertyID: propertyID,
		Statuses:   statuses,
	})
	if err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil
		}

		return err
	}

	if lv.PublicationsMashroom.Published() {
		last := lv.PublicationsMashroom.Last()
		last.UnpublishedAt = model.ServiceTime(time.Now())

		err = repo.VersionPublicationEntrySave(ctx, last)
		if err != nil {
			log.WithError(err).WithField("version", lv.ID).Error("unpublishing on mashroom")
		} else {
			last.SyncedAt = model.ServiceTime(time.Now())
			last.PublicationID = ""
			last.PublicationURL = ""
		}
	}

	err = VersionUnpublishExternal(
		ctx,
		repo,
		conns.Rightmove,
		lv.PublicationsRightmove,
		lv.PropertyID,
		"",
	)
	if err != nil {
		return err
	}

	err = VersionUnpublishExternal(
		ctx,
		repo,
		conns.Zoopla,
		lv.PublicationsZoopla,
		lv.PropertyID,
		"",
	)
	if err != nil {
		return err
	}

	err = VersionUnpublishExternal(
		ctx,
		repo,
		conns.Boomin,
		lv.PublicationsBoomin,
		lv.PropertyID,
		"",
	)
	if err != nil {
		return err
	}

	return nil
}

// VersionPublishMashroom publish on mashroom platform
func VersionPublishMashroom(
	ctx context.Context,
	repo model.Repository,
	item *listing.VersionPublicationChangeItem,
	lv *model.Version,
	user string,
	let *wrapperspb.BoolValue,
) error {
	eligible, agreed := versionEligibleForPublish(item, lv, lv.PublicationsMashroom, let)
	if !eligible {
		return nil
	}

	entry := &model.PublicationEntry{
		Platform:      listing.PublicationPlatform_PUBLICATION_PLATFORM_MASHROOM,
		ListingID:     lv.ListingID,
		PropertyID:    lv.PropertyID,
		PublicationID: lv.PropertyID,
		PublishedBy:   user,
		PublishedAt:   model.ServiceTime(time.Now()),
		LetAgreed:     agreed,
		SyncedAt:      model.ServiceTime(time.Now()),
	}

	err := repo.VersionPublicationEntrySave(ctx, entry)
	if err != nil {
		return err
	}

	lv.PublicationsMashroom = append(lv.PublicationsMashroom, entry)

	return nil
}

// VersionPublishExternal publish on one of external platforms
func VersionPublishExternal(
	ctx context.Context,
	repo model.Repository,
	external listing.ListingPublicationServiceClient,
	platform listing.PublicationPlatform,
	entries *model.PublicationEntries,
	item *listing.VersionPublicationChangeItem,
	lv *model.Version,
	user string,
	publication *listing.VersionPublicationChangeItem,
	let *wrapperspb.BoolValue,
) error {
	eligible, agreed := versionEligibleForPublish(item, lv, *entries, let)
	if !eligible {
		return nil
	}

	resp, perr := external.ListingPublicationSave(ctx, &listing.ListingPublicationSaveRequest{
		PublicationId: entries.Last().PublicationID,
		PropertyId:    lv.PropertyID,
		ListingId:     lv.ListingID,
		PropertyIvr:   lv.IVR,
		Available:     !agreed,
		Location:      render.Location(&lv.Location, true),
		Rental:        render.Rental(&lv.Rental, user == lv.People.Landlord),
		Details:       render.Details(&lv.Details),
		Description:   render.Description(&lv.Description),
	})

	if perr == nil && (resp.GetPublicationId() == "" || resp.GetPublicationUrl() == "") {
		perr = grpcerr.New(
			grpcerr.ProviderError,
			fmt.Sprintf("publishing on %v: %s", platform, lv.PropertyID),
		)
	}

	var synced model.ServiceTime

	if perr == nil {
		synced = model.ServiceTime(time.Now())
	}

	exp := VersionPublicationExpiry(*entries, publication.GetExpireAt().AsTime())

	entry := &model.PublicationEntry{
		Platform:       platform,
		ListingID:      lv.ListingID,
		PropertyID:     lv.PropertyID,
		PublicationID:  resp.GetPublicationId(),
		PublicationURL: resp.GetPublicationUrl(),
		PublishedBy:    user,
		PublishedAt:    model.ServiceTime(time.Now()),
		ExpireAt:       model.ServiceTime(exp),
		SyncedAt:       synced,
		LetAgreed:      agreed,
	}

	err := repo.VersionPublicationEntrySave(ctx, entry)
	if err != nil {
		return err
	}

	*entries = append(*entries, entry)

	return perr
}

// VersionPublicationExpiry publication expiration time
func VersionPublicationExpiry(entries model.PublicationEntries, exp time.Time) time.Time {
	oldexp := (time.Time)(entries.Last().ExpireAt)

	if oldexp.Unix() > 0 && ((exp.Unix() <= 0 && oldexp.After(time.Now())) || exp.Before(oldexp)) {
		return oldexp
	}

	if exp.Unix() <= 0 {
		exp = time.Now().AddDate(0, 0, 30)
	}

	return exp
}

func versionEligibleForPublish(
	item *listing.VersionPublicationChangeItem,
	lv *model.Version,
	entries model.PublicationEntries,
	let *wrapperspb.BoolValue,
) (eligable, agreed bool) {
	if lv.Flags.Test || lv.TenancyID != "" || lv.Status != listing.VersionStatus_VERSION_VERIFIED {
		return false, false
	}

	if item == nil && (let == nil || !entries.Published()) {
		return false, false
	}

	agreed = entries.Last().LetAgreed
	if let != nil {
		agreed = let.GetValue()
	}

	return true, agreed
}
