package render

import (
	"fmt"
	"strings"

	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
)

func engPlural(i int) int {
	if i == 0 {
		return 0
	}

	if i == 1 {
		return 1
	}

	return 2
}

func pluralize(value int, plural func(i int) int, cases ...string) string {
	idx := plural(value)
	if idx > len(cases) {
		return ""
	}

	return cases[idx]
}

func pluralizef(value int, plural func(i int) int, cases ...string) string {
	return fmt.Sprintf(pluralize(value, plural, cases...), value)
}

func renderAndList(s []string) string {
	if len(s) == 0 {
		return ""
	}

	if len(s) == 1 {
		return s[0]
	}

	p1 := strings.Join(s[:len(s)-1], ", ")
	p2 := " and " + s[len(s)-1]

	return p1 + p2
}

// GenerateDescription providers listing description from version features
func GenerateDescription(
	det *listing.Details,
	desc *listing.Description,
	loc *listing.Location,
) string {
	descr := &strings.Builder{}

	descr.WriteString("No deposit required.\n\nCould this ")

	if det.GetBedroomsNo().GetValue() > 0 {
		descr.WriteString(fmt.Sprintf("%d bedroom ", det.GetBedroomsNo().GetValue()))
	}

	descr.WriteString(fmt.Sprintf(
		"%s in %s be the one you call home?\n\n",
		generateDescriptionPropertyType(det.GetType()),
		loc.GetStreet().GetValue(),
	))

	rooms := generateDescriptionRooms(det)

	descr.WriteString(pluralize(len(rooms), engPlural, "", "There is ", "There are "))
	descr.WriteString(renderAndList(rooms))

	features := generateDescriptionFeatures(det, desc)

	var filteredFeatures []string

	for _, f := range features {
		if strings.ToLower(f) != "none of above" {
			filteredFeatures = append(filteredFeatures, f)
		}
	}

	if len(filteredFeatures) > 0 {
		descr.WriteString(" and the property features handy additions such as ")
		descr.WriteString(renderAndList(filteredFeatures))
	}

	descr.WriteString(".\n\n")

	return descr.String()
}

func generateDescriptionRooms(det *listing.Details) (rooms []string) {
	livingRoomNum := det.GetLivingroomsNo().GetValue()
	if livingRoomNum == 0 {
		livingRoomNum = 1
	}

	rooms = append(rooms, pluralizef(
		int(livingRoomNum),
		engPlural,
		"", "%d reception room", "%d reception rooms",
	))

	if det.GetBedroomsNo().GetValue() > 0 {
		rooms = append(rooms, pluralizef(
			int(det.GetBedroomsNo().GetValue()),
			engPlural,
			"", "%d bedroom", "%d bedrooms",
		))
	}

	if det.GetBathroomsNo().GetValue() > 0 {
		rooms = append(rooms, pluralizef(
			int(det.GetBathroomsNo().GetValue()),
			engPlural,
			"", "%d bathroom", "%d bathrooms",
		))
	}

	return rooms
}

func generateDescriptionFeatures(det *listing.Details, desc *listing.Description) (features []string) {
	for _, p := range det.GetAmenityOutsideSpace().GetValue() {
		a := generateDescriptionPropertyOutsideSpace(p)
		if a != "" {
			features = append(features, a)
		}
	}

	for _, p := range det.GetAmenityParking().GetValue() {
		a := generateDescriptionPropertyParking(p)
		if a != "" {
			features = append(features, a)
		}
	}

	if desc.GetTags() != nil && len(desc.GetTags().GetValue()) > 0 {
		for _, p := range desc.GetTags().GetValue() {
			f := strings.ToLower(p)
			switch f {
			case "basement", "conservatory", "fireplace", "gym", "loft", "burglar alarm", "swimming pool",
				"utility room", "tennis court":
				features = append(features, fmt.Sprintf("a %s", f))
			default:
				features = append(features, f)
			}
		}
	}

	return features
}

func generateDescriptionPropertyParking(t listing.PropertyParking) string {
	switch t {
	case listing.PropertyParking_PROPERTY_PARKING_GARAGE:
		return "garage"
	case listing.PropertyParking_PROPERTY_PARKING_OFFSTREET:
		return "off-street parking"
	case listing.PropertyParking_PROPERTY_PARKING_RESIDENTS_PARKING:
		return "resident parking"
	case listing.PropertyParking_PROPERTY_PARKING_DOUBLE_GARAGE:
		return "double garage"
	default:
		return ""
	}
}

func generateDescriptionPropertyOutsideSpace(t listing.PropertyOutsideSpace) string {
	switch t {
	case listing.PropertyOutsideSpace_PROPERTY_OUTSIDE_SPACE_BALCONY:
		return "balcony"
	case listing.PropertyOutsideSpace_PROPERTY_OUTSIDE_SPACE_TERRACE:
		return "terrace"
	case listing.PropertyOutsideSpace_PROPERTY_OUTSIDE_SPACE_COMMUNAL_GARDEN:
		return "communal garden"
	case listing.PropertyOutsideSpace_PROPERTY_OUTSIDE_SPACE_PRIVATE_GARDEN:
		return "private garden"
	default:
		return ""
	}
}

func generateDescriptionPropertyType(t listing.PropertyType) string {
	switch t {
	case listing.PropertyType_PROPERTY_TYPE_FLAT:
		return "flat"
	case listing.PropertyType_PROPERTY_TYPE_DETACHED:
		return "detached house"
	case listing.PropertyType_PROPERTY_TYPE_SEMI_DETACHED:
		return "semi-detached house"
	case listing.PropertyType_PROPERTY_TYPE_TERRACED:
		return "terrace house"
	case listing.PropertyType_PROPERTY_TYPE_STUDIO:
		return "studio"
	case listing.PropertyType_PROPERTY_TYPE_MAISONETTE:
		return "maisonette"
	case listing.PropertyType_PROPERTY_TYPE_BUNGALOW:
		return "bungalow"
	case listing.PropertyType_PROPERTY_TYPE_TOWN_HOUSE:
		return "town house"
	case listing.PropertyType_PROPERTY_TYPE_COTTAGE:
		return "cottage"
	case listing.PropertyType_PROPERTY_TYPE_APARTMENT:
		return "apartment"
	case listing.PropertyType_PROPERTY_TYPE_PENTHOUSE:
		return "penthouse"
	case listing.PropertyType_PROPERTY_TYPE_PARK_HOME:
		return "park home"
	default:
		return "property"
	}
}
