package render

import (
	"context"
	"strconv"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"

	"gitlab.mashroom.com/mashroom/backend/common/api/generic"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"google.golang.org/protobuf/types/known/durationpb"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

// Version renderer
func Version(
	ctx context.Context,
	m *model.Version,
	nearby *listing.Nearby,
	user string,
) *listing.Version {
	var lastVerified string

	if m.LastVerified != 0 {
		lastVerified = strconv.FormatUint(m.LastVerified, 10)
	}

	ver := &listing.Version{
		Id:                  strconv.FormatUint(m.ID, 10),
		ListingId:           m.ListingID,
		PropertyId:          m.PropertyID,
		TenancyId:           m.TenancyID,
		Ivr:                 m.IVR,
		Status:              m.Status,
		CreatedAt:           timestamppb.New(m.Created),
		UpdatedAt:           timestamppb.New(m.Updated),
		Description:         Description(&m.Description),
		Details:             Details(&m.Details),
		People:              People(&m.People),
		Rental:              Rental(&m.Rental, grpcctx.ContextIsAdminUser(ctx) || user == m.People.Landlord),
		Location:            Location(&m.Location, m.People.Uncensored(ctx, user)),
		Valuation:           Valuation(&m.Valuation),
		Origin:              Origin(&m.Origin),
		Nearby:              nearby,
		Publication:         Publication(m),
		Verification:        Verification(&m.Verification),
		Invitation:          Invitation(&m.Invitation),
		Validation:          Validation(&m.Validation),
		ViewingAvailability: ViewingAvailability(m.ViewingAvailability),
		Flags:               Flags(&m.Flags),
		Awards:              nil,
		LastVerified:        lastVerified,
	}

	if ver.GetDescription().GetPostPropertyPitch().GetValue() == "" {
		ver.Description.PostPropertyPitch = wrapperspb.String(GenerateDescription(
			ver.GetDetails(),
			ver.GetDescription(),
			ver.GetLocation(),
		))
	}

	return ver
}

// Origin renderer
func Origin(v *model.VersionOrigin) *listing.VersionOrigin {
	orig := &listing.VersionOrigin{}

	switch {
	case v.Listing:
		orig.Origin = &listing.VersionOrigin_Listing{Listing: true}
	case v.Preface:
		orig.Origin = &listing.VersionOrigin_Preface{Preface: true}
	case v.Tool != nil:
		switch {
		case v.Tool.Valuation:
			orig.Origin = &listing.VersionOrigin_Tool{Tool: &listing.VersionOriginTool{
				Tool: &listing.VersionOriginTool_Valuation{Valuation: true},
			}}
		case v.Tool.ExpenseTracker:
			orig.Origin = &listing.VersionOrigin_Tool{Tool: &listing.VersionOriginTool{
				Tool: &listing.VersionOriginTool_ExpenseTracker{ExpenseTracker: true},
			}}
		case v.Tool.CertificateRepository:
			orig.Origin = &listing.VersionOrigin_Tool{Tool: &listing.VersionOriginTool{
				Tool: &listing.VersionOriginTool_CertificateRepository{CertificateRepository: true},
			}}
		}
	case v.Product != nil:
		orig.Origin = &listing.VersionOrigin_Product{Product: &listing.VersionOriginProduct{
			ProductId:   v.Product.ProductID,
			ProductName: v.Product.ProductName,
		}}
	}

	return orig
}

// ViewingAvailability renderer
func ViewingAvailability(
	availability model.VersionViewingAvailability,
) *listing.ViewingAvailability {
	var slots []*listing.ViewingAvailabilitySlot

	for _, a := range availability {
		slots = append(slots, &listing.ViewingAvailabilitySlot{
			Period:      a.Period,
			StartOffset: durationpb.New(a.StartOffset),
			EndOffset:   durationpb.New(a.EndOffset),
		})
	}

	return &listing.ViewingAvailability{
		Slots: slots,
	}
}

// Flags renderer
func Flags(
	flags *model.VersionFlags,
) *listing.Flags {
	return &listing.Flags{
		Disabled:  wrapperspb.Bool(flags.Disabled),
		Test:      wrapperspb.Bool(flags.Test),
		Preface:   nil,
		Auto:      wrapperspb.Bool(flags.Auto),
		Withdrawn: nil,
	}
}

func renderOptionalStringArray(v []string) *generic.StringArrayValue {
	if v == nil {
		return nil
	}

	return &generic.StringArrayValue{Value: v}
}

// Description renderer
func Description(
	m *model.VersionDescription,
) *listing.Description {
	return &listing.Description{
		PostHeadline:         wrapperspb.String(m.PostHeadline),
		PostPropertyPitch:    wrapperspb.String(m.PostPropertyPitch),
		PostOuterDescription: wrapperspb.String(m.PostPropertyPitch),
		PostAreaPitch:        wrapperspb.String(m.PostAreaPitch),
		PostComment:          wrapperspb.String(m.PostComment),
		PostAdminPitch:       wrapperspb.String(m.PostAdminPitch),
		Photos:               renderOptionalStringArray(m.Photos),
		Plans:                renderOptionalStringArray(m.Plans),
		VideoEmbeds:          renderOptionalStringArray(m.VideoEmbeds),
		Tags:                 renderOptionalStringArray(m.Tags),
	}
}

// Details renderer
func Details(
	m *model.VersionDetails,
) *listing.Details {
	return &listing.Details{
		Type:                m.Type,
		TypeOther:           m.TypeOther.Proto(),
		Area:                m.Area.Proto(),
		BedroomsNo:          m.BedroomsNo.Proto(),
		BathroomsNo:         m.BathroomsNo.Proto(),
		LivingroomsNo:       m.LivingroomsNo.Proto(),
		KitchenNo:           m.KitchenNo.Proto(),
		AmenityFurniture:    m.AmenityFurniture,
		AmenityOutsideSpace: &listing.PropertyOutsideSpaceArrayValue{Value: m.AmenityOutsideSpace},
		AmenityParking:      &listing.PropertyParkingArrayValue{Value: m.AmenityParking},
		PetsAllowed:         m.PetsAllowed.Proto(),
	}
}

// People renderer
func People(
	m *model.VersionPeople,
) *listing.People {
	res := &listing.People{
		Manager:   wrapperspb.String(m.Manager),
		Landlord:  wrapperspb.String(m.Landlord),
		Submitter: wrapperspb.String(m.Submitter),
		Tenants:   &generic.StringArrayValue{Value: m.Tenants},
		Admin:     wrapperspb.String(m.Admin),
	}

	return res
}

// Invitation renderer
func Invitation(
	m *model.VersionInvitation,
) *listing.Invitation {
	return &listing.Invitation{
		ManagerInvitation:  VersionInvitationStatus(&m.ManagerInvitation),
		LandlordInvitation: VersionInvitationStatus(&m.LandlordInvitation),
	}
}

// VersionInvitationStatus renderer
func VersionInvitationStatus(
	m *model.VersionInvitationStatus,
) *listing.InvitationStatus {
	res := &listing.InvitationStatus{
		Pending:      m.Pending,
		Sent:         timestamppb.New(time.Time(m.Sent)),
		HostId:       m.HostID,
		InviteeEmail: m.InviteeEmail,
		InviteeId:    m.InviteeID,
		InviteePhone: m.InviteePhone,
		InviteeName:  m.InviteeName,
		InvitationId: m.InvitationID,
	}

	return res
}

// Rental renderer
func Rental(
	m *model.VersionRental,
	uncensored bool,
) *listing.Rental {
	emv := time.Time(m.EarliestMovein)

	res := &listing.Rental{
		RentalPrice:       m.RentalPrice.Proto(),
		DssAvailable:      m.DssAvailable.Proto(),
		BillsIncluded:     m.BillsIncluded.Proto(),
		BillsCovered:      &listing.PropertyBillTypeArrayValue{Value: m.BillsCovered},
		DurationMinMonths: m.DurationMinMonths.Proto(),
		DurationMaxMonths: m.DurationMaxMonths.Proto(),
		EarliestMovein: &generic.Date{
			Year:  int32(emv.Year()),
			Month: int32(emv.Month()),
			Day:   int32(emv.Day()),
		},
		TenancyDepositInsurance: m.TenancyDepositInsurance.Proto(),
		InventoryReport:         m.InventoryReport.Proto(),
		DepositReplacementFee:   m.DepositReplacementFee.Proto(),
	}

	if uncensored {
		res.BankDetails = &listing.BankDetails{
			Society:           m.BankDetails.Society,
			SortCode:          m.BankDetails.SortCode,
			AccountNumber:     m.BankDetails.AccountNumber,
			AccountName:       m.BankDetails.AccountName,
			PersonalReference: m.BankDetails.PersonalReference,
		}
	}

	return res
}

// Location renderer
func Location(
	m *model.VersionLocation,
	uncensored bool,
) *listing.Location {
	res := &listing.Location{
		Postcode:             wrapperspb.String(m.Postcode),
		Address1:             wrapperspb.String(m.Address1),
		Address2:             wrapperspb.String(m.Address2),
		Address3:             wrapperspb.String(m.Address3),
		Address4:             wrapperspb.String(m.Address4),
		Flat:                 wrapperspb.String(m.Flat),
		Building:             wrapperspb.String(m.Building),
		Street:               wrapperspb.String(m.Street),
		City:                 wrapperspb.String(m.City),
		Country:              wrapperspb.String(m.Country),
		PostTown:             wrapperspb.String(m.PostTown),
		County:               wrapperspb.String(m.County),
		AdministrativeCounty: wrapperspb.String(m.AdministrativeCounty),
		PostalCounty:         wrapperspb.String(m.PostalCounty),
		TraditionalCounty:    wrapperspb.String(m.TraditionalCounty),
		District:             wrapperspb.String(m.District),
		Address:              wrapperspb.String(m.Address),
		GeoRegion:            wrapperspb.String(m.GeoRegion),
		GeoCounty:            wrapperspb.String(m.GeoCounty),
		GeoTown:              wrapperspb.String(m.GeoTown),
	}

	if !uncensored {
		res.Flat = nil
		res.Building = nil

		if len(res.GetPostcode().GetValue()) > 3 {
			res.Postcode = wrapperspb.String(
				strings.TrimSpace(res.GetPostcode().GetValue()[:len(res.GetPostcode().GetValue())-3]),
			)
		}
	}

	if m.GPS.Latitude != 0 && m.GPS.Longitude != 0 {
		res.Gps = &listing.GpsCoordinates{
			Longitude: m.GPS.Longitude,
			Latitude:  m.GPS.Latitude,
		}
	}

	return res
}

// Valuation renderer
func Valuation(
	m *model.VersionValuation,
) *listing.Valuation {
	return &listing.Valuation{
		Estimate:   m.Estimate,
		Confidence: m.Confidence,
		Additional: &listing.ValuationAdditional{
			VariationConfidence:        m.Additional.VariationConfidence,
			ComparatorNumberConfidence: m.Additional.ComparatorNumberConfidence,
		},
		LowerBound: m.LowerBound,
		UpperBound: m.UpperBound,
	}
}

// PublicationHistory renderer
func PublicationHistory(
	entries ...model.PublicationEntries,
) (list []*listing.PublicationHistoryItem) {
	for _, es := range entries {
		for _, e := range es {
			var published, unpbulished *timestamppb.Timestamp

			if time.Time(e.PublishedAt).Unix() > 0 {
				published = timestamppb.New(time.Time(e.PublishedAt))
			}

			if time.Time(e.UnpublishedAt).Unix() > 0 {
				unpbulished = timestamppb.New(time.Time(e.UnpublishedAt))
			}

			list = append(list, &listing.PublicationHistoryItem{
				Platform:      e.Platform,
				Id:            e.PublicationID,
				Url:           e.PublicationURL,
				PublishedBy:   e.PublishedBy,
				UnpublishedBy: e.UnpublishedBy,
				PublishedAt:   published,
				UnpublishedAt: unpbulished,
				InvoiceIds:    e.InvoiceIDs,
			})
		}
	}

	return
}

// Publication renderer
func Publication(
	m *model.Version,
) *listing.Publication {
	return &listing.Publication{
		HistoryItems: PublicationHistory(
			m.PublicationsMashroom,
			m.PublicationsRightmove,
			m.PublicationsBoomin,
			m.PublicationsZoopla,
		),
		Mashroom:  PublicationItem(m.PublicationsMashroom),
		Zoopla:    PublicationItem(m.PublicationsZoopla),
		Rightmove: PublicationItem(m.PublicationsRightmove),
		Boomin:    PublicationItem(m.PublicationsBoomin),
	}
}

// PublicationItem renderer
func PublicationItem(m model.PublicationEntries) *listing.PublicationItem {
	if !m.Published() {
		return nil
	}

	last := m.Last()

	var published *timestamppb.Timestamp

	if time.Time(last.PublishedAt).Unix() > 0 {
		published = timestamppb.New(time.Time(last.PublishedAt))
	}

	return &listing.PublicationItem{
		Id:          last.PublicationID,
		Url:         last.PublicationURL,
		PublishedAt: published,
		PublishedBy: last.PublishedBy,
		InvoiceIds:  last.InvoiceIDs,
	}
}

// Verification renderer
func Verification(
	m *model.VersionVerification,
) *listing.Verification {
	return &listing.Verification{
		Address:     AdminVerifyItem(m.Address),
		Description: AdminVerifyItem(m.Description),
		Photos:      AdminVerifyItem(m.Photos),
		Documents:   AdminVerifyItem(m.Documents),
		Fee:         AdminVerifyItem(m.Fee),
	}
}

// AdminVerifyItem renderer
func AdminVerifyItem(m model.AdminVerifyItem) *listing.AdminVerifyItem {
	return &listing.AdminVerifyItem{
		Admin:      m.Admin,
		Verified:   m.Verified,
		VerifiedAt: timestamppb.New(time.Time(m.VerifiedAt)),
	}
}
