package service

import (
	"context"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/entities"
	"github.com/camunda-cloud/zeebe/clients/go/pkg/worker"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bpms"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/process"
)

// UploadProcesses uploads BPM processes if it's client is available
func (srv *serviceImpl) UploadProcesses() error {
	if srv.ZeebeClient == nil {
		return nil
	}

	for processID, file := range process.Processes {
		log.Infof("Deploying process %s from %s", processID, file)

		listBs, err := process.FS.ReadFile(file)
		if err != nil {
			return err
		}

		bpms.Deploy(srv.Config.ZeebeClient, listBs, processID)
	}

	return nil
}

// ProcessSetManager sets most recent version manager
func (srv *serviceImpl) ProcessSetManager() (request interface{}, processor bpms.Processor) {
	req := &struct {
		User     string `json:"user"`
		Property string `json:"property"`
	}{}

	return req, func(
		ctx context.Context,
		job entities.Job,
	) (reply interface{}, enhance map[string]interface{}, err error) {
		grpcctx.WithServiceUser(ctx)

		ver, err := srv.Repo.VersionGet(
			ctx,
			model.VersionSpec(ctx, listing.VersionSpecPropertyID("", req.Property), false),
		)
		if err != nil {
			return nil, nil, err
		}

		ver.People.Manager = req.User
		ver.Invitation.ManagerInvitation = model.VersionInvitationStatus{}

		srv.validateVersion(ctx, ver)

		err = srv.Repo.VersionUpdate(ctx, ver, ver.Status)
		if err != nil {
			return nil, nil, err
		}

		return nil, nil, err
	}
}

// ProcessSetLandlord sets most recent version landlord
func (srv *serviceImpl) ProcessSetLandlord() (request interface{}, processor bpms.Processor) {
	req := &struct {
		User     string `json:"user"`
		Property string `json:"property"`
	}{}

	return req, func(
		ctx context.Context,
		job entities.Job,
	) (reply interface{}, enhance map[string]interface{}, err error) {
		ctx = grpcctx.WithServiceUser(ctx)

		ver, err := srv.Repo.VersionGet(
			ctx,
			model.VersionSpec(ctx, listing.VersionSpecPropertyID("", req.Property), false),
		)
		if err != nil {
			return nil, nil, err
		}

		ver.People.Landlord = req.User
		ver.Invitation.LandlordInvitation = model.VersionInvitationStatus{}

		srv.validateVersion(ctx, ver)

		err = srv.Repo.VersionUpdate(ctx, ver, ver.Status)
		if err != nil {
			return nil, nil, err
		}

		return nil, nil, err
	}
}

// ProcessSetStatus sets most recent version status
func (srv *serviceImpl) ProcessSetStatus() (request interface{}, processor bpms.Processor) {
	req := &struct {
		Property string                `json:"property"`
		Status   listing.VersionStatus `json:"status"`
	}{}

	return req, func(
		ctx context.Context,
		job entities.Job,
	) (reply interface{}, enhance map[string]interface{}, err error) {
		ctx = grpcctx.WithServiceUser(ctx)

		ver, err := srv.Repo.VersionGet(
			ctx,
			model.VersionSpec(ctx, listing.VersionSpecPropertyID("", req.Property), false),
		)
		if err != nil {
			return nil, nil, err
		}

		err = srv.Repo.VersionUpdateStatus(ctx, ver.ID, ver.Status, req.Status)
		if err != nil {
			return nil, nil, err
		}

		return nil, nil, err
	}
}

// ProcessSetActiveOffer sets most recent version status
func (srv *serviceImpl) ProcessSetActiveOffer() (request interface{}, processor bpms.Processor) {
	req := &struct {
		Property string `json:"property"`
		Offer    string `json:"offer"`
	}{}

	return req, func(
		ctx context.Context,
		job entities.Job,
	) (reply interface{}, enhance map[string]interface{}, err error) {
		ctx = grpcctx.WithServiceUser(ctx)

		ver, err := srv.Repo.VersionGet(
			ctx,
			model.VersionSpec(ctx, listing.VersionSpecPropertyID("", req.Property), false),
		)
		if err != nil {
			return nil, nil, err
		}

		ver.Flags.ActiveOffer = req.Offer

		err = srv.Repo.VersionUpdate(ctx, ver, ver.Status)
		if err != nil {
			return nil, nil, err
		}

		return nil, nil, err
	}
}

// RegisterProcessHandlers sets up BPM process handlers, if BPM is available and returns BPM client, in case it is
func (srv *serviceImpl) RegisterProcessHandlers() (worker.JobWorker, error) {
	if srv.ZeebeClient == nil {
		return nil, nil
	}

	handler := bpms.MapHandler(map[string]bpms.Registrator{
		"set-manager":      srv.ProcessSetManager,
		"set-landlord":     srv.ProcessSetLandlord,
		"set-status":       srv.ProcessSetStatus,
		"set-active-offer": srv.ProcessSetActiveOffer,
	})

	return bpms.RegisterProcessHandler(srv.ZeebeClient, "listing-service", handler)
}
