package service

import (
	"context"
	"errors"
	"strconv"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

func (srv *serviceImpl) VersionInvitationCreate(
	ctx context.Context,
	req *listing.VersionInvitationCreateRequest,
) (*listing.VersionInvitationCreateResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetHostId())

	var invitee *account.Account

	switch {
	case req.GetInvitee().GetUserId() != "":
		var resp *account.GetResponse

		resp, err = srv.Connections.Account.Get(ctx, &account.GetRequest{
			Id: req.GetInvitee().GetUserId(),
		})
		if err != nil {
			return nil, err
		}

		invitee = resp.GetAccount()
	case req.GetInvitee().GetEmail() != "":
		var resp *account.GetResponse

		resp, err = srv.Connections.Account.GetByEmail(ctx, &account.GetByEmailRequest{
			Email: req.GetInvitee().GetEmail(),
		})
		if err != nil && !errors.Is(err, grpcerr.NotFound) {
			return nil, err
		}

		invitee = resp.GetAccount()
	}

	var invtype invitation.InvitationType

	var verinv *model.VersionInvitationStatus

	switch {
	case req.GetLandlordInvitation() != nil:
		if !rel.landlord && !rel.manager && !rel.submitter {
			return nil, errAccessDenied
		}

		invtype = invitation.InvitationType_INVITATION_TYPE_RELATED_TO_LANDLORD

		if lv.People.Landlord == "" {
			lv.People.Landlord = invitee.GetIdValue()
		}

		if lv.People.Manager == "" {
			lv.People.Manager = invitee.GetIdValue()
		}

		verinv = &lv.Invitation.LandlordInvitation
	case req.GetManagerInvitation() != nil:
		if !rel.landlord {
			return nil, errAccessDenied
		}

		invtype = invitation.InvitationType_INVITATION_TYPE_LANDLORD_TO_MANAGER

		verinv = &lv.Invitation.ManagerInvitation

		if invitee.GetIdValue() != "" && lv.People.Manager == "" {
			lv.People.Manager = invitee.GetIdValue()
		} else if lv.People.Manager == "" {
			lv.People.Manager = lv.People.Landlord
		}
	default:
		return nil, errValidationErrors
	}

	fillInvitationUserContactDetails(req.GetInvitee(), invitee)

	invresp, err := srv.Connections.Invitation.InvitationCreate(ctx, &invitation.InvitationCreateRequest{
		HostId: req.GetHostId(),
		Invitee: &invitation.User{
			UserId: req.GetInvitee().GetUserId(),
			Email:  req.GetInvitee().GetEmail(),
			Phone:  req.GetInvitee().GetPhone(),
			Name:   req.GetInvitee().GetName(),
		},
		Property: &invitation.Property{
			PropertyId: lv.PropertyID,
			VersionId:  strconv.FormatUint(lv.ID, 10),
		},
		Type:           invtype,
		Offer:          "",
		NotifyCallback: srv.Connections.SelfRemote,
		Message:        req.GetMessage(),
	})
	if err != nil {
		return nil, err
	}

	verinv.InvitationID = invresp.GetInvitation().GetId()
	verinv.Pending = true
	verinv.Sent = model.ServiceTime(invresp.GetInvitation().GetCreatedAt().AsTime())
	verinv.HostID = req.GetHostId()
	verinv.InviteeID = req.GetInvitee().GetUserId()
	verinv.InviteeEmail = req.GetInvitee().GetEmail()
	verinv.InviteePhone = req.GetInvitee().GetPhone()
	verinv.InviteeName = req.GetInvitee().GetName()

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &listing.VersionInvitationCreateResponse{
		InvitationId: invresp.GetInvitation().GetId(),
	}, nil
}

func fillInvitationUserContactDetails(user *listing.InvitationUser, acc *account.Account) {
	if acc == nil {
		return
	}

	if user.GetUserId() == "" {
		user.UserId = acc.GetIdValue()
	}

	if user.GetEmail() == "" {
		user.Email = acc.GetContactValue().GetEmailPrimaryValue()
	}

	if user.GetPhone() == "" {
		user.Phone = acc.GetContactValue().GetPhonePrimaryValue()
	}

	if user.GetName() == "" {
		user.Name = util.JoinNonblankStrings(
			" ",
			acc.GetDetailsValue().GetFirstNameValue(),
			acc.GetDetailsValue().GetLastNameValue(),
		)
	}
}

func (srv *serviceImpl) InvitationChange(
	ctx context.Context,
	req *invitation.InvitationChangeRequest,
) (*invitation.InvitationChangeResponse, error) {
	spec := listing.VersionSpecPropertyID("", req.GetInvitation().GetProperty().GetPropertyId())

	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, spec, false))
	if err != nil {
		return nil, err
	}

	switch req.GetInvitation().GetStatus() {
	case invitation.InvitationStatus_INVITATION_STATUS_ACCEPTED:
		switch req.GetInvitation().GetType() {
		case invitation.InvitationType_INVITATION_TYPE_RELATED_TO_LANDLORD:
			lv.Invitation.LandlordInvitation.Pending = false
			lv.Invitation.LandlordInvitation.InviteeID = req.GetInvitation().GetInvitee().GetUserId()
			lv.People.Landlord = req.GetInvitation().GetInvitee().GetUserId()
		case invitation.InvitationType_INVITATION_TYPE_LANDLORD_TO_MANAGER:
			lv.Invitation.ManagerInvitation.Pending = false
			lv.Invitation.ManagerInvitation.InviteeID = req.GetInvitation().GetInvitee().GetUserId()
			lv.People.Manager = req.GetInvitation().GetInvitee().GetUserId()

			if !lv.Published.After(time.Unix(0, 0)) {
				_, err = srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
					Entity: notification.Entity_ENTITY_PROPERTY_INVITATION_MANAGER_UNLISTED,
					Action: notification.Action_ACTION_ACCEPT,
					Actor: &notification.Actor{
						Role: notification.Role_ROLE_TENANT,
						User: req.GetInvitation().GetInvitee().GetUserId(),
					},
					Targets: []*notification.Target{
						{
							User: req.GetInvitation().GetHostId(),
							Role: notification.Role_ROLE_LANDLORD,
						},
					},
					Embeds: []*notification.EventEmbed{
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_INVITATION,
							EmbedValue: req.GetInvitation().GetId(),
						},
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
							EmbedValue: req.GetInvitation().GetProperty().GetPropertyId(),
						},
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
							EmbedValue: req.GetInvitation().GetHostId(),
						},
					},
				})
				if err != nil {
					log.WithError(err).Error("sending manager accepted notification")
				}
			}
		}
	case invitation.InvitationStatus_INVITATION_STATUS_DECLINED,
		invitation.InvitationStatus_INVITATION_STATUS_CANCELED,
		invitation.InvitationStatus_INVITATION_STATUS_TIMEOUT:
		switch req.GetInvitation().GetType() {
		case invitation.InvitationType_INVITATION_TYPE_RELATED_TO_LANDLORD:
			lv.Invitation.LandlordInvitation.Pending = false
		case invitation.InvitationType_INVITATION_TYPE_LANDLORD_TO_MANAGER:
			lv.Invitation.ManagerInvitation.Pending = false
			if lv.People.Manager == lv.Invitation.ManagerInvitation.InviteeID {
				lv.People.Manager = lv.People.Landlord
			}
		}
	default:
		return &invitation.InvitationChangeResponse{}, nil
	}

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationChangeResponse{}, nil
}
