package service

import (
	"context"
	"encoding/json"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

func (srv *serviceImpl) PrefaceGet(
	ctx context.Context,
	req *listing.PrefaceGetRequest,
) (*listing.PrefaceGetResponse, error) {
	lv, err := srv.Repo.PrefaceGet(ctx, req.GetClientId())
	if err != nil {
		return nil, err
	}

	if lv.ID == 0 {
		lv.Validation = srv.validate(ctx, &lv.Location, &lv.Details, &lv.Rental, nil, &lv.Description, "")
	}

	return &listing.PrefaceGetResponse{
		Preface: srv.renderPreface(lv),
	}, nil
}

func (srv *serviceImpl) PrefaceSave(
	ctx context.Context,
	req *listing.PrefaceSaveRequest,
) (*listing.PrefaceSaveResponse, error) {
	lv := &model.Preface{
		ClientID:    req.GetClientId(),
		Description: model.VersionDescription{},
		Details:     model.VersionDetails{},
		Rental:      model.VersionRental{},
		Location:    model.VersionLocation{},
		Valuation:   model.VersionValuation{},
		Validation:  model.VersionValidation{},
	}

	srv.updateDetails(&lv.Details, req.GetDetails())
	srv.updateDescription(&lv.Description, req.GetDescription(), "")
	srv.updateRental(&lv.Rental, req.GetRental())

	err := srv.updateLocation(ctx, &lv.Location, req.GetLocation(), &lv.Timezone)
	if err != nil {
		return nil, err
	}

	lv.Valuation, _, err = srv.updateValuation(ctx, &lv.Location, &lv.Details)
	if err != nil {
		log.WithError(err).Error("updating valuation")
	}

	err = srv.Repo.PrefaceSave(ctx, lv)
	if err != nil {
		return nil, err
	}

	lv.Validation = srv.validate(ctx, &lv.Location, &lv.Details, &lv.Rental, nil, &lv.Description, "")

	err = srv.Repo.PrefaceSave(ctx, lv)
	if err != nil {
		return nil, err
	}

	return &listing.PrefaceSaveResponse{
		Preface: srv.renderPreface(lv),
	}, nil
}

func (srv *serviceImpl) prefaceSlackNotification(
	ctx context.Context,
	lv *model.Version,
) error {
	if lv == nil {
		return nil
	}

	depositType := "Deposit Replacement Fee"
	if lv.Rental.TenancyDepositInsurance.GetValue() {
		depositType = "Traditional 5 week deposit"
	}

	moveinDate := time.Time(lv.Rental.EarliestMovein)
	moveInString := moveinDate.Format("Mon, 2 Jan 2006")

	extras, err := json.Marshal(struct {
		DepositType string `json:"DepositType"`
		MoveInDate  string `json:"MoveInDate"`
	}{
		DepositType: depositType,
		MoveInDate:  moveInString,
	})
	if err != nil {
		return err
	}

	_, err = srv.Connections.Notification.SendNotification(
		grpcctx.WithServiceUser(ctx),
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_PROPERTY_PREFACE,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: lv.PropertyID,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: lv.People.Landlord,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(extras),
				},
			},
		},
	)

	return err
}
