package service

import (
	"context"
	"errors"
	"testing"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/generic/optional"

	"github.com/golang/mock/gomock"
	"github.com/stretchr/testify/suite"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/mock"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

type slackNotifierSuite struct {
	suite.Suite
	rootCtx      context.Context
	rootCancel   context.CancelFunc
	ctrl         *gomock.Controller
	notification *mock.MockNotificationServiceClient
	service      *serviceImpl
}

func (s *slackNotifierSuite) SetupSuite() {
	s.rootCtx, s.rootCancel = context.WithCancel(context.Background())
}

func (s *slackNotifierSuite) SetupTest() {
	s.ctrl = gomock.NewController(s.T())
	s.notification = mock.NewMockNotificationServiceClient(s.ctrl)

	s.service = &serviceImpl{
		Config: Config{
			Connections: connections.Connections{
				Notification: s.notification,
			},
		},
	}
}

func (s *slackNotifierSuite) TearDownTest() {
	s.ctrl.Finish()
}

func (s *slackNotifierSuite) TearSuite() {
	s.rootCancel()
}

func TestSlackNotifierSuite(t *testing.T) {
	suite.Run(t, &slackNotifierSuite{})
}

func (s *slackNotifierSuite) TestSlackNotifierSuite_ok_traditional() {
	s.notification.EXPECT().SendNotification(
		grpcctx.WithServiceUser(s.rootCtx),
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_PROPERTY_PREFACE,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: "123",
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: "345",
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: `{"DepositType":"Traditional 5 week deposit","MoveInDate":"Fri, 1 Dec 2000"}`,
				},
			},
		},
	).Return(&notification.SendNotificationResponse{}, nil)

	d := time.Date(2000, 12, 1, 0, 0, 0, 0, time.UTC)
	err := s.service.prefaceSlackNotification(s.rootCtx, &model.Version{
		ID:         123,
		PropertyID: "123",
		People: model.VersionPeople{
			Landlord: "345",
		},
		Rental: model.VersionRental{
			EarliestMovein:          model.ServiceTime(d),
			TenancyDepositInsurance: optional.NewBool(true),
		},
	})

	s.Require().NoError(err)
}

func (s *slackNotifierSuite) TestSlackNotifierSuite_ok_replacement() {
	s.notification.EXPECT().SendNotification(
		grpcctx.WithServiceUser(s.rootCtx),
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_PROPERTY_PREFACE,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: "123",
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: "345",
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: `{"DepositType":"Deposit Replacement Fee","MoveInDate":"Fri, 1 Dec 2000"}`,
				},
			},
		},
	).Return(&notification.SendNotificationResponse{}, nil)

	d := time.Date(2000, 12, 1, 0, 0, 0, 0, time.UTC)
	err := s.service.prefaceSlackNotification(s.rootCtx, &model.Version{
		ID:         123,
		PropertyID: "123",
		People: model.VersionPeople{
			Landlord: "345",
		},
		Rental: model.VersionRental{
			EarliestMovein:          model.ServiceTime(d),
			TenancyDepositInsurance: optional.NewBool(false),
		},
	})

	s.Require().NoError(err)
}

func (s *slackNotifierSuite) TestSlackNotifierSuite_error() {
	s.notification.EXPECT().SendNotification(
		grpcctx.WithServiceUser(s.rootCtx),
		gomock.Any(),
	).Return(nil, errors.New("some error"))

	d := time.Date(2000, 12, 1, 0, 0, 0, 0, time.UTC)
	err := s.service.prefaceSlackNotification(s.rootCtx, &model.Version{
		ID:         123,
		PropertyID: "123",
		People: model.VersionPeople{
			Landlord: "345",
		},
		Rental: model.VersionRental{
			EarliestMovein:          model.ServiceTime(d),
			TenancyDepositInsurance: optional.NewBool(false),
		},
	})

	s.Require().Error(err)
}
