package service

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"sort"
	"strconv"
	"strings"
	"time"

	cardpayment "gitlab.mashroom.com/mashroom/backend/cardpayment/api"
	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	geo "gitlab.mashroom.com/mashroom/backend/geo/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service/publication"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service/render"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	viewing "gitlab.mashroom.com/mashroom/backend/viewing/api"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

func stringSliceContains(arr []string, el string) bool {
	for _, a := range arr {
		if a == el {
			return true
		}
	}

	return false
}

type userRelation struct {
	landlord  bool
	submitter bool
	manager   bool
	tenant    bool
	admin     bool
	superuser bool
	related   bool
	editor    bool
}

func (srv *serviceImpl) userRelation(lv *model.Version, user string) (rel userRelation) {
	rel.landlord = lv.People.Landlord == user
	rel.submitter = lv.People.Submitter == user
	rel.manager = lv.People.Manager == user
	rel.tenant = stringSliceContains(lv.People.Tenants, user)
	rel.admin = lv.People.Admin == user
	rel.superuser = user == "" || rel.admin
	rel.related = rel.landlord || rel.submitter || rel.manager || rel.tenant || rel.admin
	rel.editor = rel.superuser || rel.landlord || (rel.submitter && lv.People.Landlord == "")

	return rel
}

func (srv *serviceImpl) versionCreate(
	ctx context.Context,
	lv *model.Version,
	clientID string,
	details *listing.Details,
	description *listing.Description,
	rental *listing.Rental,
	people *listing.People,
	location *listing.NewLocation,
	viewavail *listing.ViewingAvailability,
	createfunc func(newstatus listing.VersionStatus) error,
) (err error) {
	newProperty := lv.ID == 0

	newstatus := versionCreatePreprocess(
		lv,
		description,
		rental,
	)

	oldgps := lv.Location.GPS

	if clientID != "" {
		err = srv.commitPreface(ctx, lv, clientID)
		if err != nil {
			return err
		}

		newProperty = true
	}

	srv.updateDetails(&lv.Details, details)
	srv.updateDescription(&lv.Description, description, lv.IVR)
	srv.updateRental(&lv.Rental, rental)
	srv.updatePeople(&lv.People, people)
	srv.updateViewingAvailability(&lv.ViewingAvailability, viewavail)

	err = srv.updateLocation(ctx, &lv.Location, location, &lv.Timezone)
	if err != nil {
		return err
	}

	lv.Valuation, _, err = srv.updateValuation(ctx, &lv.Location, &lv.Details)
	if err != nil {
		log.WithError(err).Error("updating valuation")
	}

	srv.validateVersion(ctx, lv)

	err = createfunc(newstatus)
	if err != nil {
		return err
	}

	if oldgps != lv.Location.GPS {
		_, err = srv.refreshNearbyCache(ctx, lv.ID, lv.Location.GPS.Longitude, lv.Location.GPS.Latitude)
		if err != nil {
			return err
		}
	}

	if clientID != "" {
		err = srv.Repo.PrefaceRemove(ctx, clientID)
		if err != nil {
			return err
		}

		err = srv.prefaceSlackNotification(ctx, lv)
		if err != nil {
			log.WithError(err).Infof(
				"Failed to send slack notification for %s",
				lv.PropertyID,
			)
		}
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if newProperty {
		go srv.serviceUserSync(bgctx, lv.People.Landlord)
		go srv.servicePropertyOpportunity(
			bgctx,
			lv.People.Landlord,
			lv.PropertyID,
			&lv.Location,
			&lv.Origin,
		)
	}

	return nil
}

func versionOrigin(origin *listing.VersionOrigin) model.VersionOrigin {
	var (
		product *model.VersionOriginProduct
		tool    *model.VersionOriginTool
	)

	if origin.GetTool() != nil {
		tool = &model.VersionOriginTool{
			Valuation:             origin.GetTool().GetValuation(),
			ExpenseTracker:        origin.GetTool().GetExpenseTracker(),
			CertificateRepository: origin.GetTool().GetCertificateRepository(),
		}
	}

	if origin.GetProduct() != nil {
		product = &model.VersionOriginProduct{
			ProductID:   origin.GetProduct().GetProductId(),
			ProductName: origin.GetProduct().GetProductName(),
		}
	}

	return model.VersionOrigin{
		Listing: origin.GetListing(),
		Preface: origin.GetPreface(),
		Product: product,
		Tool:    tool,
	}
}

func (srv *serviceImpl) VersionCreate(
	ctx context.Context,
	req *listing.VersionCreateRequest,
) (*listing.VersionCreateResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, err
	}

	if lv == nil {
		landlord := req.GetPeople().GetLandlord().GetValue()
		if landlord == "" {
			landlord = req.GetSpec().GetUserId()
		}

		manager := req.GetPeople().GetManager().GetValue()
		if manager == "" {
			manager = landlord
		}

		var ivr string

		prop := req.GetSpec().GetSelect().GetPropertyId()
		if prop == "" {
			resp, perr := srv.Connections.Property.PropertyCreate(ctx, &property.PropertyCreateRequest{
				User: req.GetSpec().GetUserId(),
			})
			if perr != nil {
				return nil, perr
			}

			prop = resp.GetProperty().GetId()
			ivr = resp.GetProperty().GetIvr()
		} else {
			resp, perr := srv.Connections.Property.PropertyGet(ctx, &property.PropertyGetRequest{
				PropertyId: prop,
			})
			if perr != nil {
				return nil, err
			}

			ivr = resp.GetProperty().GetIvr()
		}

		lv = &model.Version{
			PropertyID:  prop,
			IVR:         ivr,
			Description: model.VersionDescription{},
			Details:     model.VersionDetails{},
			People: model.VersionPeople{
				Landlord:  landlord,
				Manager:   manager,
				Submitter: req.GetSpec().GetUserId(),
			},
			Rental:              model.VersionRental{},
			Location:            model.VersionLocation{},
			Valuation:           model.VersionValuation{},
			Verification:        model.VersionVerification{},
			Validation:          model.VersionValidation{},
			ViewingAvailability: defaultViewingAvailability(),
			Origin:              versionOrigin(req.GetOrigin()),
			Invitation:          model.VersionInvitation{},
			Status:              listing.VersionStatus_VERSION_DRAFT,
			Flags:               model.VersionFlags{},
		}
	}

	rel := srv.userRelation(lv, req.GetSpec().GetUserId())

	if !rel.editor {
		return nil, errAccessDenied
	}

	err = srv.versionCreate(
		ctx,
		lv,
		req.GetSpec().GetClientId(),
		req.GetDetails(),
		req.GetDescription(),
		req.GetRental(),
		req.GetPeople(),
		req.GetLocation(),
		req.GetViewingAvailability(),
		func(newstatus listing.VersionStatus) error {
			err = srv.Repo.VersionCreate(ctx, lv, newstatus)
			if err != nil {
				if lv.ID == 0 {
					return fmt.Errorf("creating property: %w", err)
				}

				lv.ID = 0

				err = srv.Repo.VersionCreate(ctx, lv, newstatus)
				if err != nil {
					return fmt.Errorf("creating property: %w", err)
				}
			}

			return nil
		},
	)
	if err != nil {
		return nil, err
	}

	return &listing.VersionCreateResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func defaultViewingAvailability() (ms model.VersionViewingAvailability) {
	for i := 0; i < 7; i++ {
		ms = append(ms, &model.VersionViewingAvailabilitySlot{
			Period:      listing.Period_PERIOD_WEEK,
			StartOffset: time.Duration(24*i+6) * time.Hour,
			EndOffset:   time.Duration(24*i+22) * time.Hour,
		})
	}

	return
}

func (srv *serviceImpl) versionUpdate(
	ctx context.Context,
	lv *model.Version,
	details *listing.Details,
	description *listing.Description,
	rental *listing.Rental,
	people *listing.People,
	location *listing.NewLocation,
	viewavail *listing.ViewingAvailability,
	updatefunc func() error,
) (err error) {
	oldgps := lv.Location.GPS

	srv.updateDetails(&lv.Details, details)
	srv.updateDescription(&lv.Description, description, lv.IVR)
	srv.updateRental(&lv.Rental, rental)
	srv.updatePeople(&lv.People, people)
	srv.updateViewingAvailability(&lv.ViewingAvailability, viewavail)

	err = srv.updateLocation(ctx, &lv.Location, location, &lv.Timezone)
	if err != nil {
		return err
	}

	lv.Valuation, _, err = srv.updateValuation(ctx, &lv.Location, &lv.Details)
	if err != nil {
		log.WithError(err).Error("updating valuation")
	}

	srv.validateVersion(ctx, lv)

	err = updatefunc()
	if err != nil {
		return err
	}

	if oldgps != lv.Location.GPS {
		_, err = srv.refreshNearbyCache(ctx, lv.ID, lv.Location.GPS.Longitude, lv.Location.GPS.Latitude)
		if err != nil {
			return err
		}
	}

	return nil
}

func (srv *serviceImpl) VersionUpdate(
	ctx context.Context,
	req *listing.VersionUpdateRequest,
) (*listing.VersionUpdateResponse, error) {
	getopts := model.VersionSpec(ctx, req.GetSpec(), false)
	getopts.Statuses = []listing.VersionStatus{
		listing.VersionStatus_VERSION_DRAFT,
	}

	lv, err := srv.Repo.VersionGet(ctx, getopts)
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetSpec().GetUserId())

	if !rel.editor {
		return nil, errAccessDenied
	}

	err = srv.versionUpdate(
		ctx,
		lv,
		req.GetDetails(),
		req.GetDescription(),
		req.GetRental(),
		req.GetPeople(),
		req.GetLocation(),
		nil,
		func() error {
			return srv.Repo.VersionUpdate(ctx, lv, lv.Status)
		},
	)
	if err != nil {
		return nil, err
	}

	return &listing.VersionUpdateResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionValidate(
	ctx context.Context,
	req *listing.VersionValidateRequest,
) (*listing.VersionValidateResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
		UserID:    req.GetUserId(),
	})
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetUserId())

	if !rel.editor {
		return nil, errAccessDenied
	}

	srv.validateVersion(ctx, lv)

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &listing.VersionValidateResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionGet(
	ctx context.Context,
	req *listing.VersionGetRequest,
) (resp *listing.VersionGetResponse, err error) {
	if strings.HasPrefix(req.GetSpec().GetSelect().GetPropertyId(), "transient:") {
		var lv *model.Preface

		lv, err = srv.Repo.PrefaceGet(ctx, req.GetSpec().GetSelect().GetPropertyId())
		if err != nil {
			return nil, err
		}

		if lv.ID == 0 {
			lv.Validation = srv.validate(ctx, &lv.Location, &lv.Details, &lv.Rental, nil, &lv.Description, "")
		}

		preface := srv.renderPreface(lv)

		return &listing.VersionGetResponse{
			Version: &listing.Version{
				Id:          preface.GetClientId(),
				ListingId:   preface.GetClientId(),
				PropertyId:  preface.GetClientId(),
				TenancyId:   "",
				Status:      listing.VersionStatus_VERSION_DRAFT,
				CreatedAt:   preface.GetCreatedAt(),
				UpdatedAt:   preface.GetUpdatedAt(),
				Description: preface.GetDescription(),
				Details:     preface.GetDetails(),
				People: &listing.People{
					Submitter: wrapperspb.String(preface.GetClientId()),
				},
				Rental:     preface.GetRental(),
				Location:   preface.GetLocation(),
				Valuation:  preface.GetValuation(),
				Validation: preface.GetValidation(),
			},
		}, nil
	}

	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	return &listing.VersionGetResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionList(
	ctx context.Context,
	req *listing.VersionListRequest,
) (*listing.VersionListResponse, error) {
	lvs, err := srv.Repo.VersionsList(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	return &listing.VersionListResponse{
		Versions: srv.renderVersions(ctx, lvs, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionPropertyIDFilter(
	ctx context.Context,
	req *listing.VersionPropertyIDFilterRequest,
) (*listing.VersionPropertyIDFilterResponse, error) {
	var radius *model.SearchInRadius

	if req.GetInRadius() != nil {
		radius = &model.SearchInRadius{
			Longitude:      req.GetInRadius().GetCenter().GetLongitude(),
			Latitude:       req.GetInRadius().GetCenter().GetLatitude(),
			RadiusInMeters: req.GetInRadius().GetMeters(),
		}
	}

	req.Params.Pagination = &query.QueryParamsPagination{
		Page:    0,
		PerPage: 1000,
	}

	if !grpcctx.ContextIsAdminUser(ctx) && !grpcctx.ContextIsServiceUser(ctx) {
		req.Params.Filter = append(req.Params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"flags.disabled"},
			Value:    []string{"true"},
			Not:      true,
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	resp, _, err := srv.Repo.VersionFilterList(ctx, req.Params, radius)
	if err != nil {
		return nil, err
	}

	var ids []string

	for _, r := range resp {
		ids = append(ids, r.PropertyID)
	}

	return &listing.VersionPropertyIDFilterResponse{
		PropertyIds: ids,
	}, nil
}

func (srv *serviceImpl) VersionFilter(
	ctx context.Context,
	req *listing.VersionFilterRequest,
) (*listing.VersionFilterResponse, error) {
	var radius *model.SearchInRadius

	if req.GetInRadius() != nil {
		radius = &model.SearchInRadius{
			Longitude:      req.GetInRadius().GetCenter().GetLongitude(),
			Latitude:       req.GetInRadius().GetCenter().GetLatitude(),
			RadiusInMeters: req.GetInRadius().GetMeters(),
		}
	}

	if !grpcctx.ContextIsAdminUser(ctx) && !grpcctx.ContextIsServiceUser(ctx) {
		req.Params.Filter = append(req.Params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"flags.disabled"},
			Value:    []string{"true"},
			Not:      true,
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	resp, count, err := srv.Repo.VersionFilterList(ctx, req.Params, radius)
	if err != nil {
		return nil, err
	}

	return &listing.VersionFilterResponse{
		Versions: srv.renderVersions(ctx, resp, req.GetUserId()),
		Total:    int32(count),
	}, nil
}

func (srv *serviceImpl) updateVersionVerification(
	item *model.AdminVerifyItem,
	verify *wrapperspb.BoolValue,
	admin string,
) {
	if verify == nil {
		return
	}

	item.Verified = verify.GetValue()

	if verify.GetValue() {
		item.Admin = admin
		item.VerifiedAt = model.ServiceTime(time.Now())
	} else {
		item.Admin = ""
		item.VerifiedAt = model.ServiceTime{}
	}
}

func (srv *serviceImpl) VersionVerify(
	ctx context.Context,
	req *listing.VersionVerifyRequest,
) (*listing.VersionVerifyResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
		Statuses: []listing.VersionStatus{
			listing.VersionStatus_VERSION_VERIFYING,
			listing.VersionStatus_VERSION_VERIFIED,
			listing.VersionStatus_VERSION_ARCHIVE,
			listing.VersionStatus_VERSION_LET,
			listing.VersionStatus_VERSION_WITHDRAWN,
		},
	})
	if err != nil {
		return nil, err
	}

	srv.updateVersionVerification(&lv.Verification.Documents, req.GetDocuments(), req.GetUserId())
	srv.updateVersionVerification(&lv.Verification.Description, req.GetDescription(), req.GetUserId())
	srv.updateVersionVerification(&lv.Verification.Address, req.GetAddress(), req.GetUserId())
	srv.updateVersionVerification(&lv.Verification.Fee, req.GetFee(), req.GetUserId())
	srv.updateVersionVerification(&lv.Verification.Photos, req.GetPhotos(), req.GetUserId())

	srv.validateVersion(ctx, lv)

	newstatus := lv.Status

	switch {
	case lv.Status == listing.VersionStatus_VERSION_VERIFYING && verificationComplete(&lv.Verification) &&
		lv.Validation.Valid:
		newstatus = listing.VersionStatus_VERSION_VERIFIED
	case lv.Status == listing.VersionStatus_VERSION_VERIFIED && !verificationComplete(&lv.Verification):
		newstatus = listing.VersionStatus_VERSION_VERIFYING
	}

	err = srv.Repo.VersionUpdate(ctx, lv, newstatus)
	if err != nil {
		return nil, err
	}

	return &listing.VersionVerifyResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionSubmit(
	ctx context.Context,
	req *listing.VersionSubmitRequest,
) (*listing.VersionSubmitResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	userID := req.GetUserId()

	if grpcctx.ContextIsAdminUser(ctx) || grpcctx.ContextIsServiceUser(ctx) {
		userID = ""
	}

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
		UserID:    userID,
		Statuses: []listing.VersionStatus{
			listing.VersionStatus_VERSION_DRAFT,
			listing.VersionStatus_VERSION_VERIFYING,
		},
	})
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, userID)

	if !rel.editor {
		return nil, errAccessDenied
	}

	srv.validateVersion(ctx, lv)

	if !lv.Validation.Verifiable {
		return nil, errValidationErrors
	}

	if lv.Status == listing.VersionStatus_VERSION_VERIFYING {
		return &listing.VersionSubmitResponse{
			Version: srv.renderVersion(ctx, lv, req.GetUserId()),
		}, nil
	}

	needNotify := !lv.Notification.Get(model.VersionNotificationVerificationPending)

	lv.Notification.Set(model.VersionNotificationVerificationPending, true)

	err = srv.Repo.VersionUpdate(ctx, lv, listing.VersionStatus_VERSION_VERIFYING)
	if err != nil {
		return nil, err
	}

	if needNotify {
		go srv.sendNotifyPendingValidation(grpcctx.ValuesBackgroundContext(ctx), lv)
	}

	return &listing.VersionSubmitResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func (srv *serviceImpl) sendNotifyPendingValidation(ctx context.Context, lv *model.Version) {
	_, err := srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
		Actor: &notification.Actor{
			User: lv.People.Landlord,
			Role: notification.Role_ROLE_UNKNOWN,
		},
		Entity: notification.Entity_ENTITY_PROPERTY_LISTING,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: lv.People.Landlord,
				Role: notification.Role_ROLE_LANDLORD,
			},
			{
				User: lv.People.Admin,
				Role: notification.Role_ROLE_ADMIN,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: lv.PropertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: lv.People.Landlord,
			},
		},
	})

	if err != nil {
		log.WithError(err).Error("error on sending notification for listing pending verification")
	}

	if lv.LastVerified != 0 {
		_, err = srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
			Actor: &notification.Actor{
				User: lv.People.Landlord,
				Role: notification.Role_ROLE_UNKNOWN,
			},
			Entity: notification.Entity_ENTITY_PROPERTY_LISTING_AMENDMENT,
			Action: notification.Action_ACTION_CREATE,
			Targets: []*notification.Target{
				{
					User: lv.People.Admin,
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: lv.PropertyID,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: lv.People.Landlord,
				},
			},
		})

		if err != nil {
			log.WithError(err).Error("error on sending notification for listing amendment")
		}
	}
}

func (srv *serviceImpl) VersionUnsubmit(
	ctx context.Context,
	req *listing.VersionUnsubmitRequest,
) (*listing.VersionUnsubmitResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
		UserID:    req.GetUserId(),
		Statuses: []listing.VersionStatus{
			listing.VersionStatus_VERSION_VERIFYING,
		},
	})
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetUserId())

	if !rel.editor {
		return nil, errAccessDenied
	}

	err = srv.Repo.VersionCreate(ctx, lv, listing.VersionStatus_VERSION_DRAFT)
	if errors.Is(err, sql.ErrNoRows) {
		lv.ID = 0

		err = srv.Repo.VersionCreate(ctx, lv, listing.VersionStatus_VERSION_DRAFT)
	}

	if err != nil {
		return nil, err
	}

	return &listing.VersionUnsubmitResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func verificationComplete(m *model.VersionVerification) bool {
	return m.Fee.Verified &&
		m.Documents.Verified &&
		m.Address.Verified &&
		m.Description.Verified &&
		m.Photos.Verified
}

func (srv *serviceImpl) VersionFlags(
	ctx context.Context,
	req *listing.VersionFlagsRequest,
) (*listing.VersionFlagsResponse, error) {
	var flag model.FlagType

	var value interface{}

	var err error

	switch f := req.GetFlag().(type) {
	case *listing.VersionFlagsRequest_Disabled:
		flag = model.FlagDisabled
		value = f.Disabled

		if f.Disabled {
			err = publication.VersionUnpublishEverywhere(
				ctx,
				srv.Repo,
				&srv.Connections.Publication,
				req.GetSpec().GetSelect().GetPropertyId(),
				listing.VersionStatus_VERSION_VERIFYING,
				listing.VersionStatus_VERSION_VERIFIED,
				listing.VersionStatus_VERSION_LET,
				listing.VersionStatus_VERSION_ARCHIVE,
				listing.VersionStatus_VERSION_WITHDRAWN,
			)
			if err != nil {
				return nil, err
			}
		}
	case *listing.VersionFlagsRequest_Test:
		flag = model.FlagTest
		value = f.Test
	case *listing.VersionFlagsRequest_Auto:
		flag = model.FlagAuto
		value = f.Auto
	case *listing.VersionFlagsRequest_ActiveOffer:
		flag = model.FlagActiveOffer
		value = f.ActiveOffer
	}

	err = srv.Repo.VersionFlagUpdate(ctx, model.VersionSpec(ctx, req.GetSpec(), false), flag, value)
	if err != nil {
		return nil, err
	}

	return &listing.VersionFlagsResponse{}, nil
}

func (srv *serviceImpl) VersionPublish(
	ctx context.Context,
	req *listing.VersionPublishRequest,
) (*listing.VersionPublishResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	var errs []string

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
		Statuses: []listing.VersionStatus{
			listing.VersionStatus_VERSION_VERIFYING,
			listing.VersionStatus_VERSION_ARCHIVE,
			listing.VersionStatus_VERSION_WITHDRAWN,
			listing.VersionStatus_VERSION_VERIFIED,
		},
	})
	if err != nil {
		return nil, err
	}

	if !verificationComplete(&lv.Verification) && !req.GetLetAgreed().GetValue() {
		return nil, errAdminVerificationRequired
	}

	waspublished := lv.Flags.ActiveOffer != "" ||
		(lv.PublicationsRightmove.Published() && lv.PublicationsZoopla.Published())

	err = publication.VersionPublishMashroom(ctx, srv.Repo, req.GetMashroom(), lv, req.GetUserId(), req.GetLetAgreed())
	if err != nil {
		return nil, err
	}

	err = publication.VersionPublishExternal(
		ctx,
		srv.Repo,
		srv.Connections.Publication.Rightmove,
		listing.PublicationPlatform_PUBLICATION_PLATFORM_RIGHTMOVE,
		&lv.PublicationsRightmove,
		req.GetRightmove(),
		lv,
		req.GetUserId(),
		req.GetRightmove(),
		req.GetLetAgreed(),
	)
	if err != nil {
		errs = append(errs, "Rightmove: "+err.Error())

		log.WithError(err).WithField("version", lv.ID).Error("publishing on rightmove")
	}

	err = publication.VersionPublishExternal(
		ctx,
		srv.Repo,
		srv.Connections.Publication.Boomin,
		listing.PublicationPlatform_PUBLICATION_PLATFORM_BOOMIN,
		&lv.PublicationsBoomin,
		req.GetBoomin(),
		lv,
		req.GetUserId(),
		req.GetBoomin(),
		req.GetLetAgreed(),
	)
	if err != nil {
		errs = append(errs, "Boomin: "+err.Error())

		log.WithError(err).WithField("version", lv.ID).Error("publishing on boomin")
	}

	err = publication.VersionPublishExternal(
		ctx,
		srv.Repo,
		srv.Connections.Publication.Zoopla,
		listing.PublicationPlatform_PUBLICATION_PLATFORM_ZOOPLA,
		&lv.PublicationsZoopla,
		req.GetZoopla(),
		lv,
		req.GetUserId(),
		req.GetZoopla(),
		req.GetLetAgreed(),
	)
	if err != nil {
		errs = append(errs, "Zoopla: "+err.Error())

		log.WithError(err).WithField("version", lv.ID).Error("publishing on zoopla")
	}

	if lv.PublicationsRightmove.Published() && lv.PublicationsZoopla.Published() && !waspublished {
		_, err = srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_PROPERTY_PUBLICATION,
			Action: notification.Action_ACTION_CREATE,
			Targets: []*notification.Target{
				{
					User: lv.People.Manager,
					Role: notification.Role_ROLE_MANAGER,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: lv.PropertyID,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_LISTING,
					EmbedValue: lv.ListingID,
				},
			},
		})
		if err != nil {
			log.WithError(err).Error("sending property published manager to manager")
		}
	}

	if len(errs) > 0 {
		return nil, grpcerr.New(grpcerr.ProviderError, strings.Join(errs, "; "))
	}

	return &listing.VersionPublishResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionUnpublish(
	ctx context.Context,
	req *listing.VersionUnpublishRequest,
) (*listing.VersionUnpublishResponse, error) {
	id, err := strconv.ParseUint(req.GetVersionId(), 0, 64)
	if err != nil {
		return nil, err
	}

	lv, err := srv.Repo.VersionGet(ctx, &model.VersionSpecOptions{
		VersionID: id,
	})
	if err != nil {
		return nil, err
	}

	if req.GetMashroom() && lv.PublicationsMashroom.Published() {
		last := lv.PublicationsMashroom.Last()
		last.UnpublishedAt = model.ServiceTime(time.Now())
		last.UnpublishedBy = req.GetUserId()

		err = srv.Repo.VersionPublicationEntrySave(ctx, last)
		if err != nil {
			log.WithError(err).WithField("version", lv.ID).Error("unpublishing on mashroom")
		} else {
			last.SyncedAt = model.ServiceTime(time.Now())
			last.PublicationID = ""
			last.PublicationURL = ""
		}
	}

	if req.GetRightmove() {
		err = publication.VersionUnpublishExternal(
			ctx,
			srv.Repo,
			srv.Connections.Publication.Rightmove,
			lv.PublicationsRightmove,
			lv.PropertyID,
			req.GetUserId(),
		)
		if err != nil {
			return nil, err
		}
	}

	if req.GetBoomin() {
		err = publication.VersionUnpublishExternal(
			ctx,
			srv.Repo,
			srv.Connections.Publication.Boomin,
			lv.PublicationsBoomin,
			lv.PropertyID,
			req.GetUserId(),
		)
		if err != nil {
			return nil, err
		}
	}

	if req.GetZoopla() {
		err = publication.VersionUnpublishExternal(
			ctx,
			srv.Repo,
			srv.Connections.Publication.Zoopla,
			lv.PublicationsZoopla,
			lv.PropertyID,
			req.GetUserId(),
		)
		if err != nil {
			return nil, err
		}
	}

	return &listing.VersionUnpublishResponse{
		Version: srv.renderVersion(ctx, lv, req.GetUserId()),
	}, nil
}

func (srv *serviceImpl) ServiceVersionTenancyBegin(
	ctx context.Context,
	req *listing.ServiceVersionTenancyBeginRequest,
) (resp *listing.ServiceVersionTenancyBeginResponse, err error) {
	getopts := model.VersionSpec(ctx, req.GetSpec(), false)
	getopts.Statuses = []listing.VersionStatus{
		listing.VersionStatus_VERSION_VERIFIED,
		listing.VersionStatus_VERSION_LET,
		listing.VersionStatus_VERSION_WITHDRAWN,
	}

	lv, err := srv.Repo.VersionGet(ctx, getopts)
	if err != nil {
		return nil, err
	}

	lv.TenancyID = req.GetTenancyId()
	lv.People.Tenants = req.GetTenants()

	err = srv.Repo.VersionUpdate(ctx, lv, listing.VersionStatus_VERSION_LET)
	if err != nil {
		return nil, err
	}

	err = publication.VersionUnpublishEverywhere(
		ctx,
		srv.Repo,
		&srv.Connections.Publication,
		lv.PropertyID,
		listing.VersionStatus_VERSION_LET,
	)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionTenancyBeginResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) ServiceVersionTenancyEnd(
	ctx context.Context,
	req *listing.ServiceVersionTenancyEndRequest,
) (resp *listing.ServiceVersionTenancyEndResponse, err error) {
	getopts := model.VersionSpec(ctx, req.GetSpec(), false)
	getopts.Statuses = []listing.VersionStatus{
		listing.VersionStatus_VERSION_VERIFIED,
		listing.VersionStatus_VERSION_LET,
		listing.VersionStatus_VERSION_WITHDRAWN,
	}

	lv, err := srv.Repo.VersionGet(ctx, getopts)
	if err != nil {
		return nil, err
	}

	err = srv.Repo.VersionUpdateStatus(ctx, lv.ID, lv.Status, listing.VersionStatus_VERSION_ARCHIVE)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionTenancyEndResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) createMigrationInvitation(
	inv *listing.InvitationStatus,
) (s model.VersionInvitationStatus) {
	if inv == nil {
		return
	}

	return model.VersionInvitationStatus{
		Pending:      inv.GetPending(),
		Sent:         model.ServiceTime(inv.GetSent().AsTime()),
		HostID:       inv.GetHostId(),
		InviteeID:    inv.GetInviteeId(),
		InviteeEmail: inv.GetInviteeEmail(),
		InviteePhone: inv.GetInviteePhone(),
		InviteeName:  inv.GetInviteeName(),
		InvitationID: inv.GetInvitationId(),
	}
}

func (srv *serviceImpl) ServiceVersionUpdate(
	ctx context.Context,
	req *listing.ServiceVersionUpdateRequest,
) (*listing.ServiceVersionUpdateResponse, error) {
	lv, err := srv.Repo.VersionGet(
		ctx,
		model.VersionSpec(ctx, req.GetSpec(), false),
	)
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, err
	}

	if lv == nil {
		return nil, sql.ErrNoRows
	}

	err = srv.versionUpdate(
		ctx,
		lv,
		req.GetDetails(),
		req.GetDescription(),
		req.GetRental(),
		req.GetPeople(),
		req.GetLocation(),
		req.GetViewingAvailability(),
		func() error {
			err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
			if err != nil {
				return err
			}

			err = srv.serviceVersionPublication(
				ctx,
				lv,
				req.GetPublication().GetMashroom(),
				req.GetPublication().GetRightmove(),
				req.GetPublication().GetBoomin(),
				req.GetPublication().GetZoopla(),
			)
			if err != nil {
				return err
			}

			return nil
		},
	)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionUpdateResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) serviceVersionPublication(
	ctx context.Context,
	lv *model.Version,
	mashroom, rightmove, boomin, zoopla *listing.PublicationItem,
) (err error) {
	if mashroom != nil {
		t := mashroom.GetPublishedAt().AsTime()

		entry := &model.PublicationEntry{
			ListingID:      lv.ListingID,
			PropertyID:     lv.PropertyID,
			Platform:       listing.PublicationPlatform_PUBLICATION_PLATFORM_MASHROOM,
			PublicationID:  lv.PropertyID,
			PublicationURL: mashroom.GetUrl(),
			PublishedBy:    mashroom.GetPublishedBy(),
			PublishedAt:    model.ServiceTime(t),
			UnpublishedAt:  model.ServiceTime{},
			ExpireAt:       model.ServiceTime{},
			SyncedAt:       model.ServiceTime(t),
			InvoiceIDs:     mashroom.GetInvoiceIds(),
		}

		err = srv.Repo.VersionPublicationEntrySave(ctx, entry)
		if err != nil {
			return fmt.Errorf("publishing on mashroom: %w", err)
		}

		lv.PublicationsMashroom = append(lv.PublicationsMashroom, entry)
	}

	if rightmove != nil {
		t := rightmove.GetPublishedAt().AsTime()
		texp := publication.VersionPublicationExpiry(lv.PublicationsRightmove, rightmove.GetExpireAt().AsTime())

		entry := &model.PublicationEntry{
			ListingID:      lv.ListingID,
			PropertyID:     lv.PropertyID,
			Platform:       listing.PublicationPlatform_PUBLICATION_PLATFORM_RIGHTMOVE,
			PublicationID:  rightmove.GetId(),
			PublicationURL: rightmove.GetUrl(),
			PublishedBy:    rightmove.GetPublishedBy(),
			PublishedAt:    model.ServiceTime(t),
			ExpireAt:       model.ServiceTime(texp),
			SyncedAt:       model.ServiceTime(t),
			InvoiceIDs:     mashroom.GetInvoiceIds(),
		}

		err = srv.Repo.VersionPublicationEntrySave(ctx, entry)
		if err != nil {
			return fmt.Errorf("publishing on rightmove: %w", err)
		}

		lv.PublicationsRightmove = append(lv.PublicationsRightmove, entry)
	}

	if boomin != nil {
		t := boomin.GetPublishedAt().AsTime()
		texp := publication.VersionPublicationExpiry(lv.PublicationsBoomin, boomin.GetExpireAt().AsTime())

		entry := &model.PublicationEntry{
			ListingID:      lv.ListingID,
			PropertyID:     lv.PropertyID,
			Platform:       listing.PublicationPlatform_PUBLICATION_PLATFORM_BOOMIN,
			PublicationID:  boomin.GetId(),
			PublicationURL: boomin.GetUrl(),
			PublishedBy:    boomin.GetPublishedBy(),
			PublishedAt:    model.ServiceTime(t),
			ExpireAt:       model.ServiceTime(texp),
			SyncedAt:       model.ServiceTime(t),
			InvoiceIDs:     mashroom.GetInvoiceIds(),
		}

		err = srv.Repo.VersionPublicationEntrySave(ctx, entry)
		if err != nil {
			return fmt.Errorf("publishing on boomin: %w", err)
		}

		lv.PublicationsBoomin = append(lv.PublicationsBoomin, entry)
	}

	if zoopla != nil {
		t := zoopla.GetPublishedAt().AsTime()
		texp := publication.VersionPublicationExpiry(lv.PublicationsZoopla, zoopla.GetExpireAt().AsTime())

		entry := &model.PublicationEntry{
			ListingID:      lv.ListingID,
			PropertyID:     lv.PropertyID,
			Platform:       listing.PublicationPlatform_PUBLICATION_PLATFORM_ZOOPLA,
			PublicationID:  zoopla.GetId(),
			PublicationURL: zoopla.GetUrl(),
			PublishedBy:    zoopla.GetPublishedBy(),
			PublishedAt:    model.ServiceTime(t),
			ExpireAt:       model.ServiceTime(texp),
			SyncedAt:       model.ServiceTime(t),
			InvoiceIDs:     mashroom.GetInvoiceIds(),
		}

		err = srv.Repo.VersionPublicationEntrySave(ctx, entry)
		if err != nil {
			return fmt.Errorf("publishing on zoopla: %w", err)
		}

		lv.PublicationsZoopla = append(lv.PublicationsZoopla, entry)
	}

	return nil
}

func versionCreatePreprocessSensitiveFieldChanged(
	description *listing.Description,
	rental *listing.Rental,
) bool {
	return description.GetPhotos() != nil ||
		description.GetPlans() != nil ||
		description.GetVideoEmbeds() != nil ||
		description.GetPostPropertyPitch() != nil ||
		rental.GetRentalPrice() != nil
}

func versionCreatePreprocessListableFieldChanged(
	description *listing.Description,
	rental *listing.Rental,
) bool {
	onlybank := description == nil &&
		rental.GetBankDetails() != nil &&
		rental.GetRentalPrice() == nil &&
		rental.GetBillsIncluded() == nil &&
		rental.GetBillsCovered() == nil &&
		rental.GetDepositReplacementFee() == nil &&
		rental.GetDssAvailable() == nil &&
		rental.GetDurationMaxMonths() == nil &&
		rental.GetDurationMinMonths() == nil &&
		rental.GetEarliestMovein() == nil &&
		rental.GetInventoryReport() == nil &&
		rental.GetTenancyDepositInsurance() == nil

	return !onlybank
}

func versionCreatePreprocess(
	lv *model.Version,
	description *listing.Description,
	rental *listing.Rental,
) (newstatus listing.VersionStatus) {
	newstatus = lv.Status

	switch lv.Status {
	case listing.VersionStatus_VERSION_VERIFIED:
		if versionCreatePreprocessSensitiveFieldChanged(description, rental) {
			lv.Verification = model.VersionVerification{}
			newstatus = listing.VersionStatus_VERSION_DRAFT
			lv.ID = 0
			lv.MigratedID = ""

			if lv.TenancyID != "" {
				lv.ListingID = ""
				lv.TenancyID = ""
				lv.People.Tenants = nil
				lv.People.Viewers = nil
				lv.People.Offerers = nil
				lv.Flags.ActiveOffer = ""
			}

			lv.Notification.Set(model.VersionNotificationVerificationPending, false)
		}
	case listing.VersionStatus_VERSION_VERIFYING:
		lv.Verification = model.VersionVerification{}
		newstatus = listing.VersionStatus_VERSION_DRAFT
	case listing.VersionStatus_VERSION_DRAFT:
		lv.Verification = model.VersionVerification{}
		// keep as is
	case listing.VersionStatus_VERSION_LET:
		if versionCreatePreprocessListableFieldChanged(description, rental) {
			lv.Verification = model.VersionVerification{}
			lv.ID = 0
			lv.ListingID = ""
			lv.TenancyID = ""
			lv.MigratedID = ""
			lv.People.Tenants = nil
			lv.People.Viewers = nil
			lv.People.Offerers = nil
			lv.Flags.ActiveOffer = ""
			newstatus = listing.VersionStatus_VERSION_DRAFT

			lv.Notification.Set(model.VersionNotificationVerificationPending, false)
		}
	default:
		lv.Verification = model.VersionVerification{}
		newstatus = listing.VersionStatus_VERSION_DRAFT
		lv.ID = 0
	}

	return
}

func (srv *serviceImpl) servicePropertyOpportunity(
	ctx context.Context,
	user, propertyID string,
	loc *model.VersionLocation,
	origin *model.VersionOrigin,
) {
	req := &closecom.LeadOpportunitySaveRequest{
		UserId: user,
	}

	switch {
	case origin.Tool != nil:
		switch {
		case origin.Tool.Valuation:
			req.Opportunity = &closecom.LeadOpportunitySaveRequest_Tool{
				Tool: &closecom.Tool{
					ToolName:     "Valuation",
					PropertyId:   propertyID,
					PropertyName: render.Location(loc, true).FormatAddress(),
				},
			}
		case origin.Tool.ExpenseTracker:
			req.Opportunity = &closecom.LeadOpportunitySaveRequest_Tool{
				Tool: &closecom.Tool{
					ToolName:     "Expense Tracker",
					PropertyId:   propertyID,
					PropertyName: render.Location(loc, true).FormatAddress(),
				},
			}
		case origin.Tool.CertificateRepository:
			req.Opportunity = &closecom.LeadOpportunitySaveRequest_Tool{
				Tool: &closecom.Tool{
					ToolName:     "Certificate Repository",
					PropertyId:   propertyID,
					PropertyName: render.Location(loc, true).FormatAddress(),
				},
			}
		}
	case origin.Product != nil:
		req.Opportunity = &closecom.LeadOpportunitySaveRequest_Product{
			Product: &closecom.Product{
				ProductId:       origin.Product.ProductID,
				ProductName:     origin.Product.ProductName,
				PropertyId:      propertyID,
				PropertyName:    render.Location(loc, true).FormatAddress(),
				PropertyCreated: true,
			},
		}
	default:
		req.Opportunity = &closecom.LeadOpportunitySaveRequest_Property{
			Property: &closecom.Property{
				PropertyId:   propertyID,
				PropertyName: render.Location(loc, true).FormatAddress(),
			},
		}
	}

	_, err := srv.Connections.Closecom.LeadOpportunitySave(ctx, req)
	if err != nil {
		log.WithError(err).Error("saving property opportunity")
	}
}

func (srv *serviceImpl) ServiceVersionCreate(
	ctx context.Context,
	req *listing.ServiceVersionCreateRequest,
) (*listing.ServiceVersionCreateResponse, error) {
	ctx = grpcctx.WithServiceUser(ctx)

	lv, err := srv.Repo.VersionGet(
		ctx,
		model.VersionSpec(ctx, listing.VersionSpecPropertyID("", req.GetPropertyId()), false),
	)
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, err
	}

	if lv == nil {
		lv = &model.Version{
			Created:             req.GetCreatedAt().AsTime(),
			Updated:             req.GetUpdatedAt().AsTime(),
			PropertyID:          req.GetPropertyId(),
			TenancyID:           req.GetTenancyId(),
			ListingID:           req.GetListingId(),
			IVR:                 req.GetIvr(),
			MigratedID:          req.GetMigratedId(),
			ID:                  0,
			Description:         model.VersionDescription{},
			Details:             model.VersionDetails{},
			People:              model.VersionPeople{},
			Rental:              model.VersionRental{},
			Location:            model.VersionLocation{},
			Valuation:           model.VersionValuation{},
			Verification:        model.VersionVerification{},
			Validation:          model.VersionValidation{},
			Origin:              versionOrigin(req.GetOrigin()),
			ViewingAvailability: defaultViewingAvailability(),
			Invitation: model.VersionInvitation{
				ManagerInvitation: srv.createMigrationInvitation(
					req.GetInvitation().GetManagerInvitation(),
				),
				LandlordInvitation: srv.createMigrationInvitation(
					req.GetInvitation().GetLandlordInvitation(),
				),
			},
			PublicationsMashroom:  nil,
			PublicationsZoopla:    nil,
			PublicationsRightmove: nil,
			Status:                req.GetStatus(),
			Flags: model.VersionFlags{
				Disabled: req.GetFlags().GetDisabled().GetValue(),
				Test:     req.GetFlags().GetTest().GetValue() || grpcctx.ContextIsTestUser(ctx),
				Auto:     req.GetFlags().GetAuto().GetValue(),
			},
		}
	}

	if lv.Created.Unix() <= 0 {
		lv.Created = time.Now()
	}

	if lv.Updated.Unix() <= 0 {
		lv.Updated = time.Now()
	}

	err = srv.versionCreate(
		ctx,
		lv,
		req.GetClientId(),
		req.GetDetails(),
		req.GetDescription(),
		req.GetRental(),
		req.GetPeople(),
		req.GetLocation(),
		req.GetViewingAvailability(),
		func(newstatus listing.VersionStatus) error {
			err = srv.Repo.VersionServiceCreate(ctx, lv, req.GetStatus())
			if err != nil {
				if lv.ID == 0 {
					return fmt.Errorf("creating property: %w", err)
				}

				lv.ID = 0

				err = srv.Repo.VersionServiceCreate(ctx, lv, req.GetStatus())
				if err != nil {
					return fmt.Errorf("creating property: %w", err)
				}
			}

			err = srv.serviceVersionPublication(
				ctx,
				lv,
				req.GetPublication().GetMashroom(),
				req.GetPublication().GetRightmove(),
				req.GetPublication().GetBoomin(),
				req.GetPublication().GetZoopla(),
			)
			if err != nil {
				return err
			}

			return nil
		},
	)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionCreateResponse{
		Version: srv.renderVersion(ctx, lv, ""),
	}, nil
}

func (srv *serviceImpl) commitPreface(ctx context.Context, lv *model.Version, clientID string) (err error) {
	var preface *model.Preface

	preface, err = srv.Repo.PrefaceGet(ctx, clientID)
	if err != nil {
		return err
	}

	srv.updateDetails(&lv.Details, render.Details(&preface.Details))
	srv.updateDescription(&lv.Description, render.Description(&preface.Description), lv.IVR)
	srv.updateRental(&lv.Rental, render.Rental(&preface.Rental, true))

	lv.Origin = model.VersionOrigin{
		Preface: true,
	}

	err = srv.updateLocation(
		ctx,
		&lv.Location,
		&listing.NewLocation{Raw: render.Location(&preface.Location, true)},
		&lv.Timezone,
	)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) VersionPermissionCheck(
	ctx context.Context,
	req *listing.VersionPermissionCheckRequest,
) (*listing.VersionPermissionCheckResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetSpec().GetUserId())

	var granted bool

	switch req.GetRole() {
	case listing.VersionRole_VERSION_ROLE_TENANT:
		granted = rel.tenant
	case listing.VersionRole_VERSION_ROLE_LANDLORD:
		granted = rel.landlord
	case listing.VersionRole_VERSION_ROLE_MANAGER:
		granted = rel.manager
	case listing.VersionRole_VERSION_ROLE_SUBMITTER:
		granted = rel.submitter
	case listing.VersionRole_VERSION_ROLE_ADMIN:
		granted = rel.admin
	}

	return &listing.VersionPermissionCheckResponse{
		Granted: granted,
	}, nil
}

func (srv *serviceImpl) ServiceVersionSetAdmin(
	ctx context.Context,
	req *listing.ServiceVersionSetAdminRequest,
) (*listing.ServiceVersionSetAdminResponse, error) {
	ctx = grpcctx.WithServiceUser(ctx)

	err := srv.Repo.VersionSetAdmin(ctx, model.VersionSpec(ctx, req.GetSpec(), false), req.GetAdmin())
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionSetAdminResponse{}, nil
}

func (srv *serviceImpl) ServiceVersionUpdateMedia(
	ctx context.Context,
	req *listing.ServiceVersionUpdateMediaRequest,
) (*listing.ServiceVersionUpdateMediaResponse, error) {
	ctx = grpcctx.WithServiceUser(ctx)

	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, err
	}

	lv.Description.Plans = updateMedia(lv.Description.Plans, req.GetPlans())
	lv.Description.Photos = updateMedia(lv.Description.Photos, req.GetPhotos())
	lv.Description.VideoEmbeds = updateMedia(lv.Description.VideoEmbeds, req.GetVideoEmbeds())

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionUpdateMediaResponse{}, nil
}

func updateMedia(arr []string, changes *listing.FileChanges) []string {
	var newarr []string

	sort.Strings(changes.GetRemoveFiles())

	for _, a := range arr {
		idx := sort.SearchStrings(changes.GetRemoveFiles(), a)

		if idx < len(changes.GetRemoveFiles()) && changes.GetRemoveFiles()[idx] == a {
			continue
		}

		newarr = append(newarr, a)
	}

	sortedadd := make([]string, len(changes.GetAddFiles()))

	copy(sortedadd, changes.GetAddFiles())

	sort.Strings(sortedadd)

	for _, a := range changes.GetAddFiles() {
		idx := sort.SearchStrings(sortedadd, a)

		if idx < len(sortedadd) && sortedadd[idx] == a {
			continue
		}

		newarr = append(newarr, a)
	}

	return newarr
}

func (srv *serviceImpl) VersionViewingAvailability(
	ctx context.Context,
	req *listing.VersionViewingAvailabilityRequest,
) (*listing.VersionViewingAvailabilityResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	srv.updateViewingAvailability(&lv.ViewingAvailability, &listing.ViewingAvailability{
		Slots: req.GetSlots(),
	})

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &listing.VersionViewingAvailabilityResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionListingServicesSelect(
	ctx context.Context,
	req *listing.VersionListingServicesSelectRequest,
) (*listing.VersionListingServicesSelectResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	var products []*invoice.Product

	for _, p := range req.GetServices() {
		products = append(products, &invoice.Product{
			ServiceId: p,
			Details: &invoice.ProductDetails{
				Property: &invoice.ProductDetailsProperty{
					Id:      wrapperspb.String(p),
					Listing: wrapperspb.String(lv.ListingID),
				},
			},
		})
	}

	_, err = srv.Connections.Invoice.OrderSave(ctx, &invoice.OrderSaveRequest{
		Userid:            req.GetSpec().GetUserId(),
		Products:          products,
		Periodic:          false,
		CorrelationType:   invoice.CorrelationTypeProperty,
		CorrelationId:     lv.PropertyID,
		CorrelationTarget: invoice.CorrelationTargetListingServices,
	})
	if err != nil {
		return nil, err
	}

	return &listing.VersionListingServicesSelectResponse{
		Version: srv.renderVersion(ctx, lv, req.GetSpec().GetUserId()),
	}, nil
}

func (srv *serviceImpl) VersionListingServicesPay(
	ctx context.Context,
	req *listing.VersionListingServicesPayRequest,
) (*listing.VersionListingServicesPayResponse, error) {
	lv, err := srv.Repo.VersionGet(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	resp, err := srv.Connections.Invoice.OrderList(ctx, &invoice.OrderListRequest{
		Userid: req.GetSpec().GetUserId(),
		Params: &query.QueryParams{
			Filter: []*query.QueryParamsFilter{
				{
					Fields:   []string{"correlation_type"},
					Value:    []string{invoice.CorrelationTypeProperty},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"correlation_id"},
					Value:    []string{lv.PropertyID},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"correlation_target"},
					Value:    []string{invoice.CorrelationTargetListingServices},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"status"},
					Value:    []string{invoice.OrderStatus_ORDER_PENDING.String()},
					Operator: query.QueryFilterOperator_EQUALS,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	if len(resp.GetOrders()) == 0 {
		return nil, grpcerr.Newf(grpcerr.InvalidState, "no pending orders for property %s", lv.PropertyID)
	}

	ord := resp.GetOrders()[0]

	comm, err := srv.Connections.Invoice.OrderCommit(ctx, &invoice.OrderCommitRequest{
		Userid: req.GetSpec().GetUserId(),
		Id:     ord.GetId(),
	})
	if err != nil {
		return nil, err
	}

	pay, err := srv.Connections.Invoice.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId: comm.GetInvoice().GetInvoiceId(),
		User:      req.GetSpec().GetUserId(),
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Stripe{
				Stripe: &invoice.PaymentSystemStripeRequest{
					Type: cardpayment.PaymentType_CARD,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &listing.VersionListingServicesPayResponse{
		StripeId:     pay.GetPayment().GetSystem().GetStripe().GetResponse().GetStripeId(),
		ClientSecret: pay.GetPayment().GetSystem().GetStripe().GetResponse().GetClientSecret(),
	}, nil
}

func (srv *serviceImpl) ServiceVersionWithdraw(
	ctx context.Context,
	req *listing.ServiceVersionWithdrawRequest,
) (*listing.ServiceVersionWithdrawResponse, error) {
	lvs, err := srv.Repo.VersionsList(ctx, model.VersionSpec(ctx, req.GetSpec(), false))
	if err != nil {
		return nil, err
	}

	for _, lv := range lvs {
		rel := srv.userRelation(lv, req.GetSpec().GetUserId())

		if !rel.editor {
			return nil, errAccessDenied
		}

		err = srv.Repo.VersionUpdateStatus(ctx, lv.ID, lv.Status, listing.VersionStatus_VERSION_WITHDRAWN)
		if err != nil {
			return nil, err
		}
	}

	return &listing.ServiceVersionWithdrawResponse{}, nil
}

func (srv *serviceImpl) ServiceGetAverageRentPrices(
	ctx context.Context,
	req *listing.ServiceGetAverageRentPricesRequest,
) (*listing.ServiceGetAverageRentPricesResponse, error) {
	ms, err := srv.Repo.VersionAverageRentPrices(ctx, &model.VersionAverageRentOptions{
		Country: req.GetCountry(),
		Region:  req.GetRegion(),
		County:  req.GetCounty(),
		Town:    req.GetTown(),
		Statuses: []listing.VersionStatus{
			listing.VersionStatus_VERSION_VERIFIED,
			listing.VersionStatus_VERSION_LET,
		},
	})
	if err != nil {
		return nil, err
	}

	resp := &listing.ServiceGetAverageRentPricesResponse{
		Time:     timestamppb.Now(),
		Bedrooms: make(map[int64]*listing.AverageRentPrices),
	}

	for _, m := range ms {
		resp.Bedrooms[int64(m.Bedrooms)] = &listing.AverageRentPrices{
			Total: int64(m.Total),
			Price: m.Price,
		}
	}

	return resp, nil
}

func (srv *serviceImpl) ServicePublicationExtend(
	ctx context.Context,
	req *listing.ServicePublicationExtendRequest,
) (*listing.ServicePublicationExtendResponse, error) {
	ctx = grpcctx.WithServiceUser(ctx)

	opts := model.VersionSpec(ctx, req.GetSpec(), false)

	opts.Statuses = []listing.VersionStatus{
		listing.VersionStatus_VERSION_VERIFIED,
		listing.VersionStatus_VERSION_VERIFYING,
		listing.VersionStatus_VERSION_DRAFT,
	}

	v, err := srv.Repo.VersionGet(ctx, opts)
	if err != nil {
		return nil, err
	}

	var errs []error

	zooplaPub := v.PublicationsZoopla.Last()
	if zooplaPub == nil {
		zooplaPub = &model.PublicationEntry{
			ListingID:  v.ListingID,
			PropertyID: v.PropertyID,
			Platform:   listing.PublicationPlatform_PUBLICATION_PLATFORM_ZOOPLA,
			ExpireAt:   model.ServiceTime(time.Now()),
		}

		v.PublicationsZoopla = append(v.PublicationsZoopla, zooplaPub)
	}

	found := false

	for _, inv := range zooplaPub.InvoiceIDs {
		if inv == req.GetInvoiceId() {
			found = true

			break
		}
	}

	if !found {
		zooplaPub.ExpireAt = model.ServiceTime(time.Time(zooplaPub.ExpireAt).Add(time.Hour * 24 * 30))
	}

	err = srv.Repo.VersionPublicationEntrySave(ctx, zooplaPub)
	if err != nil {
		errs = append(errs, err)
	}

	rightmovePub := v.PublicationsRightmove.Last()
	if rightmovePub == nil {
		rightmovePub = &model.PublicationEntry{
			ListingID:  v.ListingID,
			PropertyID: v.PropertyID,
			Platform:   listing.PublicationPlatform_PUBLICATION_PLATFORM_RIGHTMOVE,
			ExpireAt:   model.ServiceTime(time.Now()),
		}

		v.PublicationsRightmove = append(v.PublicationsRightmove, rightmovePub)
	}

	found = false

	for _, inv := range rightmovePub.InvoiceIDs {
		if inv == req.GetInvoiceId() {
			found = true

			break
		}
	}

	if !found {
		rightmovePub.ExpireAt = model.ServiceTime(time.Time(rightmovePub.ExpireAt).Add(time.Hour * 24 * 30))
	}

	err = srv.Repo.VersionPublicationEntrySave(ctx, rightmovePub)
	if err != nil {
		errs = append(errs, err)
	}

	boominPub := v.PublicationsBoomin.Last()
	if boominPub == nil {
		boominPub = &model.PublicationEntry{
			ListingID:  v.ListingID,
			PropertyID: v.PropertyID,
			Platform:   listing.PublicationPlatform_PUBLICATION_PLATFORM_BOOMIN,
			ExpireAt:   model.ServiceTime(time.Now()),
		}

		v.PublicationsBoomin = append(v.PublicationsBoomin, boominPub)
	}

	found = false

	for _, inv := range boominPub.InvoiceIDs {
		if inv == req.GetInvoiceId() {
			found = true

			break
		}
	}

	if !found {
		boominPub.ExpireAt = model.ServiceTime(time.Time(boominPub.ExpireAt).Add(time.Hour * 24 * 30))
	}

	err = srv.Repo.VersionPublicationEntrySave(ctx, boominPub)
	if err != nil {
		errs = append(errs, err)
	}

	if len(errs) > 0 {
		return nil, fmt.Errorf("updating publications: %v", errs)
	}

	return &listing.ServicePublicationExtendResponse{}, nil
}

func (srv *serviceImpl) serviceUserSyncLeadStats(ctx context.Context, user string) {
	count, err := srv.Repo.VersionPropertiesCount(ctx, user)
	if err != nil {
		log.WithError(err).WithField("user", user).Error("getting properties count")

		return
	}

	_, err = srv.Connections.Closecom.LeadStatsSave(ctx, &closecom.LeadStatsSaveRequest{
		UserId: user,
		Stats: &closecom.LeadStatsSaveRequest_Properties{
			Properties: &closecom.LeadStatsProperties{
				PropertiesCount: int32(count[user]),
			},
		},
	})
	if err != nil {
		log.WithError(err).WithField("user", user).Error("sending closecom properties stats")
	}
}

func (srv *serviceImpl) serviceUserSync(ctx context.Context, user string) {
	srv.serviceUserSyncLeadStats(ctx, user)
}

func (srv *serviceImpl) ServiceUserSync(
	ctx context.Context,
	req *listing.ServiceUserSyncRequest,
) (*listing.ServiceUserSyncResponse, error) {
	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	go srv.serviceUserSync(bgctx, req.GetUserId())

	return &listing.ServiceUserSyncResponse{}, nil
}

func (srv *serviceImpl) refreshNearbyCache(
	ctx context.Context,
	id uint64,
	lon, lat float64,
) (cache *model.NearbyCache, err error) {
	cache = &model.NearbyCache{
		VersionID: id,
	}

	var points []*geo.PointRelative

	resp, err := srv.getLocations(ctx, geo.PointType_POINT_METRO_STATION, lon, lat)
	if err != nil {
		return nil, err
	}

	points = append(points, resp...)

	resp, err = srv.getLocations(ctx, geo.PointType_POINT_RAILWAY_STATION, lon, lat)
	if err != nil {
		return nil, err
	}

	points = append(points, resp...)

	resp, err = srv.getLocations(ctx, geo.PointType_POINT_SCHOOL, lon, lat)
	if err != nil {
		return nil, err
	}

	points = append(points, resp...)

	resp, err = srv.getLocations(ctx, geo.PointType_POINT_HOSPITAL, lon, lat)
	if err != nil {
		return nil, err
	}

	points = append(points, resp...)

	for _, p := range points {
		cache.Stations = append(cache.Stations, &model.NearbyPoint{
			ID:          p.GetPoint().GetId(),
			Type:        p.GetPoint().GetType(),
			Name:        p.GetPoint().GetName(),
			Description: p.GetPoint().GetDescription(),
			Longitude:   p.GetPoint().GetLocation().GetLongitude(),
			Latitude:    p.GetPoint().GetLocation().GetLatitude(),
			Count:       p.GetPoint().GetCount(),
			Distance:    p.GetDistance(),
		})
	}

	err = srv.Repo.VersionNearbyInsert(ctx, cache)

	return
}

func (srv *serviceImpl) getLocations(
	ctx context.Context,
	pointType geo.PointType,
	lon, lat float64,
) ([]*geo.PointRelative, error) {
	resp, err := srv.Connections.Geo.GetStations(ctx, &geo.GetStationsRequest{
		PointType: pointType,
		Center: &geo.Location{
			Latitude:  lat,
			Longitude: lon,
		},
		RadiusInMeters: 10000,
		Limit:          2,
	})
	if err != nil {
		return nil, err
	}

	return resp.Points, nil
}

func (srv *serviceImpl) ServicePropertyCount(
	ctx context.Context,
	req *listing.ServicePropertyCountRequest,
) (*listing.ServicePropertyCountResponse, error) {
	count, err := srv.Repo.VersionPropertiesCount(ctx, req.GetLandlords().GetValue()...)
	if err != nil {
		return nil, err
	}

	return &listing.ServicePropertyCountResponse{
		Count: count,
	}, nil
}

func (srv *serviceImpl) ServiceVersionRelatedResource(
	ctx context.Context,
	req *listing.ServiceVersionRelatedResourceRequest,
) (*listing.ServiceVersionRelatedResourceResponse, error) {
	opts := model.VersionSpec(ctx, req.GetSpec(), false)

	lv, err := srv.Repo.VersionGet(ctx, opts)
	if err != nil {
		return nil, err
	}

	switch req.GetResource() {
	case listing.RelatedResource_RELATED_RESOURCE_VIEWING:
		lv.People.Viewers.Set(req.GetUserId(), viewing.ViewingState(req.GetState()))
	case listing.RelatedResource_RELATED_RESOURCE_OFFER:
		lv.People.Offerers.Set(req.GetUserId(), offer.OfferState(req.GetState()))
	}

	err = srv.Repo.VersionUpdate(ctx, lv, lv.Status)
	if err != nil {
		return nil, err
	}

	return &listing.ServiceVersionRelatedResourceResponse{
		VersionId: strconv.FormatUint(lv.ID, 10),
		ListingId: lv.ListingID,
	}, nil
}

func (srv *serviceImpl) VersionBankDetailsGet(
	ctx context.Context,
	req *listing.VersionBankDetailsGetRequest,
) (*listing.VersionBankDetailsGetResponse, error) {
	getopts := model.VersionSpec(ctx, req.GetSpec(), false)

	lv, err := srv.Repo.VersionGet(ctx, getopts)
	if err != nil {
		return nil, err
	}

	rel := srv.userRelation(lv, req.GetSpec().GetUserId())

	if !rel.editor {
		return nil, errAccessDenied
	}

	return &listing.VersionBankDetailsGetResponse{
		BankDetails: &listing.BankDetails{
			Society:           lv.Rental.BankDetails.Society,
			SortCode:          lv.Rental.BankDetails.SortCode,
			AccountNumber:     lv.Rental.BankDetails.AccountNumber,
			AccountName:       lv.Rental.BankDetails.AccountName,
			PersonalReference: lv.Rental.BankDetails.PersonalReference,
		},
	}, nil
}
