package service

import (
	"context"
	"strings"
	"time"

	address "gitlab.mashroom.com/mashroom/backend/address/api"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/optional"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	geo "gitlab.mashroom.com/mashroom/backend/geo/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/intel"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
)

func (srv *serviceImpl) updateDescription(
	m *model.VersionDescription,
	data *listing.Description,
	ivr string,
) {
	if data == nil {
		return
	}

	if data.GetPostHeadline() != nil {
		m.PostHeadline = data.GetPostHeadline().GetValue()
	}

	if data.GetPostPropertyPitch() != nil {
		m.PostPropertyPitch = data.GetPostPropertyPitch().GetValue()
	}

	if data.GetPostAreaPitch() != nil {
		m.PostAreaPitch = data.GetPostAreaPitch().GetValue()
	}

	if data.GetPostComment() != nil {
		m.PostComment = data.GetPostComment().GetValue()
	}

	if data.GetPhotos() != nil {
		m.Photos = make([]string, 0)

		for _, p := range util.UniqueStrings(data.GetPhotos().GetValue()) {
			m.Photos = append(m.Photos, document.EncodeVariant(p, &document.ImageVariant{
				Variants: &document.ImageVariant_IvrWatermark{
					IvrWatermark: &document.ImageVariantIVRWatermark{
						Ivr: ivr,
					},
				},
			}))
		}
	}

	if data.GetPlans() != nil {
		m.Plans = make([]string, 0)

		for _, p := range util.UniqueStrings(data.GetPlans().GetValue()) {
			m.Plans = append(m.Plans, document.EncodeVariant(p, &document.ImageVariant{
				Variants: &document.ImageVariant_IvrWatermark{
					IvrWatermark: &document.ImageVariantIVRWatermark{
						Ivr: ivr,
					},
				},
			}))
		}
	}

	if data.GetVideoEmbeds() != nil {
		m.VideoEmbeds = util.UniqueStrings(data.GetVideoEmbeds().GetValue())
	}

	if data.GetTags() != nil {
		m.Tags = data.GetTags().GetValue()
	}
}

func (srv *serviceImpl) updateDetails(
	m *model.VersionDetails,
	data *listing.Details,
) {
	if data == nil {
		return
	}

	if data.GetTypeOther() != nil {
		m.TypeOther = optional.NewString(data.GetTypeOther().GetValue())
	}

	if data.GetArea() != nil {
		m.Area = optional.NewFloat64(data.GetArea().GetValue())
	}

	if data.GetBedroomsNo() != nil {
		m.BedroomsNo = optional.NewInt32(data.GetBedroomsNo().GetValue())
	}

	if data.GetBathroomsNo() != nil {
		m.BathroomsNo = optional.NewInt32(data.GetBathroomsNo().GetValue())
	}

	if data.GetLivingroomsNo() != nil {
		m.LivingroomsNo = optional.NewInt32(data.GetLivingroomsNo().GetValue())
	}

	if data.GetKitchenNo() != nil {
		m.KitchenNo = optional.NewInt32(data.GetKitchenNo().GetValue())
	}

	if data.GetAmenityFurniture() != 0 {
		m.AmenityFurniture = data.GetAmenityFurniture()
	}

	if data.GetAmenityOutsideSpace() != nil {
		m.AmenityOutsideSpace = data.GetAmenityOutsideSpace().GetValue()
	}

	if data.GetAmenityParking() != nil {
		m.AmenityParking = data.GetAmenityParking().GetValue()
	}

	if data.GetPetsAllowed() != nil {
		m.PetsAllowed = optional.NewBool(data.GetPetsAllowed().GetValue())
	}

	if data.GetType() != 0 {
		m.Type = data.GetType()

		if m.Type == listing.PropertyType_PROPERTY_TYPE_STUDIO {
			m.BedroomsNo = nil
			m.BathroomsNo = nil
			m.LivingroomsNo = nil
			m.KitchenNo = nil
		}
	}
}

func (srv *serviceImpl) updatePeople(
	m *model.VersionPeople,
	data *listing.People,
) {
	if data == nil {
		return
	}

	if data.GetLandlord() != nil {
		m.Landlord = data.GetLandlord().GetValue()
	}

	if data.GetManager() != nil {
		m.Manager = data.GetManager().GetValue()
	}

	if data.GetSubmitter() != nil {
		m.Submitter = data.GetSubmitter().GetValue()
	}

	if m.Manager == "" {
		m.Manager = m.Landlord
	}
}

func (srv *serviceImpl) updateRental(
	m *model.VersionRental,
	data *listing.Rental,
) {
	if data == nil {
		return
	}

	if data.GetRentalPrice() != nil {
		m.RentalPrice = optional.NewFloat64(data.GetRentalPrice().GetValue())
	}

	if data.GetDssAvailable() != nil {
		m.DssAvailable = optional.NewBool(data.GetDssAvailable().GetValue())
	}

	if data.GetBillsIncluded() != nil {
		m.BillsIncluded = optional.NewBool(data.GetBillsIncluded().GetValue())
	}

	if data.GetBillsCovered() != nil {
		m.BillsCovered = data.GetBillsCovered().GetValue()
	}

	if data.GetDurationMinMonths() != nil {
		m.DurationMinMonths = optional.NewInt32(data.GetDurationMinMonths().GetValue())
	}

	if data.GetDurationMaxMonths() != nil {
		m.DurationMaxMonths = optional.NewInt32(data.GetDurationMaxMonths().GetValue())
	}

	if data.GetEarliestMovein() != nil {
		em := data.GetEarliestMovein()
		d := time.Date(int(em.GetYear()), time.Month(em.GetMonth()), int(em.GetDay()), 0, 0, 0, 0, time.UTC)

		m.EarliestMovein = model.ServiceTime(d)
	}

	if data.GetTenancyDepositInsurance() != nil {
		m.TenancyDepositInsurance = optional.NewBool(data.GetTenancyDepositInsurance().GetValue())
	}

	if data.GetInventoryReport() != nil {
		m.InventoryReport = optional.NewBool(data.GetInventoryReport().GetValue())
	}

	if data.GetDepositReplacementFee() != nil {
		m.DepositReplacementFee = optional.NewFloat64(data.GetDepositReplacementFee().GetValue())
	}

	if data.GetBankDetails().GetAccountNumber() != "" && data.GetBankDetails().GetSortCode() != "" {
		m.BankDetails = model.BankDetails{
			Society:           data.GetBankDetails().GetSociety(),
			SortCode:          data.GetBankDetails().GetSortCode(),
			AccountNumber:     data.GetBankDetails().GetAccountNumber(),
			AccountName:       data.GetBankDetails().GetAccountName(),
			PersonalReference: data.GetBankDetails().GetPersonalReference(),
		}
	}
}

func (srv *serviceImpl) updateTimezone(ctx context.Context, lon, lat float64, tz *string) {
	if tz == nil {
		return
	}

	resp, _ := srv.Connections.Geo.ListTimezones(ctx, &geo.TimezonesRequest{
		TimezonesRequestData: []*geo.TimezonesRequestData{
			{
				Id:        0,
				Longitude: lon,
				Latitude:  lat,
			},
		},
	})

	rdata := resp.GetTimezonesResponseData()

	if len(rdata) > 0 {
		*tz = rdata[0].Timezone
	}
}

func (srv *serviceImpl) updateLocationRaw(
	ctx context.Context,
	m *model.VersionLocation,
	data *listing.Location,
	tz *string,
) {
	if data == nil {
		return
	}

	if data.GetPostcode() != nil {
		m.Postcode = data.GetPostcode().GetValue()
	}

	if data.GetAddress() != nil {
		m.Address = data.GetAddress().GetValue()
	}

	if data.GetAddress1() != nil {
		m.Address1 = data.GetAddress1().GetValue()
	}

	if data.GetAddress2() != nil {
		m.Address2 = data.GetAddress2().GetValue()
	}

	if data.GetAddress3() != nil {
		m.Address3 = data.GetAddress3().GetValue()
	}

	if data.GetAddress4() != nil {
		m.Address4 = data.GetAddress4().GetValue()
	}

	if data.GetCity() != nil {
		m.City = data.GetCity().GetValue()
	}

	if data.GetStreet() != nil {
		m.Street = data.GetStreet().GetValue()
	}

	if data.GetBuilding() != nil {
		m.Building = data.GetBuilding().GetValue()
	}

	if data.GetFlat() != nil {
		m.Flat = data.GetFlat().GetValue()
	}

	if data.GetGps() != nil {
		m.GPS.Longitude = data.GetGps().GetLongitude()
		m.GPS.Latitude = data.GetGps().GetLatitude()

		srv.updateTimezone(ctx, data.GetGps().GetLongitude(), data.GetGps().GetLatitude(), tz)
	}

	if data.GetCountry() != nil {
		m.Country = data.GetCountry().GetValue()
	}

	if data.GetPostTown() != nil {
		m.PostTown = data.GetPostTown().GetValue()
	}

	if data.GetCounty() != nil {
		m.County = data.GetCounty().GetValue()
	}

	if data.GetAdministrativeCounty() != nil {
		m.AdministrativeCounty = data.GetAdministrativeCounty().GetValue()
	}

	if data.GetPostalCounty() != nil {
		m.PostalCounty = data.GetPostalCounty().GetValue()
	}

	if data.GetTraditionalCounty() != nil {
		m.TraditionalCounty = data.GetTraditionalCounty().GetValue()
	}

	if data.GetDistrict() != nil {
		m.District = data.GetDistrict().GetValue()
	}
}

func (srv *serviceImpl) updateLocationUdprn(
	ctx context.Context,
	m *model.VersionLocation,
	udprn string,
	tz *string,
) error {
	if udprn == "" {
		return nil
	}

	addrresp, err := srv.Connections.Address.GetAddressByUdprn(ctx, &address.UdprnRequest{
		UdprnUrl: udprn,
	})
	if err != nil {
		return err
	}

	addr := addrresp.GetAddress()

	m.GPS = model.GpsCoordinates{
		Longitude: addr.GetGps().GetLongitude(),
		Latitude:  addr.GetGps().GetLatitude(),
	}

	srv.updateTimezone(ctx, addr.GetGps().GetLongitude(), addr.GetGps().GetLatitude(), tz)

	m.Postcode = addr.GetPostcode()
	m.Address = addr.GetAddress()
	m.Address1 = addr.GetAddress1()
	m.Address2 = addr.GetAddress2()
	m.Address3 = addr.GetAddress3()
	m.Address4 = addr.GetAddress4()
	m.Flat = addr.GetFlat()
	m.Building = addr.GetBuilding()
	m.Street = addr.GetStreet()
	m.City = addr.GetCity()
	m.Country = addr.GetCountry()
	m.County = addr.GetCounty()
	m.AdministrativeCounty = addr.GetAdministrativeCounty()
	m.District = addr.GetDistrict()
	m.PostalCounty = addr.GetPostalCounty()
	m.PostTown = addr.GetPostTown()
	m.TraditionalCounty = addr.GetTraditionalCounty()

	if len(strings.TrimSpace(m.Address)) == 0 {
		m.Address = strings.Join(util.NonemptyStrings(
			addr.GetAddress1(),
			addr.GetAddress2(),
			addr.GetAddress3(),
			addr.GetAddress4(),
		), " ")
	}

	return nil
}

func (srv *serviceImpl) updateLocation(
	ctx context.Context,
	m *model.VersionLocation,
	data *listing.NewLocation,
	tz *string,
) error {
	if data == nil {
		return nil
	}

	if data.GetUdprn() != nil {
		err := srv.updateLocationUdprn(ctx, m, data.GetUdprn().GetValue(), tz)
		if err != nil {
			return err
		}
	}

	if data.GetRaw() != nil {
		srv.updateLocationRaw(ctx, m, data.GetRaw(), tz)
	}

	resp, err := srv.Connections.Geo.GetNearestTownsForLocationList(ctx, &geo.GetNearestTownsForLocationListRequest{
		Locations: []*geo.Location{
			{
				Longitude: m.GPS.Longitude,
				Latitude:  m.GPS.Latitude,
			},
		},
		RadiusInMeters: 5000,
	})
	if err != nil {
		log.WithError(err).Errorf("getting geo location for property gps: %v", m.GPS)
	}

	if len(resp.GetTowns()) > 0 {
		town := resp.GetTowns()[0]
		m.GeoTown = town.GetName()
		m.GeoRegion = town.GetRegion()
		m.GeoCounty = town.GetCounty()
	}

	return nil
}

func (srv *serviceImpl) updateValuation(
	ctx context.Context,
	loc *model.VersionLocation,
	det *model.VersionDetails,
) (v model.VersionValuation, raw string, err error) {
	if loc == nil || det == nil || loc.Street == "" || loc.Postcode == "" || det.Type == 0 ||
		det.BedroomsNo.GetValue() == 0 || det.AmenityFurniture == 0 {
		return
	}

	var typ intel.TypeValue

	switch det.Type {
	case listing.PropertyType_PROPERTY_TYPE_DETACHED:
		typ = intel.TypeValueDetached
	case listing.PropertyType_PROPERTY_TYPE_SEMI_DETACHED:
		typ = intel.TypeValueSemiDetached
	case listing.PropertyType_PROPERTY_TYPE_TERRACED:
		typ = intel.TypeValueTerraced
	case listing.PropertyType_PROPERTY_TYPE_STUDIO:
		typ = intel.TypeValueStudio
	default:
		typ = intel.TypeValueOther
	}

	var furnished intel.FurnishedValue

	switch det.AmenityFurniture {
	case listing.PropertyFurniture_PROPERTY_FURNITURE_FURNISHED:
		furnished = intel.FurnishedValueFurnished
	default:
		furnished = intel.FurnishedValueUnfurnished
	}

	resp, err := srv.IntelClient.Lettings(ctx, &intel.ValuationRequest{
		Location: intel.ValuationRequestLocation{
			Postcode: loc.Postcode,
			Building: loc.Building,
			Flat:     loc.Flat,
			Street:   loc.Street,
		},
		Details: intel.ValuationRequestDetails{
			Type:       typ,
			BedroomsNo: int(det.BedroomsNo.GetValue()),
			Furnished:  furnished,
			Ownership:  1,
		},
		EstimatedValue: intel.ValuationRequestEstimatedValue{
			InitialAskingPriceLow:  0,
			InitialAskingPriceHigh: 1000000,
		},
	})
	if err != nil {
		return
	}

	raw = resp.Raw

	v.Estimate = resp.AvmOutput.Estimate
	v.Confidence = resp.AvmOutput.Confidence
	v.Additional.VariationConfidence = resp.AvmOutput.Additional.VariationConfidence
	v.Additional.ComparatorNumberConfidence = resp.AvmOutput.Additional.ComparatorNumberConfidence
	v.LowerBound = resp.AvmOutput.LowerBound
	v.UpperBound = resp.AvmOutput.UpperBound

	return
}

func (srv *serviceImpl) updateViewingAvailability(
	m *model.VersionViewingAvailability,
	data *listing.ViewingAvailability,
) {
	if data == nil {
		return
	}

	var arr model.VersionViewingAvailability

	for _, s := range data.GetSlots() {
		arr = append(arr, &model.VersionViewingAvailabilitySlot{
			Period:      s.GetPeriod(),
			StartOffset: s.GetStartOffset().AsDuration(),
			EndOffset:   s.GetEndOffset().AsDuration(),
		})
	}

	*m = arr
}
