package service

import (
	"context"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
)

func (srv *serviceImpl) validateVersion(ctx context.Context, lv *model.Version) {
	lv.Validation = srv.validate(
		ctx,
		&lv.Location,
		&lv.Details,
		&lv.Rental,
		&lv.People,
		&lv.Description,
		lv.PropertyID,
	)
}

func (srv *serviceImpl) validate(
	ctx context.Context,
	loc *model.VersionLocation,
	det *model.VersionDetails,
	rent *model.VersionRental,
	peop *model.VersionPeople,
	desc *model.VersionDescription,
	propertyID string,
) model.VersionValidation {
	val := model.VersionValidation{
		Valid:      true,
		Verifiable: true,
	}

	srv.validateVersionLocation(&val, loc)
	srv.validateVersionDetails(&val, det)
	srv.validateVersionRental(&val, rent)
	srv.validateVersionPeople(ctx, &val, peop)
	srv.validateVersionDescription(ctx, &val, desc, peop, propertyID)
	srv.validateVersionDocuments(ctx, &val, peop, propertyID)

	return val
}

func (srv *serviceImpl) validateVersionLocation(
	val *model.VersionValidation,
	loc *model.VersionLocation,
) {
	if val == nil || loc == nil {
		return
	}

	if loc.Postcode == "" {
		val.Valid = false
		val.Verifiable = false

		val.Location.Postcode = append(
			val.Location.Postcode,
			listing.ValidationError_EMPTY_VALUE,
		)
	}

	if loc.City == "" {
		val.Valid = false
		val.Verifiable = false

		val.Location.City = append(
			val.Location.City,
			listing.ValidationError_EMPTY_VALUE,
		)
	}
}

func (srv *serviceImpl) validateVersionDetails(
	val *model.VersionValidation,
	det *model.VersionDetails,
) {
	if det == nil || val == nil {
		return
	}

	if det.AmenityFurniture == 0 {
		val.Valid = false
		val.Verifiable = false

		val.Details.AmenityFurniture = append(
			val.Details.AmenityFurniture,
			listing.ValidationError_EMPTY_VALUE,
		)
	}

	if len(det.AmenityOutsideSpace) == 0 {
		val.Valid = false
		val.Verifiable = false

		val.Details.AmenityOutsideSpace = append(
			val.Details.AmenityOutsideSpace,
			listing.ValidationError_EMPTY_VALUE,
		)
	}

	for _, v := range det.AmenityOutsideSpace {
		if v == 0 {
			val.Valid = false
			val.Verifiable = false

			val.Details.AmenityOutsideSpace = append(
				val.Details.AmenityOutsideSpace,
				listing.ValidationError_INVALID_VALUE,
			)

			break
		}
	}

	if len(det.AmenityParking) == 0 {
		val.Valid = false
		val.Verifiable = false

		val.Details.AmenityParking = append(
			val.Details.AmenityParking,
			listing.ValidationError_EMPTY_VALUE,
		)
	}

	for _, v := range det.AmenityParking {
		if v == 0 {
			val.Valid = false
			val.Verifiable = false

			val.Details.AmenityParking = append(
				val.Details.AmenityParking,
				listing.ValidationError_INVALID_VALUE,
			)

			break
		}
	}
}

func (srv *serviceImpl) validateVersionRental(
	val *model.VersionValidation,
	rent *model.VersionRental,
) {
	if val == nil || rent == nil {
		return
	}

	if rent.RentalPrice.GetValue() <= 0 {
		val.Valid = false
		val.Verifiable = false

		val.Rental.RentalPrice = append(
			val.Rental.RentalPrice,
			listing.ValidationError_INVALID_VALUE,
		)
	}
}

func (srv *serviceImpl) validateVersionPeople(
	ctx context.Context,
	val *model.VersionValidation,
	peop *model.VersionPeople,
) {
	if val == nil || peop == nil {
		return
	}

	p, err := srv.Connections.Account.Get(ctx, &account.GetRequest{
		Id: peop.Landlord,
	})

	if p == nil || err != nil {
		val.Valid = false
		val.Verifiable = false

		val.People.Landlord = append(
			val.People.Landlord,
			listing.ValidationError_INVALID_VALUE,
		)
	}
}

func (srv *serviceImpl) validateVersionDescription(
	ctx context.Context,
	val *model.VersionValidation,
	desc *model.VersionDescription,
	peop *model.VersionPeople,
	propertyID string,
) {
	if val == nil || desc == nil || peop == nil || propertyID == "" {
		return
	}

	if len(desc.Photos) == 0 {
		invs, err := srv.Connections.Invoice.InvoiceList(ctx, &invoice.InvoiceListRequest{
			User:     peop.Landlord,
			Property: propertyID,
			Statuses: []invoice.InvoiceStatus{
				invoice.InvoiceStatus_INVOICE_FULFILLED,
			},
			Types: []extras.ServiceType{
				extras.ServiceType_PHOTOS_FLOOR_PLAN,
			},
		})
		if err != nil || len(invs.GetInvoices()) == 0 {
			val.Valid = false

			val.Description.Photos = append(
				val.Description.Photos,
				listing.ValidationError_EMPTY_VALUE,
			)
		}
	}
}

func (srv *serviceImpl) validateVersionDocuments(
	ctx context.Context,
	val *model.VersionValidation,
	peop *model.VersionPeople,
	propertyID string,
) {
	if val == nil || peop == nil || propertyID == "" {
		return
	}

	exdocs, _ := srv.Connections.Document.List(ctx, &document.ListRequest{
		User:     peop.Landlord,
		Property: propertyID,
		TypeCodes: []string{
			document.TypeCertificate,
		},
	})

	found := false

	for _, d := range exdocs.GetDocuments() {
		if d.Code == "landlord_epc_certificate" {
			found = true
			break
		}
	}

	if found {
		return
	}

	invs, _ := srv.Connections.Invoice.InvoiceList(ctx, &invoice.InvoiceListRequest{
		User:     peop.Landlord,
		Property: propertyID,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
		Types: []extras.ServiceType{
			extras.ServiceType_CERTIFICATE,
		},
	})

	for _, i := range invs.GetInvoices() {
		for _, s := range i.GetServices() {
			if s.GetService() != nil && s.GetService().GetCode() == "landlord_epc_certificate" {
				found = true

				break
			}
		}

		if found {
			break
		}
	}

	if found {
		return
	}

	val.Valid = false

	val.Documents.EPC = append(
		val.Documents.EPC,
		listing.ValidationError_EMPTY_VALUE,
	)
}
