// Package listed provides ensured publication of listed properties
package listed

import (
	"context"
	"time"

	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service/render"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
)

// Config for a task
type Config struct {
	Repository  model.Repository
	Connections connections.Publications
}

// New returns new task instance
func New(config *Config) (schedule.Handler, error) {
	err := config.Connections.Validate()
	if err != nil {
		return nil, err
	}

	return &handlerImpl{
		Config: *config,
	}, nil
}

type handlerImpl struct {
	Config
}

func (h *handlerImpl) Name() string {
	return "publication-sync-listed"
}

func (h *handlerImpl) Consume(
	ctx context.Context,
	req *schedule.Reminder_ConsumeRequest,
	data []byte,
	now time.Time,
) error {
	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	go h.syncExternal(bgctx, h.Connections.Rightmove, listing.PublicationPlatform_PUBLICATION_PLATFORM_RIGHTMOVE)
	go h.syncExternal(bgctx, h.Connections.Boomin, listing.PublicationPlatform_PUBLICATION_PLATFORM_BOOMIN)
	go h.syncExternal(bgctx, h.Connections.Zoopla, listing.PublicationPlatform_PUBLICATION_PLATFORM_ZOOPLA)

	return nil
}

func (h *handlerImpl) syncExternal(
	ctx context.Context,
	external listing.ListingPublicationServiceClient,
	platform listing.PublicationPlatform,
) {
	entries, err := h.Repository.VersionPublicationEntryListUnsynced(
		ctx,
		platform,
	)
	if err != nil {
		log.WithError(err).Errorf("Getting %v unsynced listing list", platform)
	}

	for _, e := range entries {
		var resp *listing.ListingPublicationSaveResponse

		var ver *model.Version

		ver, err = h.Repository.VersionGet(ctx, &model.VersionSpecOptions{
			ListingID: e.ListingID,
			Statuses:  []listing.VersionStatus{listing.VersionStatus_VERSION_VERIFIED},
		})
		if err != nil {
			log.WithError(err).WithField("listing", e.ListingID).Error("getting version")

			continue
		}

		resp, err = external.ListingPublicationSave(ctx, &listing.ListingPublicationSaveRequest{
			PublicationId: e.PublicationID,
			PropertyId:    e.PropertyID,
			ListingId:     e.ListingID,
			PropertyIvr:   ver.IVR,
			Available:     !e.LetAgreed,
			Location:      render.Location(&ver.Location, true),
			Rental:        render.Rental(&ver.Rental, true),
			Details:       render.Details(&ver.Details),
			Description:   render.Description(&ver.Description),
		})
		if err != nil || resp.GetPublicationId() == "" || resp.GetPublicationUrl() == "" {
			log.WithError(err).WithField("property", e.PropertyID).Errorf("publishing on %v", platform)

			continue
		}

		e.SyncedAt = model.ServiceTime(time.Now())
		e.PublicationID = resp.GetPublicationId()
		e.PublicationURL = resp.GetPublicationUrl()

		err = h.Repository.VersionPublicationEntrySave(ctx, e)
		if err != nil {
			log.WithError(err).WithField("property", e.PropertyID).Errorf("saving %v publication", platform)
		}
	}
}
