// Package unlisted provides ensured unpublication of unlisted properties
package unlisted

import (
	"context"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/service/publication"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
)

// Config for a task
type Config struct {
	Repository  model.Repository
	Connections connections.Publications
}

// New returns new task instance
func New(config *Config) (schedule.Handler, error) {
	err := config.Connections.Validate()
	if err != nil {
		return nil, err
	}

	return &handlerImpl{
		Config: *config,
	}, nil
}

type handlerImpl struct {
	Config
}

func (h *handlerImpl) Name() string {
	return "publication-sync-unlisted"
}

func (h *handlerImpl) Consume(
	ctx context.Context,
	req *schedule.Reminder_ConsumeRequest,
	data []byte,
	now time.Time,
) error {
	bgctx := grpcctx.WithServiceUser(grpcctx.ValuesBackgroundContext(ctx))

	go h.syncExternal(bgctx, h.Connections.Rightmove, listing.PublicationPlatform_PUBLICATION_PLATFORM_RIGHTMOVE)
	go h.syncExternal(bgctx, h.Connections.Boomin, listing.PublicationPlatform_PUBLICATION_PLATFORM_BOOMIN)
	go h.syncExternal(bgctx, h.Connections.Zoopla, listing.PublicationPlatform_PUBLICATION_PLATFORM_ZOOPLA)
	go h.syncLet(bgctx)

	return nil
}

func (h *handlerImpl) syncExternal(
	ctx context.Context,
	external listing.ListingPublicationServiceClient,
	platform listing.PublicationPlatform,
) {
	resp, err := external.ListingPublicationList(ctx, &listing.ListingPublicationListRequest{})
	if err != nil {
		log.WithError(err).Errorf("Getting %v listing list", platform)
	}

	for {
		var ls *listing.ListingPublicationListResponse

		ls, err = resp.Recv()
		if err != nil {
			break
		}

		for _, l := range ls.GetListings() {
			var res bool

			res, err = h.Repository.VersionPublicationEntryPublished(
				ctx,
				platform,
				l.GetPropertyId(),
			)
			if err != nil {
				log.WithError(err).WithField("property", l.GetPropertyId()).
					Errorf("checking published state for %v", platform)

				continue
			}

			if !res {
				_, err = external.ListingPublicationRemove(ctx, &listing.ListingPublicationRemoveRequest{
					PropertyId: l.GetPropertyId(),
				})
				log.WithError(err).WithField("property", l.GetPropertyId()).
					Errorf("deleting unpublished property from %v", platform)
			}
		}
	}
}

func (h *handlerImpl) syncLet(ctx context.Context) {
	pids, err := h.Repository.VersionPublicationLetPropertyIDs(
		ctx,
	)
	if err != nil {
		return
	}

	for _, p := range pids {
		err = publication.VersionUnpublishEverywhere(
			ctx,
			h.Repository,
			&h.Connections,
			p,
			listing.VersionStatus_VERSION_LET,
		)
		if err != nil {
			log.WithError(err).WithField("property", p).
				Errorf("unpublishing let property")
		}
	}
}
