// Package unviewed handles schedule for listings having few viewings
package unviewed

import (
	"context"
	"encoding/json"
	"errors"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"

	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	gmailparser "gitlab.mashroom.com/mashroom/backend/gmailparser/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	"gitlab.mashroom.com/mashroom/backend/listing/internal/model"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	viewing "gitlab.mashroom.com/mashroom/backend/viewing/api"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// Config for a task
type Config struct {
	Repository   model.Repository
	Viewing      viewing.ViewingServiceClient
	Offer        offer.OfferServiceClient
	Gmailparser  gmailparser.GmailparserServiceClient
	Invoice      invoice.InvoiceServiceClient
	Notification notification.NotificationServiceClient
}

// New returns new task instance
func New(config *Config) (schedule.Handler, error) {
	return &handlerImpl{
		Config: *config,
	}, nil
}

type handlerImpl struct {
	Config
}

func (h *handlerImpl) Name() string {
	return "publication-unviewed"
}

func (h *handlerImpl) Consume(
	ctx context.Context,
	req *schedule.Reminder_ConsumeRequest,
	data []byte,
	now time.Time,
) error {
	go h.process(now)

	return nil
}

func (h *handlerImpl) process(now time.Time) {
	since := now.AddDate(0, 0, -7)

	bgctx := grpcctx.ValuesBackgroundContext(context.Background())

	pids, err := h.Repository.VersionPublicationPropertyListingIDs(
		bgctx,
		listing.PublicationPlatform_PUBLICATION_PLATFORM_MASHROOM,
		since,
	)
	if err != nil {
		return
	}

	after := timestamppb.New(since)

	for len(pids) > 0 {
		ps := pids

		if len(ps) > 100 {
			ps, pids = pids[:100], pids[100:]
		} else {
			pids = nil
		}

		h.processSlice(bgctx, after, ps)
	}
}

func (h *handlerImpl) processSlice(ctx context.Context, after *timestamppb.Timestamp, ps []model.PropertyListingID) {
	viewings, err := h.Viewing.ServiceViewingCount(ctx, &viewing.ServiceViewingCountRequest{
		Selector: &viewing.ServiceViewingCountRequest_Properties{
			Properties: &viewing.PropertyListingArrayValue{
				Value: func() (res []*viewing.PropertyListing) {
					for _, p := range ps {
						res = append(res, &viewing.PropertyListing{
							Property: p.PropertyID,
							Listing:  p.ListingID,
						})
					}

					return
				}(),
				After: after,
			},
		},
	})
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return
	}

	offers, err := h.Offer.ServiceOfferCount(ctx, &offer.ServiceOfferCountRequest{
		Selector: &offer.ServiceOfferCountRequest_Properties{
			Properties: &offer.PropertyListingArrayValue{
				Value: func() (res []*offer.PropertyListing) {
					for _, p := range ps {
						res = append(res, &offer.PropertyListing{
							Property: p.PropertyID,
							Listing:  p.ListingID,
						})
					}

					return
				}(),
				After: after,
			},
		},
	})
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return
	}

	enquiries, err := h.Gmailparser.ServiceEnquiryCount(ctx, &gmailparser.ServiceEnquiryCountRequest{
		Selector: &gmailparser.ServiceEnquiryCountRequest_Properties{
			Properties: &gmailparser.PropertyListingArrayValue{
				Value: func() (res []*gmailparser.PropertyListing) {
					for _, p := range ps {
						res = append(res, &gmailparser.PropertyListing{
							Property: p.PropertyID,
							Listing:  p.ListingID,
						})
					}

					return
				}(),
				After: after,
			},
		},
	})
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return
	}

	invoices, err := h.Invoice.ServiceCount(ctx, &invoice.ServiceCountRequest{
		Selector: &invoice.ServiceCountRequest_Properties{
			Properties: &invoice.ServiceCountProperties{
				Properties: func() (res []string) {
					for _, p := range ps {
						res = append(res, p.PropertyID)
					}

					return
				}(),
				ServiceCodes:  []string{extras.ProductLandlordPhotographsFloorPlan.String()},
				OrderStatuses: []invoice.InvoiceStatus{invoice.InvoiceStatus_INVOICE_FULFILLED},
			},
		},
	})
	if err != nil {
		return
	}

	for _, p := range ps {
		ver, _ := h.Repository.VersionGet(ctx, &model.VersionSpecOptions{
			PropertyID: p.PropertyID,
			ListingID:  p.ListingID,
		})

		if ver == nil {
			continue
		}

		v := viewings.GetCounts()[p.PropertyID]
		o := offers.GetCounts()[p.PropertyID]
		e := enquiries.GetCounts()[p.PropertyID]
		i := invoices.GetCounts()[p.PropertyID]

		total := v + o + e

		data := map[string]interface{}{
			"first_week_performance": map[string]interface{}{
				"viewings":                v,
				"offers":                  o,
				"enquiries":               e,
				"is_well_performing":      total > 20,
				"rent_too_high":           ver.Rental.RentalPrice.GetValue() > ver.Valuation.UpperBound,
				"has_professional_photos": i > 0,
			},
		}

		var bs []byte

		bs, err = json.Marshal(data)
		if err != nil {
			continue
		}

		_, err = h.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_PROPERTY_VIEWING_RARITY,
			Action: notification.Action_ACTION_REMIND,
			Targets: []*notification.Target{
				{
					User: ver.People.Landlord,
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: p.PropertyID,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(bs),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("error sent to landlord: %v from property: %v", ver.People.Landlord, p)
		}

		err = h.Repository.VersionNotificationUpdate(
			ctx,
			&model.VersionSpecOptions{PropertyID: p.PropertyID},
			model.VersionNotificationViewingRarity,
			true,
		)
		if err != nil {
			log.WithError(err).Errorf("error update version notification flag for property: %v", p)
		}
	}
}
